/***************************************************************************
BEGIN_FILE_PROLOG:

        COPYRIGHT @ 2000, Raytheon Systems Company, its vendors,
        and suppliers, ALL RIGHTS RESERVED.


       ==========================================================================
       The HDF object/attribute converters for conversion of
       hybrid files are taken from the HDF Group's HDF4 to hdf5
       converter H4H5TOOLS source code and been adopted to this
       converter's working.

         Copyright Notice and License Terms for 
         H4H5TOOLS Software Library and Utilities
         -----------------------------------------------------------------------------

         H4H5TOOLS Software Library and Utilities
         Copyright 2006-2010 by The HDF Group.

         NCSA H4H5TOOLS Software Library and Utilities
         Copyright 1998-2006 by the Board of Trustees of the University of Illinois.

         All rights reserved.

         Contributors:  National Center for Supercomputing Applications (NCSA) at 
         the University of Illinois.
         Redistribution and use in source and binary forms, with or without 
         modification, are permitted for any purpose (including commercial purposes) 
         provided that the following conditions are met:

         1. Redistributions of source code must retain the above copyright notice, 
            this list of conditions, and the following disclaimer.

         2. Redistributions in binary form must reproduce the above copyright notice, 
            this list of conditions, and the following disclaimer in the documentation 
            and/or materials provided with the distribution.

         3. In addition, redistributions of modified forms of the source or binary 
            code must carry prominent notices stating that the original code was 
            changed and the date of the change.

         4. All publications or advertising materials mentioning features or use of 
            this software are asked, but not required, to acknowledge that it was 
            developed by The HDF Group and by the National Center for Supercomputing 
            Applications at the University of Illinois at Urbana-Champaign and 
            credit the contributors.

         5. Neither the name of The HDF Group, the name of the University, nor the 
            name of any Contributor may be used to endorse or promote products derived 
            from this software without specific prior written permission from The HDF
            Group, the University, or the Contributor, respectively.

         DISCLAIMER: 
         THIS SOFTWARE IS PROVIDED BY THE HDF GROUP AND THE CONTRIBUTORS "AS IS" 
         WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED.  In no event
         shall The HDF Group or the Contributors be liable for any damages suffered 
         by the users arising out of the use of this software, even if advised of 
         the possibility of such damage. 

         ==============================================================================

FILENAME:
        convert.c

DESCRIPTION:
        This file contains the functions that make up the command-
	line utility to convert HDF-EOS objects from HDF4 to HDF5.
        The HDF object/attribute converters for conversion of
        hybrid files are taken from the HDF Groups HDF4 to hdf5
        converter source code and been adopted this converter's
        working.

AUTHOR:
        Ray Milburn / Emergent Information Technologies, Inc.

HISTORY:
        15-Jun-00 RM                 Initial version
         1-Aug-08 Abe Taaheri        Modified for correcting compression, 
                                     unlimited dimension, ... This version
                                     integrated into HDF-EOS Plug-in for
                                     HDFView 2.4
         9-Mar-11 Abe Taaheri        Modified for correcting group/field,
                                     attributes, fillvalue, compression,
                                     hdf written fields
         2-Apr-11 Abe Taaheri        Modified for writing Core and Archive
                                     Metadata from he2 to he5 file

         20-Apr-11 Abe Taaheri       Modified for writing dimension scales
                                     from he2 to he5 file
         20-Apr-12 Abe Taaheri       Modified for preserving original chunk
                                     sizes
         30-Apr-12 Abe Taaheri       Modified for writing more Global attributes, 
                                     and handling hybrid files containing swath
                                     or HDF4 written SDS
END_FILE_PROLOG:
***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>

#include "hdf.h"
#include "HdfEosDef.h"

#include "hdf5.h"
#include <assert.h>

#define HECONVERT 1
#include "HE5_HdfEosDef.h"

#define HAVE_LIBHDFEOS

/***************************************************************************
  Defines and data structure needed for the utiltity.
***************************************************************************/
#define CONVERT_VERSION "Version 5-1.8.8, 01-May-12"

#define CONVERT_FALSE 0
#define CONVERT_TRUE  1

#define CONVERT_SWATH 10
#define CONVERT_GRID  20
#define CONVERT_POINT 30
#define CONVERT_META  40

#define CONVERT_SUCCESS   0
#define CONVERT_FAIL     -1
#define CONVERT_PROBLEM  -2

#define CONVERT_SWATH_DIMS      100
#define CONVERT_GRID_DIMS       200
#define CONVERT_SWATH_GEOFIELD  300
#define CONVERT_SWATH_DATAFIELD 310
#define CONVERT_SWATH_ATTRIBUTE 320
#define CONVERT_SWATH_DIMSCALE  330
#define CONVERT_GRID_DATAFIELD  400
#define CONVERT_GRID_ATTRIBUTE  410
#define CONVERT_GRID_DIMSCALE   420
#define CONVERT_POINT_ATTRIBUTE 500
#define WARNING                 1000
#define ERROR_READ_INPUT        2000
#define ERROR_MEMORY            3000

#define CONVERT_NO_TILE -99

#define CONVERT_VERBOSE_HEAD "CONVERT"
#ifdef  MAX_NC_NAME
#define HE4HE5_MAX_NC_NAME MAX_NC_NAME
#else 
#define HE4HE5_MAX_NC_NAME H4_MAX_NC_NAME
#endif

#ifdef  MAX_VAR_DIMS
#define HE4HE5_MAX_VAR_DIMS MAX_VAR_DIMS
#else
#define HE4HE5_MAX_VAR_DIMS H4_MAX_VAR_DIMS  /* = 32 */
#endif
#define HE4_MAX_VAR_DIMS MAX_VAR_DIMS



#define HDF4_VDATA_CLASS "CLASS"
#define _HDF_CHK_TBL_CLASS "_HDF_CHK_TBL_"
#define VDATALABEL  "Vdata"
#define HDF4_REF_NUM        "HDF4_REF_NUM"
#define HDF4_VDATA_VERSION_ATTR "VERSION"
#define HDF4_OBJECT_NAME    "HDF4_OBJECT_NAME"
#define H425_NONEWATTRS  2
#define HDF4_VDATA          "HDF4_VDATA"
#define H425_ALLATTRS    1
#define MAXREF_LENGTH    32
#define OBJECT_HASHSIZE  128
#define VD_HASHSIZE      64
#define HDF5_TABLE_TITLE "TITLE"
#define HDF4_OBJECT_TYPE    "HDF4_OBJECT_TYPE"
#define HDF5TABLE_SPE_VERSION "1.0"
#define VDATA_CLASS "TABLE"
#define ORI_SLASH '/'
#define CHA_SLASH '_'
#define HE4HE5_DEFAULT_NUMELEM_IN_CHUNK  512
#define HDF4_SDS            "HDF4_SDS"
#define HDF4_PALETTE        "HDF4_PALETTE"
#define HDF4_IMAGE          "HDF4_IMAGE"
#define HDF4_VGROUP         "HDF4_VGROUP"
#define ATTR "ATTR"
#define H425_ALLATTRS    1

typedef struct
{
  int   totalNumAttr; /*total number of attributes for this dim scale */
  int   sds_buffsize; /* buffer size of data read for the dim scale */
  int32 sds_dtype;    /* hdf4 SDS data type */
  int32 dimsize;
  void* sds_data;      /* data read for the dim scale */
  char  name[256];    /* dimscale name */
  char *sdsattr_name[100];
  void* sds_attdata[100];      /* attribute data*/
  hid_t he5_atype[100];         /* data type of each attribute */
} Hdf4DimScaleInfo;


typedef struct
{
    int total;
    char **name;
} HdfInfo;

typedef struct
{
    int total_hdfeos_Used_SDS;
    char *Used_SDS[1000];
} Hdf5Info;

Hdf5Info hdfeos5Info;

/*** adding a macro for detecting overflow when casting between types, 
     following HDF5 library approach***/

#define H425_CHECK_OVERFLOW(var,vartype,casttype) \
{                                                 \
  casttype _tmp_overflow=(casttype)(var);         \
  assert ((var)==(vartype)_tmp_overflow);          \
}

/***************************************************************************
  Global variables
***************************************************************************/
char inNameGlobal[1025];     /* input file */
char outNameGlobal[1025];    /* output file */
int verboseModeGlobal;       /* verbose mode flag */
int compressModeGlobal;      /* compression mode flag */
int numTilesGlobal;          /* number of tiles (for compression) */
int fileCreatedGlobal;	     /* already created output file */

/***************************************************************************
  Function prototypes (all contained in this file)
***************************************************************************/
int ConvertEOSGlobalMetadata(char *inFileName, char *outFileName);
intn GetEOSMetadata(int32 sd_id, int out_id, char *inFileName, 
		    char *outFileName);
intn ConvertEOSAttr(int32 fid_in, int fid_out, char *attr);
intn ConvertOtherGlobalAttr(int32 sd_id, int fid_out, 
			    char *inFileName,char *outFileName);

int GetInput(int argc, char *argv[]);
void DisplayHelpAndExit(char *progName);
void DisplayVersionAndExit(char *progName);
int GetNumberObjects(HdfInfo *infoPtr,int objType);
int DoSwathConversion(HdfInfo swathInfo);
int DoGridConversion(HdfInfo swathInfo);
int DoPointConversion(HdfInfo swathInfo);
int ConvertSWorGDDims(int32 inObj,int outObj,int typeFlag);
int ConvertSwathDimMap(int32 inFile,int outFile);
int ConvertSwathIndexMap(int32 inFile,int outFile);
int ConvertSWorGDFieldInfo(int32 inSG,hid_t outSG,int outFile,
			   char *SGname,int typeFlag);
int ConvertHDFEOSAttribute(int32 inObj,int outObj,int typeFlag);
int ConvertFieldAttribute(int32 inObj,int outObj,int typeFlag);
int GetHDF5TypeFromHDF4Type(int32 hdf4Type);
int ConvertGridProjInfo(int32 inFile,int outFile);
void CalculateNumberOfChunks(int32 inSG, char *fieldname, 
			     int numDims,hsize_t *realDims,
                             hsize_t *chunkDims);
void FlushSwath(int outFile,hid_t *outSwath, char *SWname);
void FlushGrid(int outFile, hid_t *outGrid, char *GDname);
int defineUnlimitedDims(int32 inSG, hid_t outSG, char *fieldName, char *dimList, 
 			char *maxdimList, int32 dim[], int typeFlag);

int RecoverOutputFid(hid_t HDFfid, int *fid_out, char *outFileName);
char* make_objname_no(
		      char* refstr,
		      const char* path_name,
		      const char*objstr);
char* make_objname_yes(
		       char* obj_name,
		       const char* path_name);
char* get_obj_aboname(
		      char* obj_name,
		      char* refstr,
		      const char* path_name,
		      const char*objstr,/*
					  struct name_table*name_hashtab,*/
		      int tablesize,
		      int* check_nameptr );
char *correct_name(const char* oldname);
void ZeroMemory(void*s,size_t n);
char* get_h5datasetname(
			const char* h5groupname,
                        const char* h5dsetname, 
			uint16 obj_ref,
                        const char* hdf4obj,
			int tablesize,
			int* check_nameptr);
int He4toHe5vdata(int32 file_id,
		  int32 vdata_id,
		  char* outFileName,
		  char* h5groupname, 
		  char* h5dsetname,
		  int   attr_flag);
char* trans_obj_name(
		     int32 obj_tag,
		     int32 h4toh5_index);
int conv_int_str(
		 uint32 num, 
		 char* str_num);
char *correct_name2(const char* oldname);

/* for some reason prototype for isdigit causes syntax error in linux, so
   we will not use prototype in linux */

#if defined(LINUX) || defined(LINUX32) || defined(LINUX64) 

# else
int isdigit(int);
#endif
/*
#ifndef LINUX
int isdigit(int);
# else

#endif
*/
int CallocErrorMsg(int);
int ConvertPointLevelData(int32 inPTid, int outPTid);
int GetHDF5ClassFromHDF4Type(int32);

void strupr_S(char *str);
void strlwr_S(char *str);

hid_t H5Acreate_safe(hid_t loc_id, const char *name, hid_t type_id, 
		     hid_t space_id, hid_t acpl_id);
hid_t mkstr(size_t size, H5T_str_t pad);
int  h4type_to_h5type(const int32 h4type, 
		      hid_t* h5memtype,
		      size_t* h4memsize,
		      size_t* h4size, 
		      hid_t *h5type);
int  H4toH5datatype(const   int32 h4type, 
		    hid_t*  h5typeptr,
		    hid_t*  h5memtypeptr,
		    size_t* h4sizeptr,
		    size_t* h4memsizeptr);
int CopyDimscalesdsToHe5(char *outFileName, int32 sds_id);
int h4_transnumattr( 
		    hid_t h5g,
		    const char *refname,
		    uint16 group_ref);

/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
          separating()
***************************************************************************/
void separating(char *stringlist, int32 *nstring, char *strings[], char strs[])
{
  char		*astring;
  
  *nstring=0;
  astring=strtok(stringlist, strs);
  while (astring!=NULL)
    {
      strings[*nstring]=astring;  
      astring=strtok(NULL, strs);
      *nstring=*nstring+1;
    } 
}


int32  checkobjNameForUlimitedSize(int32 inSG, char *fieldName)
{
  intn status;
  int32 sdInterfaceID=-1;
  int32 sds_id;
  int32 actualdims[HE4_MAX_VAR_DIMS];
  char            name[1024];
  int32           data_type=0;
  int32           size;
  int32           dimid;	/* SDS dimension ID */
  int32           dum;	        /* Dummy variable */
  int32           fid;

  /* Get HDF-EOS file ID and SDS interface ID */
  status = SWchkswid(inSG, "checkobjNameForUlimitedSize", &fid, &sdInterfaceID, &dum);
  if(status != 0)
    {
      return(status);
    }
  
  status = SWSDfldsrch(inSG, sdInterfaceID, fieldName,
		       &sds_id, &dum, &dum, &dum, actualdims,
		       &dum);

  /* only consider first dim for sds, since in HDF4
     that is the only dim than can be unlimited */

  dimid = SDgetdimid(sds_id, 0);
  SDdiminfo(dimid, name, &size, &data_type, &dum);

    if(size == 0 && actualdims[0] > 0) /* dimension is unlimited */
    {
      return(1);
    }
  else
    {
      return(0);
    }
}



int32 createDimsInMaxdimlist(hid_t outSG, char *maxdimList)
{
  char *tempstr = NULL;
  tempstr = (char *)malloc((strlen(maxdimList)+1)*sizeof(char));
  if(tempstr != NULL)
    {
      int i=0;
      int32 foundDim = 0;
      int32 status;
      int32 numDims;                      /* number of dimensions */
      hsize_t dims[32];
       int32 count, countdims;
      int32 bufSize;                      /* string buffer size */
      int retVal = CONVERT_SUCCESS;
      char **ptr2 = NULL;
      char **ptr3 = NULL;
      size_t *slen2 = NULL;
      size_t *slen3 = NULL;
      
      char *dimName= NULL, *swathdimlist_dimname= NULL, *maxdimlist_dimname= NULL;

      dimName = (char *)malloc(1800 * sizeof(char));
      swathdimlist_dimname = (char *)malloc(500 * sizeof(char));
      maxdimlist_dimname = (char *)malloc(500 * sizeof(char));

      strcpy(tempstr, maxdimList);

      count = HE5_EHparsestr(tempstr,',',NULL,NULL);
      if( count == 0 ) return(-1);
      ptr2 = (char **) calloc(count,sizeof(char *));
      if (ptr2 == NULL)
	{
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return (CallocErrorMsg(__LINE__));
	}
      slen2 = (size_t *) calloc(count,sizeof(long));
      if (slen2 == NULL)
	{
	  free(ptr2);
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return (CallocErrorMsg(__LINE__));
	}
      
      count = HE5_EHparsestr(tempstr,',',ptr2,slen2);

      /* check to see if first dim in maxdimlist exist. 
	 I HDF4 only first dim can be unlimited
	 and this functions only deals with unlimited dimensions */

      memmove(maxdimlist_dimname,ptr2[0],(int) slen2[0]);
      maxdimlist_dimname[slen2[0]] = '\0';
      
      numDims = HE5_SWinqdims(outSG,dimName,dims);

      countdims = HE5_EHparsestr(dimName,',',NULL,NULL);
     if( countdims == 0 )
	{
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return(-1);
	}
      ptr3 = (char **) calloc(countdims,sizeof(char *));
      if (ptr3 == NULL)
	{
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return (CallocErrorMsg(__LINE__));
	}
      slen3 = (size_t *) calloc(countdims,sizeof(long));
      if (slen3 == NULL)
	{
	  free(ptr3);
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return (CallocErrorMsg(__LINE__));
	}
      countdims = HE5_EHparsestr(dimName,',',ptr3,slen3);
      for(i=0; i<countdims; i++)
	{
	  memmove(swathdimlist_dimname,ptr3[i],(int) slen3[i]);
	  swathdimlist_dimname[slen3[i]] = '\0';

	  if(strcmp(swathdimlist_dimname,maxdimlist_dimname) == 0)
	    {
	      /* dimension exist. break */
	      foundDim = 1;
	      break;
	    }
	}
      
      if(foundDim == 1)
	{
	  free(ptr3);
	  free(ptr2);
	  free(slen3);
	  free(slen2);
	  free(tempstr);
	  free(dimName);
	  free(swathdimlist_dimname);
	  free(maxdimlist_dimname);
	  return(0);
	}
      else
	{
	  status = HE5_SWdefdim(outSG, maxdimlist_dimname, H5S_UNLIMITED);
	}

      if(ptr3 != NULL )
	{
	  free(ptr3);
	  ptr3 = NULL;
	}
     if(ptr2 != NULL )
	{
	  free(ptr2);
	  ptr2 = NULL;
	}
     if(slen3 != NULL )
	{
	  free(slen3);
	  slen3 = NULL;
	}
     if(slen2 != NULL )
	{
	  free(slen2);
	  slen2 = NULL;
	}
     if(tempstr != NULL )
	{
	  free(tempstr);
	  tempstr = NULL;
	}

     free(dimName);
     free(swathdimlist_dimname);
     free(maxdimlist_dimname);
    }
  else
    {
      free(tempstr);
      return(-1);
    }
  free(tempstr);
  return(0);
}


  /*-------------------------------------------------------------------
   * Function:	getHDF4Dimscaleattrs
   *
   * Purpose:     gets attributes of HDF4 dimscale sds (written by native HDF4)
   *             
   *              
   * Return:	FAIL if failed, SUCCEED if successful.
   *
   * In :	        
   sds_id:        SDS identifier for HDF4 sds
   DimscaleInfo   to hold info for dimscale

  Note: H4 to H5 (of The HDF Group) translaor, and some of its utilities 
        were used with some appropriate modifications to comnvert HDF4 
        wirtten SDSs and their attributes to HDF-EOS5 attributes
   *------------------------------------------------------------------
   */	
herr_t getHDF4Dimscaleattrs(int32 sds_id, Hdf4DimScaleInfo *DimscaleInfo)
{
  char*     sdstempattr_name;                   /* a temporary variable to 
						   store SDS attribute name */

  int32     count_sdsattdata; /* number of elements of SDS attributes.*/
  uint32    temp_countattdata;/* temporary variable to describe the number of
				 SDS attributes */
  hsize_t   h5_temp_countattdata[1];
  int32     sds_atype;        /* SDS attribute datatype */

  size_t    he4_amemsize;     /* size of attribute data type in memory. */
  size_t    he4_asize;        /* size of attribute data type in disk. */

  herr_t    ret=CONVERT_SUCCESS;  /* the temporary variable to check 
				     the "return" status  of
				     HDF5 APIs */
  int       i; 
  HdfInfo   hdf4Info;          /* temporary variable. */
  char indstr[10];
  char *temp;
  char *sdsattr_name=NULL;

  sdsattr_name = (char *)malloc(2*HE4HE5_MAX_NC_NAME * sizeof(char));

      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  HE5_EHset_error_on(1, 0); /* suppress only HDF5 errors */
	}
      else
	{
	  HE5_EHset_error_on(2, 0);/* suppress both HDFEOS5 and  HDF5 errors */
	}
 
  for (i = 0;i < DimscaleInfo->totalNumAttr; i++) {
    DimscaleInfo->sds_attdata[i] = NULL;

    if (SDattrinfo(sds_id,i,sdsattr_name,&sds_atype, &count_sdsattdata) == 
	CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot get attribute information for attribute %d of %s.", 
	     i,"HDF4 written dimension scale");
	}
      free(sdsattr_name);
      return CONVERT_FAIL;
    }

    /* convert HDF4 attribute data type to HDF-EOS5 attribute data type. */
    if((DimscaleInfo->he5_atype[i] = GetHDF5TypeFromHDF4Type(sds_atype))== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot get SDS attribute data type for sds dimscale %s.", 
	       DimscaleInfo->name);
	}
      free(sdsattr_name);
      return CONVERT_FAIL;
    }

    he4_amemsize = H5Tget_size(DimscaleInfo->he5_atype[i]);
    if(count_sdsattdata != 0) {
      /* allocating memory for SDS attribute data. */
      DimscaleInfo->sds_attdata[i] = malloc(he4_amemsize * count_sdsattdata+1);
      if(DimscaleInfo->sds_attdata[i] == NULL) {
	free(sdsattr_name);               
	return (CallocErrorMsg(__LINE__));
      }
      if(SDreadattr(sds_id,i,(VOIDP)DimscaleInfo->sds_attdata[i])== CONVERT_FAIL) {
	puts("cannot read SDS attribute.\n");
	free(DimscaleInfo->sds_attdata[i] );
	free(sdsattr_name);
	return CONVERT_FAIL;
      }

   if((DimscaleInfo->he5_atype[i] == H5T_NATIVE_CHAR) ||
       (DimscaleInfo->he5_atype[i] == H5T_NATIVE_SCHAR) ||
       (DimscaleInfo->he5_atype[i] == H5T_NATIVE_UCHAR) )
      {
	temp = (char *)DimscaleInfo->sds_attdata[i];
	temp[he4_amemsize * count_sdsattdata]='\0';
	DimscaleInfo->sds_attdata[i] = (void *)temp;
      }

      /*     DimscaleInfo->sds_attdata[i][he4_amemsize * count_sdsattdata]='\0';*/

      h5_temp_countattdata[0] = (hsize_t)count_sdsattdata;
    }
    else {/* treat the number of data as 1 for empty attribute*/

      temp_countattdata = 1;
      h5_temp_countattdata[0] = (hsize_t)temp_countattdata;
      strcpy(DimscaleInfo->sds_attdata[i] ,"");
      DimscaleInfo->he5_atype[i] = GetHDF5TypeFromHDF4Type(DFNT_CHAR);
       
      if((DimscaleInfo->he5_atype[i] == H5T_NATIVE_CHAR) ||
	 (DimscaleInfo->he5_atype[i] == H5T_NATIVE_SCHAR) ||
	 (DimscaleInfo->he5_atype[i] == H5T_NATIVE_UCHAR) )
	{
	  temp = (char *)DimscaleInfo->sds_attdata[i];
	  temp[he4_amemsize * count_sdsattdata]='\0';
	  DimscaleInfo->sds_attdata[i] = (void *)temp;
	}
 
      if(sdsattr_name[0] == '\0') {
	sdstempattr_name = (char *)calloc((strlen("HDF4_SDS")+strlen("ATTR")+8), sizeof(char));
	strcpy(sdstempattr_name,"HDF4_SDS");
	sprintf(indstr, "%d",i);
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,"ATTR");
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,indstr);
	strncpy(sdsattr_name,sdstempattr_name,strlen(sdstempattr_name));
	free(sdstempattr_name);
      }
    }

    DimscaleInfo->sdsattr_name[i] = (char *)malloc((strlen(sdsattr_name) + 1)*sizeof(char));
      if(DimscaleInfo->sdsattr_name[i] == NULL) {               
	return (CallocErrorMsg(__LINE__));
      }
     strcpy(DimscaleInfo->sdsattr_name[i],sdsattr_name);
  }

  if(ret != CONVERT_SUCCESS)
    {
      free(sdsattr_name);
      return CONVERT_FAIL;
    }

  HE5_EHset_error_on(2, 0);/* reset both HDFEOS5 and HDF5 error writing */
  if(sdsattr_name != NULL)
    {
      free(sdsattr_name);
      sdsattr_name = NULL;
    }
  return ret;
}

  /*-------------------------------------------------------------------
   * Function:	sds_convetattrs 
   *
   * Purpose:     converts attributes of HDF4 fields (written by native HDF4)
   *              into hdf-eos5 field attributes
   *              
   * Return:	FAIL if failed, SUCCEED if successful.
   *
   * In :	        
   sds_id:        SDS identifier for HDF4 sds
   he5_fieldname: hdf-eos5 field name
   numAttrs:      number of HDF4 sds attributes
   outObj:        swath or grid ID
   typeFlag:      swath or grid type

  Note: H4 to H5 (of The HDF Group) translaor, and some of its utilities 
        were used with some appropriate modifications to comnvert HDF4 
        wirtten SDSs and their attributes to HDF-EOS5 attributes
   *------------------------------------------------------------------
   */	

int sds_convertattrs(int outObj, 
		    int32 sds_id, 
		    char *he5_fieldname,
		    int   numAttrs,
		    int   typeFlag)
{
   char*     sdstempattr_name;                   /* a temporary variable to 
						   store SDS attribute name */

  int32     count_sdsattdata; /* number of elements of SDS attributes.*/
  uint32    temp_countattdata;/* temporary variable to describe the number of
				 SDS attributes */
  hsize_t   h5_temp_countattdata[1];
  int32     sds_atype;        /* SDS attribute datatype */

  size_t    he4_amemsize;     /* size of attribute data type in memory. */
  size_t    he4_asize;        /* size of attribute data type in disk. */

  hid_t     he5_atype;        /* HDF5 attribute data type in disk */
  void*     sds_attdata;      /* attribute data*/
  herr_t    ret=CONVERT_SUCCESS;  /* the temporary variable to check 
				     the "return" status  of
				     HDF5 APIs */
  int       i;
  HdfInfo hdf4Info;          /* temporary variable. */
  char indstr[10];
  char      *sdsattr_name=NULL;

  sdsattr_name = (char *)malloc(2*HE4HE5_MAX_NC_NAME * sizeof(char));

      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  HE5_EHset_error_on(1, 0); /* suppress only HDF5 errors */
	}
      else
	{
	  HE5_EHset_error_on(2, 0);/* suppress both HDFEOS5 and  HDF5 errors */
	}
 

  for (i = 0;i < numAttrs; i++) {
    sds_attdata = NULL;

    if (SDattrinfo(sds_id,i,sdsattr_name,&sds_atype, &count_sdsattdata) == 
	CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot start SD attribute information for attribute %d of field %s.", 
	     i,he5_fieldname);
	}
      free(sdsattr_name);
      return CONVERT_FAIL;
    }


    /* convert HDF4 attribute data type to HDF-EOS5 attribute data type. */
    if((he5_atype = GetHDF5TypeFromHDF4Type(sds_atype))== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert SDS attribute data type from H4 to He5 for attr. %d of field %s.", 
	       i,he5_fieldname);
	}
      free(sdsattr_name);
      return CONVERT_FAIL;
    }

    he4_amemsize = H5Tget_size(he5_atype);
    if(count_sdsattdata != 0) {
      /* allocating memory for SDS attribute data. */
      sds_attdata = malloc(he4_amemsize * count_sdsattdata + 1);
      if(sds_attdata == NULL) {
	free(sdsattr_name);  
	return (CallocErrorMsg(__LINE__));
      }

      if(SDreadattr(sds_id,i,(VOIDP)sds_attdata)== CONVERT_FAIL) {
	puts("cannot read SDS attribute.\n");
	free(sds_attdata);
	free(sdsattr_name);
	return CONVERT_FAIL;
      }

      if((he5_atype == H5T_NATIVE_CHAR) ||
	 (he5_atype == H5T_NATIVE_SCHAR) ||
	 (he5_atype == H5T_NATIVE_UCHAR) )
	{
	  char *temp;
	  temp = (char *)sds_attdata;
	  temp[he4_amemsize * count_sdsattdata]='\0';
	  sds_attdata = (void *)temp;
	}

      h5_temp_countattdata[0] = (hsize_t)count_sdsattdata;
    }
    else {/* treat the number of data as 1 for empty attribute*/
      temp_countattdata = 1;
      h5_temp_countattdata[0] = (hsize_t)temp_countattdata;
      sds_attdata = malloc(sizeof(char) + 1);
      strcpy(sds_attdata,"");
      he5_atype = GetHDF5TypeFromHDF4Type(DFNT_CHAR);
       
      /* if an attribute doesn't have a name, a default name is set. */
      if(sdsattr_name[0] == '\0') {
	sdstempattr_name = (char *)calloc((strlen("HDF4_SDS")+strlen("ATTR")+8), sizeof(char));
	strcpy(sdstempattr_name,"HDF4_SDS");
	sprintf(indstr, "%d",i);
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,"ATTR");
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,indstr);
	strncpy(sdsattr_name,sdstempattr_name,strlen(sdstempattr_name));
	free(sdstempattr_name);
      }
    }
 
    switch (typeFlag)
      {
      case CONVERT_SWATH_ATTRIBUTE:
        {
	  hid_t dummy_ntype;
	  hsize_t error_flag;
	  int err_level = 0;
	  hsize_t dummy_count;
	  /* first check that if attribute is already written */
	  /* if not, then write it to the field */


	  /* suppress hdfeos5 error for HE5_SWlocattrinfo to avoid group 
	     attribute error reports (too many misleading error message)*/
	  error_flag = 2;
	  error_flag = HE5_EHset_error_on((int) error_flag, err_level);
	  ret = HE5_SWlocattrinfo(outObj, he5_fieldname,sdsattr_name, 
				  &dummy_ntype, &dummy_count);
	  error_flag = 0;
	  error_flag = HE5_EHset_error_on((int) error_flag, err_level);
	  if (ret != CONVERT_SUCCESS) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf(" Setting attribute %s for field %s\n",sdsattr_name,he5_fieldname); 
	      }
	    ret = HE5_SWwritelocattr(outObj, he5_fieldname, 
				     sdsattr_name, he5_atype, 
				     h5_temp_countattdata, sds_attdata);
	  }
	}
	break;
      case CONVERT_GRID_ATTRIBUTE:
        {
	  hid_t dummy_ntype;
	  hsize_t dummy_count;
	  /* first check that if attribute is already written */
	  /* if not, then write it to the field */
	  ret = HE5_GDlocattrinfo(outObj, he5_fieldname,sdsattr_name, 
				  &dummy_ntype, &dummy_count);
	  if (ret != CONVERT_SUCCESS) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf(" Setting attribute %s for field %s\n",sdsattr_name,he5_fieldname); 
	      }
	    ret = HE5_GDwritelocattr(outObj, he5_fieldname, 
				     sdsattr_name, he5_atype, 
				     h5_temp_countattdata, sds_attdata);
	  }
	}
	break;
      case CONVERT_POINT_ATTRIBUTE:
        {
	  hid_t dummy_ntype;
	  hsize_t dummy_count;
	  /* first check that if attribute is already written */
	  /* if not, then write it to the field */
	  ret = HE5_PTlocattrinfo(outObj, he5_fieldname,sdsattr_name, 
				  &dummy_ntype,&dummy_count);
	  if (ret != CONVERT_SUCCESS) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf(" Setting attribute %s for field %s\n",sdsattr_name,he5_fieldname); 
	      }
	    ret = HE5_PTwritelocattr(outObj, he5_fieldname, 
				     sdsattr_name, he5_atype, 
				     h5_temp_countattdata, sds_attdata);
	  }
	}
	break;
      }
    
    if(ret != CONVERT_SUCCESS)
      {
	free(sdsattr_name);
	free(sds_attdata);
	return CONVERT_FAIL;
      }
    else
      {
	free(sds_attdata);
      }
  }

  free(sdsattr_name);
  HE5_EHset_error_on(2, 0);/* reset both HDFEOS5 and HDF5 error writing */
  return ret;
}


  /*-------------------------------------------------------------------
   * Function:	sds_HDF4convetattrs 
   *
   * Purpose:     converts attributes of HDF4 SDS (written by native HDF4)
   *              into hdf5 SDS attributes
   *              
   * Return:	FAIL if failed, SUCCEED if successful.
   *
   * In :	        
   sds_id:        SDS identifier for HDF4 sds
   h5csds_namee:  hdf5 SDS name
   numAttrs:      number of HDF4 sds attributes
   outObj:        swath or grid ID

  Note: H4 to H5 (of The HDF Group) translaor, and some of its utilities 
        were used with some appropriate modifications to comnvert HDF4 
        wirtten SDSs and their attributes to HDF-EOS5 attributes
   *------------------------------------------------------------------
   */	

int sds_HDF4convertattrs(hid_t h5_group, 
			 int32 sds_id, 
			 char *h5csds_name,
			 int   numAttrs)
{
   char*     sdstempattr_name;                   /* a temporary variable to 
						   store SDS attribute name */

  int32     count_sdsattdata; /* number of elements of SDS attributes.*/
  uint32    temp_countattdata;/* temporary variable to describe the number of
				 SDS attributes */
  hsize_t   h5_temp_countattdata[1];
  int32     sds_atype;        /* SDS attribute datatype */

  size_t    h4_amemsize;     /* size of attribute data type in memory. */
  size_t    h4_asize;        /* size of attribute data type in disk. */

  hid_t     h5_atype;        /* HDF5 attribute data type in disk */
  void*     sds_attdata;      /* attribute data*/
  herr_t    ret=CONVERT_SUCCESS;  /* the temporary variable to check 
				     the "return" status  of
				     HDF5 APIs */
  int       i;
  HdfInfo hdf4Info;          /* temporary variable. */
  char indstr[10];
  hid_t     h5_amemtype; /* HDF5 attribute data type in memory */
  hid_t     h5a_sid; /* HDF5 attribute space id */
  hid_t     h5str_type; /* HDF5 attribute data type in disk when the data type is string.*/
  hid_t     h5_dset;
  hid_t     h5str_memtype; /* HDF5 attribute data type in memory when the data type is string. */
  hid_t     h5a_id; /* HDF5 attribute id */
  herr_t    sret; /* the flag to check the "return" status of HDF5 APIs. */
  hsize_t   h5dims[HE4HE5_MAX_VAR_DIMS]; /* an array of size of each dimension of HDF5
					  dataset converted from SDS*/
  char      *sdsattr_name;


  sdsattr_name = (char *)malloc(2*HE4HE5_MAX_NC_NAME * sizeof(char));

      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  HE5_EHset_error_on(1, 0); /* suppress only HDF5 errors */
	}
      else
	{
	  HE5_EHset_error_on(2, 0);/* suppress both HDFEOS5 and  HDF5 errors */
	}
 
    h5_dset = H5Dopen(h5_group, h5csds_name);
    if(h5_dset < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot open HDF5 dataset %s to write attributes.\n", h5csds_name);
	}
      return CONVERT_FAIL;
    }
     
  for (i = 0;i < numAttrs; i++) {
    sds_attdata = NULL;

    if (SDattrinfo(sds_id,i,sdsattr_name,&sds_atype, &count_sdsattdata) == 
	CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot start SD attribute information for attribute %d of field %s.", 
	     i,sdsattr_name);
	}
      H5Dclose(h5_dset);
      return CONVERT_FAIL;
    }


    /* convert HDF4 attribute data type to HDF5 attribute data type. */
    if(h4type_to_h5type(sds_atype,&h5_amemtype,&h4_amemsize,
			&h4_asize,&h5_atype)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("\n Cannot convert SD attribute data type for %s.from H4 to H5", sdsattr_name);
	}
      H5Dclose(h5_dset);
      return CONVERT_FAIL;
    }

    if(count_sdsattdata != 0) {
      /* allocating memory for SDS attribute data. */
      sds_attdata = malloc(h4_amemsize * count_sdsattdata);
      if(sds_attdata == NULL) {               
	return (CallocErrorMsg(__LINE__));
      }

      if(SDreadattr(sds_id,i,(VOIDP)sds_attdata)== CONVERT_FAIL) {
	puts("cannot read SDS attribute.\n");
	free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }

      h5_temp_countattdata[0] = (hsize_t)count_sdsattdata;
    }
    else {/* treat the number of data as 1 for empty attribute*/
      temp_countattdata = 1;
      h5_temp_countattdata[0] = (hsize_t)temp_countattdata;
      strcpy(sds_attdata,"");

      if(h4type_to_h5type(sds_atype,&h5_amemtype,&h4_amemsize,
			  &h4_asize,&h5_atype)== CONVERT_FAIL) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("\n Cannot convert SD attribute data type from H4 to H5\n");
	  }
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
       
      /* if an attribute doesn't have a name, a default name is set. */
      if(sdsattr_name[0] == '\0') {
	sdstempattr_name = (char *)calloc((strlen("HDF4_SDS")+strlen("ATTR")+8), sizeof(char));
	strcpy(sdstempattr_name,"HDF4_SDS");
	sprintf(indstr, "%d",i);
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,"ATTR");
	strcat(sdstempattr_name,"_");
	strcat(sdstempattr_name,indstr);
	strncpy(sdsattr_name,sdstempattr_name,strlen(sdstempattr_name));
	free(sdstempattr_name);
      }
    }
 
     
    if (h5_atype == H5T_STRING) {

      h5a_sid = H5Screate(H5S_SCALAR);
      if (h5a_sid < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot create HDF5 attribute space\n");
	  }
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
     
      if(count_sdsattdata != 0){
	temp_countattdata = count_sdsattdata*h4_asize;
	/*H425_CHECK_OVERFLOW(temp_countattdata,uint32,size_t);*/
	if ((h5str_type = mkstr((size_t)temp_countattdata,
				 H5T_STR_SPACEPAD))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot make HDF5 string type\n");
	  }
	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}

	if ((h5str_memtype = mkstr(count_sdsattdata*h4_amemsize,
				    H5T_STR_SPACEPAD))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot make hdf5 memory string type\n");
	  }
	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}

	h5a_id = H5Acreate_safe(h5_dset,sdsattr_name,h5str_type,
			    h5a_sid,H5P_DEFAULT);

	if (h5a_id <0) {
	  /* H5Acreate() can fail just because this attribute is too long. This could be rejected,
	   * the following code tries to split it. */
	  /*
	  if (transattrs_split(h5_dset, sdsattr_name, h5a_sid, sds_attdata, count_sdsattdata) < 0) {
	  */
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot create hdf5 attribute id for attribute %s\n", sdsattr_name);
	    }

	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}
	else if (sds_attdata) {
	  sret = H5Awrite(h5a_id,h5str_memtype,(void *)sds_attdata);
	  if (sret <0) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("cannot write hdf5 attribute\n");
	      }
	    H5Sclose(h5a_sid);
	    H5Aclose(h5a_id);
	    if (sds_attdata) free(sds_attdata);
	    H5Dclose(h5_dset);
	    return CONVERT_FAIL;
	  }
	}
      }
      else {/* treat the number of data as 1 for empty attribute*/
	
	if ((h5str_type = mkstr((size_t)1*h4_asize,
				 H5T_STR_SPACEPAD))<0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot make HDF5 string type\n");
	    }
	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}

	if ((h5str_memtype = mkstr(1*h4_amemsize,
				    H5T_STR_SPACEPAD))<0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot make hdf5 memory string type\n");
	    }
	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}

	h5a_id = H5Acreate_safe(h5_dset,sdsattr_name,h5str_type,
			    h5a_sid,H5P_DEFAULT);
	if (h5a_id <0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot create hdf5 attribute id for attribute %s\n", sdsattr_name);
	    }
	  H5Sclose(h5a_sid);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}

	sret = H5Awrite(h5a_id,h5str_memtype,(void *)sds_attdata);
	if (sret <0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot write hdf5 attribute\n");
	    }
	  H5Sclose(h5a_sid);
	  H5Aclose(h5a_id);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}
      }
      sret = H5Sclose(h5a_sid);
      /* An HDF4 attribute could be converted into an HDF5 attribute. */
      if (h5a_id >= 0) sret = H5Aclose(h5a_id);
    }
	 
    else {
      if(count_sdsattdata == 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("the attribute data must not be none\n");
	  }
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
      h5dims[0] = count_sdsattdata;
      h5a_sid =  H5Screate_simple(1,h5dims,NULL);
      
      if (h5a_sid < 0)  {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot create HDF5 attribute space\n");
	  }
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
      
      h5a_id = H5Acreate_safe(h5_dset,sdsattr_name,h5_atype,
			      h5a_sid,H5P_DEFAULT);
      
      if(h5a_id <0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot create HDF5 attribute id for attrinue %s\n", sdsattr_name);
	  }
	H5Sclose(h5a_sid);
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
      
      if (sds_attdata) {
	sret = H5Awrite(h5a_id,h5_amemtype,(void *)sds_attdata);
	if(sret <0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot write HDF5 attribute\n");
	    }
	  H5Sclose(h5a_sid);
	  H5Aclose(h5a_id);
	  if (sds_attdata) free(sds_attdata);
	  H5Dclose(h5_dset);
	  return CONVERT_FAIL;
	}
      }
      sret = H5Sclose(h5a_sid);
      if(sret < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot close HDF5 space\n");
	  }
	H5Aclose(h5a_id);
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
      sret = H5Aclose(h5a_id);
      if(sret < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot close HDF5 attribute interface \n");
	  }
	if (sds_attdata) free(sds_attdata);
	H5Dclose(h5_dset);
	return CONVERT_FAIL;
      }
    }
    
    if(count_sdsattdata !=0)
      if (sds_attdata) free(sds_attdata);
  }   
  if(ret != CONVERT_SUCCESS)
    {
      ret = CONVERT_FAIL;
    }
  
  H5Dclose(h5_dset);
  HE5_EHset_error_on(2, 0);/* reset both HDFEOS5 and HDF5 error writing */
  return ret;
}



/****************************************************************************
FUNCTION:
        main()

DESCRIPTION:
	Main driver or start up function for conversion utility.

INPUTS:
        Name            Description                     Units   Min     Max

        argc		number of command line arguments
	argv		array of command line arguments

OUTPUTS:
        Name            Description                     Units   Min     Max

        NONE

FUNCTIONS_CALLED:

        GetInput		- Parse command line arguments
	DisplayHelpAndExit	- Display usage message
	GetNumberObjects	- Count number of specified type of 	
				HDF-EOS objects in file
	DoSwathConversion	- Convert swath objects from HDF4 to HDF5
	DoGridConversion	- Convert grid objects from HDF4 to HDF5
	DoPointConversion	- Convert point objects from HDF4 to HDF5

RETURNS:
        Program exit code

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int main (int argc, char *argv[])
{
    int numObjects;            /* number of HDF-EOS objects */
    int status;                /* return status */
    HdfInfo hdf4Info;          /* input file HDF-EOS object info */
    int num_EOS_Metadata;      /* number of EOS metadata in HDF4 file */

/***************************************************************************
  Parse command line - set up globals, if a problem show usage
  message to user and exit.
***************************************************************************/
    status = GetInput(argc,argv);
    if (status == CONVERT_FAIL)
    {
        DisplayHelpAndExit(argv[0]);
    }

/***************************************************************************
  Verbose message.
***************************************************************************/
    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("\n%s - copying HDF-EOS objects:\n",CONVERT_VERBOSE_HEAD);
        printf("    Input file:  %s\n",inNameGlobal);
        printf("    Output file:  %s\n",outNameGlobal);
    }

    /* initilize total number od SDSes that are hdfeos fields to 0 */

    hdfeos5Info.total_hdfeos_Used_SDS = 0;

/***************************************************************************
  Get number of Swath objects - do conversion if necessary.
***************************************************************************/
    numObjects = GetNumberObjects(&hdf4Info,CONVERT_SWATH);
    if (numObjects > 0)
    {
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("\n%s - Converting %d Swath objects\n",
                   CONVERT_VERBOSE_HEAD,numObjects);
        }
        status = DoSwathConversion(hdf4Info);
    }
    else if (numObjects == CONVERT_FAIL)
    {
    }
    else
    {
        puts("Message:  No Swath objects in input file.\n");
    }

/***************************************************************************
  Get number of Grid objects - do conversion if necessary.
***************************************************************************/
    numObjects = GetNumberObjects(&hdf4Info,CONVERT_GRID);
    if (numObjects > 0)
    {
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("\n%s - Converting %d Grid objects\n",
                   CONVERT_VERBOSE_HEAD,numObjects);
        }
        status = DoGridConversion(hdf4Info);
    }
    else if (numObjects == CONVERT_FAIL)
    {
    }
    else
    {
        puts("Message:  No Grid objects in input file.\n");
    }

/***************************************************************************
  Get number of Point objects - do nothing.....yet (this will change)
***************************************************************************/
    numObjects = GetNumberObjects(&hdf4Info,CONVERT_POINT);
    if (numObjects > 0)
    {
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("\n%s - Converting %d Point objects\n",
                   CONVERT_VERBOSE_HEAD,numObjects);
        }
        status = DoPointConversion(hdf4Info);
    }
    else if (numObjects == CONVERT_FAIL)
    {
    }
    else
    {
        puts("Message:  No Point objects in input file.\n");
    }

/***************************************************************************
  Get EOS Metadata (CoreMetadata, ArchiveMetadata)- do conversion if necessary.
  Also convert all other global attributes besides CoreMetadata and ArchiveMetadata
***************************************************************************/
    status = ConvertEOSGlobalMetadata(inNameGlobal, outNameGlobal);
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("Done - Global EOS metadata - \n");
      }

 /***************************************************************************
  Get remaining SDS and vdata (witten by HDF4)- do conversion if necessary.
  Also convert their attributes. These SDS's will go into HDFEOS group in 
  output hdf-eos5 file. Also we will handle remaining dimension scales here.
***************************************************************************/
    status = ConvertLoneSDSes(inNameGlobal, outNameGlobal);
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("Done - HDF4 written SDSes, their attributes and dimension scales - \n");
      }

    status = ConvertLoneVdatas(inNameGlobal, outNameGlobal);
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("Done - HDF4 written Vdatas, and their attributes- \n");
      }

/***************************************************************************
  All's well that ends well.
***************************************************************************/
    return 0;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	GetInput()

DESCRIPTION:
	Parse command line inputs, set global flags based on inputs, 
	and load global variables.

INPUTS:
        Name            Description                     Units   Min     Max

        argc		number of command line arguments
	argv		array of command line arguments

OUTPUTS:
        Name            Description                     Units   Min     Max

        NONE

FUNCTIONS_CALLED:

	DisplayHelpAndExit	- Display usage message

RETURNS:
        CONVERT_SUCCESS		- everything is ready to go
	CONVERT_FAIL		- can not continue

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int GetInput(int argc, char *argv[])
{
    int retVal;                      /* return value */
    int loopCount;                   /* for loop counter */
    int inFileLoc;                   /* index of -i flag */
    int outFileLoc;                  /* index of -o flag */
    int helpFlag;                    /* flag if user needs usage help */
    int compValueLoc;                /* index of -c flag */
    FILE *inFile = (FILE *) NULL;    /* in file - file pointer */
    FILE *outFile = (FILE *) NULL;   /* out file - file pointer */

/***************************************************************************
  Initializations
***************************************************************************/
    inFileLoc = 0;
    outFileLoc = 0;
    compValueLoc = 0;
    verboseModeGlobal = CONVERT_FALSE;
    compressModeGlobal = CONVERT_FALSE;
    numTilesGlobal = CONVERT_NO_TILE;
    fileCreatedGlobal = CONVERT_FALSE;
    helpFlag = CONVERT_FALSE;
    retVal = CONVERT_SUCCESS;

/***************************************************************************
  If only the program name is entered then we need to help the user.
***************************************************************************/
    if (argc == 1)
    {
        helpFlag = CONVERT_TRUE;
    }

/***************************************************************************
  Loop over arguments after program name.
***************************************************************************/
    for (loopCount = 1; loopCount < argc; loopCount++)
    {

/***************************************************************************
  Found the -i flag, next argument should be the input file name.
***************************************************************************/
        if (strcmp(argv[loopCount],"-i") == 0)
        {
            inFileLoc = loopCount + 1;
        }

/***************************************************************************
  Found the -o flag, next argument should be the output file name.
***************************************************************************/
        if (strcmp(argv[loopCount],"-o") == 0)
        {
            outFileLoc = loopCount + 1;
        }

/***************************************************************************
  Found the -h flag, set helpFlag to true.
***************************************************************************/
        if (strcmp(argv[loopCount],"-h") == 0)
        {
            helpFlag = CONVERT_TRUE;
        }

/***************************************************************************
  Found the -v flag, set verbose flag to true.
***************************************************************************/
        if (strcmp(argv[loopCount],"-v") == 0)
        {
            verboseModeGlobal = CONVERT_TRUE;
        }

/***************************************************************************
  Found the -version flag, set version flag to true.
***************************************************************************/
        if (strcmp(argv[loopCount],"-version") == 0)
        {
            DisplayVersionAndExit(argv[0]);
        }

/***************************************************************************
  Found the -c flag, do compression.
***************************************************************************/
        if (strcmp(argv[loopCount],"-c") == 0)
        {
            compressModeGlobal = CONVERT_TRUE;
            compValueLoc = loopCount + 1;
        }
    }  /* end for */

/***************************************************************************
  See if the user entered "-i -o" with no file names....
***************************************************************************/
    if (abs(inFileLoc - outFileLoc) < 2)
    {
        helpFlag = CONVERT_TRUE;
    }

/***************************************************************************
  The user either wanted or needed help.  Do it now.
***************************************************************************/
    if (helpFlag == CONVERT_TRUE)
    {
        DisplayHelpAndExit(argv[0]);
    }

/***************************************************************************
  Check to see if input file exists, has read permission and is a
  valid HDF4 file.
***************************************************************************/
    if (inFileLoc > 0)
    {
        inFile = fopen(argv[inFileLoc],"r");
        if ((Hishdf(argv[inFileLoc]) == 0) || (inFile == (FILE *) NULL))
        {
            printf("ERROR:   input file (%s) must exist and be HDF4 file.\n",argv[inFileLoc]);
            retVal = CONVERT_FAIL;
        }
    }
    else
    {
        puts("ERROR:   input file needed!\n");
        retVal = CONVERT_FAIL;
    }

/***************************************************************************
  Check to see if output file does not exist.
***************************************************************************/
    if (outFileLoc > 0)
    {
        outFile = fopen(argv[outFileLoc],"r");
        if (outFile != (FILE *) NULL)
        {
            printf("ERROR:  out file (%s) exists, file name must be new file.\n",
                   argv[outFileLoc]);
            retVal = CONVERT_FAIL;
           retVal = CONVERT_SUCCESS;
        }
    }
    else
    {
        puts("ERROR:   output file needed!\n");
        retVal = CONVERT_FAIL;
    }

/***************************************************************************
  Close files (we opened them for checking purposes).
***************************************************************************/
    if (inFile != (FILE *) NULL)
    {
        fclose(inFile);
    }

    if (outFile != (FILE *) NULL)
    {
        fclose(outFile);
    }

/***************************************************************************
  If all is well, load global file names with what the user input.
***************************************************************************/
    if (retVal == CONVERT_SUCCESS)
    {
        strcpy(inNameGlobal,argv[inFileLoc]);
        strcpy(outNameGlobal,argv[outFileLoc]);
    }

    if ((compValueLoc > 0) && (compValueLoc < argc))
    {
        char temString[30];
        int len;
        int count;
        int testVal;

        strcpy(temString,argv[compValueLoc]);
        len = (int) strlen(temString); 

        for (count = 0; count < len; count++)
        {
            testVal = isdigit(temString[count]);
            if (testVal == 0)
            {
                break;
            }
        }

        if (testVal != 0)
        {
            numTilesGlobal = atoi(temString);
        }
    }

    return retVal;
}







/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	GetNumberObjects()

DESCRIPTION:
	Determine the number of specified HDF4 HDF-EOS objects
	are in the file.

INPUTS:
        Name            Description                     Units   Min     Max

	objType		type of object to look for (SWATH/POINT/GRID)

OUTPUTS:
        Name            Description                     Units   Min     Max

        infoPtr		information about found objects

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure

RETURNS:
	Number of specified HDF-EOS objects or CONVERT_FAIL on failure.

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int GetNumberObjects(HdfInfo *infoPtr,int objType)
{
    char *nameList=NULL;                /* list of object names */
    int numObjs = 0;               /* number of objects found */
    int32 strBufSize;              /* length of object name list */
    int32 objects;                 /* number of objects */

/***************************************************************************
  Determine how many swath objects there are and get the list of 
  swath names.
***************************************************************************/
    switch (objType)
    {
        case CONVERT_SWATH:
        {
            objects = SWinqswath(inNameGlobal,NULL,&strBufSize);
            if (objects > 0)
            {
                nameList = (char *) calloc(1,(int) strBufSize + 1);
                if (nameList == NULL)
                {
                    objects = CallocErrorMsg(__LINE__);
                }
                else
                {
                    objects = SWinqswath(inNameGlobal,nameList,&strBufSize);
                }
            }
        }
        break;

/***************************************************************************
  Determine how many grid objects there are and get the list of 
  grid names.
***************************************************************************/
        case CONVERT_GRID:
        {
            objects = GDinqgrid(inNameGlobal,NULL,&strBufSize);
            if (objects > 0)
            {
                nameList = (char *) calloc(1,(int) strBufSize + 1);
                if (nameList == NULL)
                {
                    objects = CallocErrorMsg(__LINE__);
                }
                else
                {
                    objects = GDinqgrid(inNameGlobal,nameList,&strBufSize);
                }
            }
        }
        break;

/***************************************************************************
  Determine how many point objects there are and get the list of 
  point names.
***************************************************************************/
        case CONVERT_POINT:
        {
            objects = PTinqpoint(inNameGlobal,NULL,&strBufSize);
            if (objects > 0)
            {
                nameList = (char *) calloc(1,(int) strBufSize + 1);
                if (nameList == NULL)
                {
                    objects = CallocErrorMsg(__LINE__);
                }
                else
                {
                    objects = PTinqpoint(inNameGlobal,nameList,&strBufSize);
                }
            }
        }
        break;

    }  /* end switch */  


/***************************************************************************
  Set up structure to contain an array of object names.
***************************************************************************/
    infoPtr->total = numObjs = (int) objects;
    if (numObjs > 0)
    {
        char **ptr = NULL;
        char objName[500];
        int loopCount;
        long count;
        size_t *slen = NULL;

        infoPtr->name = (char **) calloc(numObjs,sizeof(char *));
        if (infoPtr->name == NULL)
        {
	  free(nameList);
	  nameList = NULL;
	  return (CallocErrorMsg(__LINE__));
        }

/***************************************************************************
  See how many names are in the list.
***************************************************************************/
        count = HE5_EHparsestr(nameList,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
        if (ptr == NULL)
	  {
	    free(nameList);
	    nameList = NULL;
	    return (CallocErrorMsg(__LINE__));
	  }
        slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
	  {
	    free(ptr);
	    free(nameList);
	    nameList = NULL;
	    return (CallocErrorMsg(__LINE__));
	  }

        count = HE5_EHparsestr(nameList,',',ptr,slen);

/***************************************************************************
  Loop over the names and place them in the structure.
***************************************************************************/
        for (loopCount = 0; loopCount < infoPtr->total; loopCount++)
        {
            memmove(objName,ptr[loopCount],(int) slen[loopCount]);
            objName[slen[loopCount]] = '\0';

            infoPtr->name[loopCount] = (char *) calloc(1,strlen(objName) + 1);
            if (infoPtr->name[loopCount] == NULL)
            {
                numObjs = CallocErrorMsg(__LINE__);
                break;
            }
            strcpy(infoPtr->name[loopCount],objName);
        }
        free(ptr);
        free(slen);
    }

    free(nameList);
    nameList = NULL;
    return numObjs;
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertEOSGlobalMetadata()

DESCRIPTION:
	Convert EOS metadata  from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

       inFileName       input file name
       outFileName      output file name

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

        SDstart
        SDend
        H5Fopen
        H5Fclose

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Apr 11          Abe Taaheri             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertEOSGlobalMetadata(char *inFileName, char *outFileName)
{
    int retVal = CONVERT_SUCCESS;           /* return value */
    int32 sd_id;
    int outFid;

    sd_id = SDstart(inFileName, DFACC_RDWR);

    if (fileCreatedGlobal == CONVERT_FALSE)
    {
      outFid = (int) H5Fopen(outFileName,H5F_ACC_TRUNC,H5P_DEFAULT);
        fileCreatedGlobal = CONVERT_TRUE;
    }
    else
    {
      outFid = (int) H5Fopen(outFileName,H5F_ACC_RDWR,H5P_DEFAULT);
    }

    if (outFid == CONVERT_FAIL) 
    {
        SDend(sd_id);
        retVal = CONVERT_FAIL; 
        printf("ERROR: Unable to open output file %s\n",outFileName);
        return retVal;
    }

 /* transfer metadata and Global attributes from input HDF file to output file */
    retVal = GetEOSMetadata( sd_id, outFid, inFileName, outFileName);

    SDend(sd_id);
    H5Fclose(outFid);
    
    return retVal;
}

/*************************************************************************

MODULE: GetEOSMetadata

PURPOSE: 
    This module sends requests to GetAttr to read user-defined core,
    archive, and other EOS metadata and write it to output HDF-EOS5
    file.
 

RETURN VALUE:

Value    	Description
-----		-----------
CONVERT_SUCCESS	Returns CONVERT_SUCCESS if the metadata 
                reading/writinng was successful.
                Returns CONVERT_FAIL otherwise

HISTORY:
Version    Date     Programmer      Code     Reason
-------    ----     ----------      ----     ------
           04/11    Abe Taaheri              

NOTES:

**************************************************************************/

intn GetEOSMetadata
(
 int32 sd_id,	        /* the file id for input HDF file */ 
 int out_id,	        /* the file id for ouput file */
 char *inFileName, char *outFileName
 )
  
{
  intn i;                /* used to loop through attrs */
  intn j;
  intn status=CONVERT_SUCCESS;           /* receives result of processing
					    from Getattr function */     
  char attrname[256];    /* holds the file_name string */ 
  char *root = NULL;     /* will point to attr type */
  
  /* set up a loop to check for and read the three
     types of metadata -- core, archive, metadata*/
  
  for ( i = 1; i < 3; i++ )
    {
      /* set up root to hold correct string */ 
      switch( i )
        {
	case 1:
	  root = "CoreMetadata";
	  break;
	case 2:
	  root = "ArchiveMetadata";
	  break;
        }  
      
      /* First, try root name alone */
      status = ConvertEOSAttr( sd_id, out_id, root );
      if(status == ERROR_MEMORY || status == CONVERT_FAIL)
	{
	  return(CONVERT_FAIL);
	}
      /* Now, try concatenating sequence numbers */
      for ( j = 0; j <= 9; j++ )
        {
	  sprintf( attrname, "%s.%d", root, j );
	  status = ConvertEOSAttr( sd_id, out_id, attrname );
	  if(status == WARNING || status == ERROR_READ_INPUT) continue;
	  if(status == ERROR_MEMORY || status == CONVERT_FAIL)
	    {
	      return(CONVERT_FAIL);
	    }
        }
      
    }  /* end outer for loop */

  /* Convert all other Global Attributes besides CoreMetadata and ArchiveMetadata */
  /* skip HDFEOSVersion, which is not needed */

  status = ConvertOtherGlobalAttr(sd_id, out_id, inFileName, outFileName);

  if(status !=CONVERT_FAIL) status = CONVERT_SUCCESS;

  /* return the value of the function to main function */
  return ( status );
  
}  /* end GetEOSMetadata function */


/*************************************************************************

MODULE: ConvertEOSAttr

PURPOSE: 
    This module reads the user-defined core, and archive metadata
    from the original (input) HDF-EOS2 file and writes it to output
    hdfeos5 file.  

RETURN VALUE:
Type = intn
Value    	Description
-----		-----------
CONVERT_SUCCESS	Returns CONVERT_SUCCESS if the metadata 
                reading/writinng was successful.
                Returns CONVERT_FAIL otherwise

HISTORY:
Version    Date     Programmer      Code     Reason
-------    ----     ----------      ----     ------
           04/11    Abe Taaheri              

NOTES:

**************************************************************************/
intn ConvertEOSAttr
(
  int32 fid_in   ,	/* the SD id for input HDF4 file */ 
  int fid_out,	        /* the file id for ouput he5 file */
  char *attr		/* the filename handle of the attribute to move */ 
 )
  
{
  intn status = CONVERT_SUCCESS; /* this is the var that holds return val */
  int32 my_attr_index;           /* holds return val from SDfindattr */ 
  int32 data_type;	         /* holds attribute's data type */ 
  int32 n_values;                /* stores # of vals of the attribute */
  char *file_data = NULL;                 /* char ptr used to allocate temp space
					   * during transfer of attribute info */
  int                retVal = CONVERT_SUCCESS;	/* SDPS toolkit ret value */
  herr_t             HDF5status = CONVERT_SUCCESS;;
  hid_t              datid;
  hid_t              ggid;
  hid_t              atype;
  hid_t              HDFfid;
  size_t             size;
  hid_t              sid2;
  int		     attrlen=0, numBytesRead = 0;
  char              *Buffer_temp = NULL;
  static int         OnetimeFlage = 0;
  char *attr_name;
  char *new_attr_name;
  char *dynamicMsg;

  attr_name = (char *)malloc(MAX_NC_NAME * sizeof(char));
  new_attr_name  = (char *)malloc((MAX_NC_NAME +3)* sizeof(char));
  dynamicMsg = (char *)malloc(64* sizeof(char));

  if (verboseModeGlobal == CONVERT_TRUE && OnetimeFlage == 0)
    {
      printf("\n Searching for Core and Archive Metadata.\n");
      OnetimeFlage = 1;
    }

  /* look for attribute in the HDF file */
  my_attr_index = SDfindattr( fid_in, attr );
  
  /* only proceed if the attribute was found */
  if ( my_attr_index != -1 )
    {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("Found attr_index for attribute %s. Writing it to He5 file.\n", attr);
	}
    }
  else
    {
      free(attr_name);
      free(new_attr_name);
      free(dynamicMsg);
      return( WARNING);
    }
  
  /* get size of HDF file attribute  */
  status = SDattrinfo( fid_in, my_attr_index, attr_name, &data_type, &n_values );
  
  /* only proceed if successful read of attribute info */
  if ( status == -1 )
    {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("Cannot read  info for attribute %s.\n", attr);
	}
      free(attr_name);
      free(new_attr_name);
      free(dynamicMsg);
      return( ERROR_READ_INPUT);
    } 
  
  /* attempt to allocate memory for HDF file attribute contents */
  file_data = ( char * ) calloc( (n_values+1), sizeof(char) );
  
  if ( file_data == NULL )
    {
      printf("Error: Unable to allocate %d bytes for %s\n",
	       n_values, attr );
      free(attr_name);
      free(new_attr_name);
      free(dynamicMsg);
      return ( ERROR_MEMORY );
    }
      
  /* read attribute from the HDF4 file */
  status = SDreadattr( fid_in, my_attr_index, file_data );
  if(status == -1 )
    {
      /* first free the allocated memory */
      free( file_data );
      printf("cannot read data for attribute %s.\n", attr);
      free(attr_name);
      free(new_attr_name);
      free(dynamicMsg);
      return( ERROR_READ_INPUT);
    }
  
  /* Write metadata to output HDF5 file */
 
  HDFfid =(hid_t) fid_out;

  /* probe: open group "HDFEOS INFORMATION" */
  status = H5Eset_auto(NULL, NULL);
  ggid = H5Gopen(HDFfid,"HDFEOS INFORMATION");

  /* if it doesn't exist, create it */
  if( ggid < 0 ) 
    {
      ggid = H5Gcreate(HDFfid, "HDFEOS INFORMATION", 0);
      if (ggid == -1)
	{
	  printf("Cannot create \"HDFEOS INFORMATION\" group in output HDF-EOS5 file.\n");
	  free(file_data);
	  file_data = NULL;
	  free(attr_name);
	  free(new_attr_name);
	  free(dynamicMsg);
	  return(CONVERT_FAIL);
	}
    }

  /* 
     -------------------------------------------------------
     |Create dataset  "coremetadata, coremetadata.0, etc   |
     |and attach to the "HDFEOS INFORMATION"  group        | 
     -------------------------------------------------------
  */      
  
  sid2    = H5Screate(H5S_SCALAR);
  if (sid2 == -1)
    {
      printf("Cannot create dataspace for %s dataset.\n",attr);
      H5Gclose(ggid);
      HDF5status = CONVERT_FAIL;
    }
  
  if(HDF5status == CONVERT_SUCCESS)
    {
      atype   = H5Tcopy(H5T_C_S1);
      size = (size_t)MAX_ORDER;
      status  = H5Tset_size(atype,size);
      if (status == -1)
	{
	  printf("Cannot set the total size for atomic datatype.\n");
	  H5Gclose(ggid);
	  H5Sclose(sid2);
	  HDF5status = CONVERT_FAIL;
	}
    }

  attrlen = strlen(file_data);
  numBytesRead = attrlen * sizeof(char);

  if(HDF5status == CONVERT_SUCCESS)
    {
      file_data[ numBytesRead ] = '\0';
      datid = H5Dopen(ggid, attr);
      if(datid == -1)
	{
	  datid   = H5Dcreate(ggid,attr,atype,sid2,H5P_DEFAULT);
	  if (datid == -1)
	    {
	      printf("Cannot create %s dataset.",attr);
	      H5Gclose(ggid);
	      H5Sclose(sid2);
	      HDF5status = CONVERT_FAIL;
	    }
	}
    }
  
  /* 
     -----------------------------------------------------
     |    Write data to the dataset attr                 | 
     -----------------------------------------------------
  */ 
  if(HDF5status == CONVERT_SUCCESS)
    {
       Buffer_temp = (char *) calloc((int)(MAX_ORDER + 1), sizeof(char));
      if( Buffer_temp ==  NULL)
	{
	  /* error message is:
	     "Unable to allocate memory for the hdf attribute" */
	  free(file_data);
	  file_data = NULL;
	  free(attr_name);
	  free(new_attr_name);
	  free(dynamicMsg);
	  retVal = CallocErrorMsg(__LINE__);
	  H5Dclose(datid);
	  H5Gclose(ggid);
	  H5Sclose(sid2);
	  retVal = CONVERT_FAIL;
	  return(retVal);
	}
      strcpy(Buffer_temp, file_data);
      status = H5Dwrite(datid,atype,H5S_ALL,H5S_ALL,H5P_DEFAULT,Buffer_temp);
      free(Buffer_temp);
      Buffer_temp = (char *) NULL;
      if (status == -1)
	{
	  printf("Cannot write in data to the %s dataset.\n",attr); 
	  H5Dclose(datid);
	  H5Gclose(ggid);
	  H5Sclose(sid2);
	  HDF5status = CONVERT_FAIL;
	}
    }
  
  
  /* 
     ----------------------------------------------------
     |     Release (close) the created objects          |
     ----------------------------------------------------
  */

  if(HDF5status == CONVERT_SUCCESS)
    {
      status = H5Sclose(sid2);
      if (status == -1)
	{
	  printf("Cannot release the dataspace ID.\n");
	  H5Dclose(datid);
	  H5Gclose(ggid);
	  HDF5status = CONVERT_FAIL;
	}
    }
  if(HDF5status == CONVERT_SUCCESS)
    {
      status = H5Dclose(datid);
      if (status == -1)
	{
	  printf("Cannot release the dataset ID.\n");
	  H5Gclose(ggid);
	  HDF5status = CONVERT_FAIL;
	}
    }
  
  if(HDF5status != CONVERT_SUCCESS)
    {
      /* error message is:
	 "Unable to set the HDF file attribute" */
      free( file_data );
      free(Buffer_temp);
      Buffer_temp = (char *) NULL;
      return(HDF5status);
    }

  /* free dynamically allocated memory */
  free( file_data );
  free(Buffer_temp);

  free( attr_name);
  free( new_attr_name);
  free( dynamicMsg);
  /* return the value of status to ConvertEOSAttr */
  return (HDF5status );
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	DoSwathConversion()

DESCRIPTION:
	Convert Swaths from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

	swathInfo	information about the swath

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure
	ConvertSWorGDDims	- convert swath or grid dimensions
	ConvertSwathDimMap	- convert swath geolocation mappings
	ConvertSwathIndexMap	- convert swath indexed mappings
	ConvertSWorGDFieldInfo	- convert swath or grid field data 
				(geolocation or data field)
	ConvertHDFEOSAttribute	- convert swath, point, or grid attribute 
				information

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int DoSwathConversion(HdfInfo swathInfo)
{
    int retVal = CONVERT_SUCCESS;           /* return value */
    int loopCount;                          /* for loop counter */
    int32 inSWFid;                          /* input file HDF4 file id */
    int outSWFid;                           /* output file HDF5 file id */

/***************************************************************************
  Open input file for Swath access.
***************************************************************************/
    inSWFid = SWopen(inNameGlobal,DFACC_READ);
    if (inSWFid == CONVERT_FAIL) 
    {
        retVal = CONVERT_FAIL; 
        printf("ERROR: Unable to open swath %s\n",inNameGlobal);
        return retVal;
    }

/***************************************************************************
  Open output file for Swath creating.
***************************************************************************/
    if (fileCreatedGlobal == CONVERT_FALSE)
    {
        outSWFid = HE5_SWopen(outNameGlobal,H5F_ACC_TRUNC);
        fileCreatedGlobal = CONVERT_TRUE;
    }
    else
    {
        outSWFid = HE5_SWopen(outNameGlobal,H5F_ACC_RDWR);
    }

    if (outSWFid == CONVERT_FAIL) 
    {
        SWclose(inSWFid);
        retVal = CONVERT_FAIL; 
        printf("ERROR: Unable to open swath %s\n",outNameGlobal);
        return retVal;
    }

/***************************************************************************
  Loop over total number of swaths.
***************************************************************************/
    for (loopCount = 0; loopCount < swathInfo.total; loopCount++)
    {
        char *swathName;              /* name of swath */
        int status;                   /* return status */
        int32 inSWid;                 /* HDF4 swath id (input) */
        int outSWid;                  /* HDF5 swath id (output) */

/***************************************************************************
  Load name of swath into local variable.
***************************************************************************/
        swathName = (char *) calloc(1,strlen(swathInfo.name[loopCount]) + 1);
        if (swathName == NULL)
        {
            retVal = CallocErrorMsg(__LINE__);
            break;
        }
        strcpy(swathName,swathInfo.name[loopCount]);

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Begin - convert Swath - %s\n",swathName);
        }

/***************************************************************************
  Attach to input file swath.
***************************************************************************/
        inSWid = SWattach(inSWFid,swathName);
        if (inSWid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            free(swathName);
            puts("ERROR:  SWattach fail.\n");
            break;
        }

/***************************************************************************
  Create a swath in the output file.
***************************************************************************/
        outSWid = HE5_SWcreate(outSWFid,swathName);
        if (outSWid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            free(swathName);
            puts("ERROR:  HE5_SWcreate fail.\n");
            break;
        }

/***************************************************************************
  Now, convert the swath.
***************************************************************************/
        status = ConvertSWorGDDims(inSWid,outSWid,CONVERT_SWATH_DIMS);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Swath Dimensions - %s\n",swathName);
        }

        status = ConvertSwathDimMap(inSWid,outSWid);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Swath Dimension Mappings - %s\n",swathName);
        }

        status = ConvertSwathIndexMap(inSWid,outSWid);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Swath Indexed Mappings - %s\n",swathName);
        }

        status = ConvertSWorGDFieldInfo(inSWid,outSWid,outSWFid,swathName,CONVERT_SWATH_GEOFIELD);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Swath Geolocation Fields - %s\n",swathName);
        }

        status = ConvertSWorGDFieldInfo(inSWid,outSWid,outSWFid,swathName,CONVERT_SWATH_DATAFIELD);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Swath Data Fields - %s\n",swathName);
        }

        status = ConvertHDFEOSAttribute(inSWid,outSWid,CONVERT_SWATH_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Swath Attributes - %s\n",swathName);
         }

	/* Convert dimension scales */
	status = ConvertDimScales(inSWid,outSWid,CONVERT_SWATH_DIMSCALE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Swath Dimension Scales - %s\n",swathName);
        }
	/* Convert Field attributes written by HDF calls (Note: HDFEOS2 does not have
	   APIs to write/read field attributes */
       status = ConvertFieldAttribute(inSWid,outSWid,CONVERT_SWATH_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Swath Field Attributes - %s\n",swathName);
	    printf("End - convert Swath - %s\n",swathName);
        }

/***************************************************************************
  Free the name and detach.
***************************************************************************/
        free(swathName);
        SWdetach(inSWid);
        HE5_SWdetach(outSWid);
    }

/***************************************************************************
  Close the file - done swath conversion
***************************************************************************/
    SWclose(inSWFid);
    HE5_SWclose(outSWFid);
    
    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	DoGridConversion()

DESCRIPTION:
	Convert Grids from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

	gridInfo	information about the grid

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure
	ConvertGridProjInfo	- convert grid projection information
	ConvertSWorGDDims	- convert swath or grid dimensions
	ConvertSWorGDFieldInfo	- convert swath or grid field data 
				(geolocation or data field)
	ConvertHDFEOSAttribute	- convert swath, point, or grid attribute 
				information

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int DoGridConversion(HdfInfo gridInfo)
{
    int retVal = CONVERT_SUCCESS;           /* return value */
    int loopCount;                          /* for loop counter */
    int32 inGDFid;                          /* input file HDF4 file id */
    int outGDFid;                           /* output file HDF5 file id */

/***************************************************************************
  Open input file for Grid access.
***************************************************************************/
    inGDFid = GDopen(inNameGlobal,DFACC_READ);
    if (inGDFid == CONVERT_FAIL) 
    {
        printf("ERROR:  Unable to open Grid %s\n",inNameGlobal);
        return CONVERT_FAIL;
    }

/***************************************************************************
  Open output file for Grid creating.
***************************************************************************/
    if (fileCreatedGlobal == CONVERT_FALSE)
    {
        outGDFid = HE5_GDopen(outNameGlobal,H5F_ACC_TRUNC);
        fileCreatedGlobal = CONVERT_TRUE;
    }
    else
    {
        outGDFid = HE5_GDopen(outNameGlobal,H5F_ACC_RDWR);
    }

    if (outGDFid == CONVERT_FAIL) 
    {
        GDclose(inGDFid);
        printf("ERROR:  Unable to open Grid %s\n",outNameGlobal);
        return CONVERT_FAIL;
    }

/***************************************************************************
  Loop over total number of grids.
***************************************************************************/
    for (loopCount = 0; loopCount < gridInfo.total; loopCount++)
    {
        char *gridName;                   /* grid name */
        double outUpLeft[2];              /* upper left point (HDF5) */
        double outLowRight[2];            /* lower right point (HDF5) */
        int status;                       /* return status */
        int inLoop;                       /* for inner for loop counter */
        int32 inGDid;                     /* input grid id (HDF4) */
        int32 inXDim;                     /* input X dimension (HDF4) */
        int32 inYDim;                     /* input Y dimension (HDF4) */
        float64 inUpLeft[2];              /* upper left point (HDF4) */
        float64 inLowRight[2];            /* upper left point (HDF4) */
        long outXDim;                     /* output X dimension (HDF5) */
        long outYDim;                     /* output X dimension (HDF5) */
        int outGDid;                      /* output grid id (HDF5) */

/***************************************************************************
  Load name of grid into local variable.
***************************************************************************/
        gridName = (char *) calloc(1,strlen(*(gridInfo.name+loopCount)) + 1);
        if (gridName == NULL)
        {
            return (CallocErrorMsg(__LINE__));
        }
        strcpy(gridName,*(gridInfo.name+loopCount));

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Begin - convert Grid - %s\n",gridName);
        }

/***************************************************************************
  Attach to input file grid.
***************************************************************************/
        inGDid = GDattach(inGDFid,gridName);
        if (inGDid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            puts("ERROR:  HDF4 GDattach failure.\n");
            free(gridName);
            break;
        }

/***************************************************************************
  Get grid information.
***************************************************************************/
        status = GDgridinfo(inGDid,&inXDim,&inYDim,inUpLeft,inLowRight);
        if (inGDid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            puts("ERROR:  HDF4 GDgridinfo failure.\n");
            free(gridName);
            break;
        }

/***************************************************************************
  Convert grid information to HDF5 input data types (which are
  native data types).
***************************************************************************/
        outXDim = (long) inXDim;
        outYDim = (long) inYDim;
        for (inLoop = 0; inLoop < 2; inLoop++)
        {
            outUpLeft[inLoop] = (double) inUpLeft[inLoop];
            outLowRight[inLoop] = (double) inLowRight[inLoop];
        }

/***************************************************************************
  Create a grid in the output file.
***************************************************************************/
        outGDid = HE5_GDcreate(outGDFid,gridName,outXDim,outYDim,outUpLeft,
                               outLowRight);
        if (outGDid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            puts("ERROR:  HE5_GDcreate failure.\n");
            free(gridName);
            break;
        }
        else
        {
            if (verboseModeGlobal == CONVERT_TRUE)
            {
                 printf("Done - Grid Information - %s\n",gridName);
            }
        }

/***************************************************************************
  Now, convert the grid.
***************************************************************************/
        status = ConvertGridProjInfo(inGDid,outGDid);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Grid Projection Information - %s\n",gridName);
        }

        status = ConvertSWorGDDims(inGDid,outGDid,CONVERT_GRID_DIMS);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Grid Dimensions - %s\n",gridName);
        }

        status = ConvertSWorGDFieldInfo(inGDid,outGDid,outGDFid,gridName,CONVERT_GRID_DATAFIELD);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Grid Data Fields - %s\n",gridName);
        }

        status = ConvertHDFEOSAttribute(inGDid,outGDid,CONVERT_GRID_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Grid Attributes - %s\n",gridName);
         }
	/* Convert dimension scales */
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Start - Grid Dimension Scales - %s\n",gridName);
        }
	status = ConvertDimScales(inGDid,outGDid,CONVERT_GRID_DIMSCALE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Grid Dimension Scales - %s\n",gridName);
        }
	/* Convert Field attributes written by HDF calls (Note: HDFEOS2 does not have
	   APIs to write/read field attributes */
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Start - Grid Field Attributes - %s\n",gridName);
	}
       status = ConvertFieldAttribute(inGDid,outGDid,CONVERT_GRID_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Grid Field Attributes - %s\n",gridName);
            printf("End - convert Grid - %s\n",gridName);
         }

/***************************************************************************
  Free the name and detach.
***************************************************************************/
        free(gridName);
        GDdetach(inGDid);
        HE5_GDdetach(outGDid);
    }

/***************************************************************************
  Close the file - done grid conversion
***************************************************************************/
    GDclose(inGDFid);
    HE5_GDclose(outGDFid);
    
    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	DoPointConversion()

DESCRIPTION:
	Convert Point from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

	pointInfo	information about the grid

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:


RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Dec 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int DoPointConversion(HdfInfo pointInfo)
{
    int retVal = CONVERT_SUCCESS;           /* return value */
    int loopCount;                          /* for loop counter */
    int32 inPTFid;                          /* input file HDF4 file id */
    int outPTFid;                           /* output file HDF5 file id */

/***************************************************************************
  Open input file for Point access.
***************************************************************************/
    inPTFid = PTopen(inNameGlobal,DFACC_READ);
    if (inPTFid == CONVERT_FAIL) 
    {
        printf("ERROR:  Unable to open Point %s\n",inNameGlobal);
        return CONVERT_FAIL;
    }

/***************************************************************************
  Open output file for Point creating.
***************************************************************************/
    if (fileCreatedGlobal == CONVERT_FALSE)
    {
        outPTFid = HE5_PTopen(outNameGlobal,H5F_ACC_TRUNC);
        fileCreatedGlobal = CONVERT_TRUE;
    }
    else
    {
        outPTFid = HE5_PTopen(outNameGlobal,H5F_ACC_RDWR);
    }

    if (outPTFid == CONVERT_FAIL) 
    {
        PTclose(inPTFid);
        printf("ERROR:  Unable to open Point %s\n",outNameGlobal);
        return CONVERT_FAIL;
    }

/***************************************************************************
  Loop over total number of points.
***************************************************************************/
    for (loopCount = 0; loopCount < pointInfo.total; loopCount++)
    {
        char *pointName;
        int outPTid;
        int status;                       /* return status */
        int32 inPTid;

/***************************************************************************
  Load name of point into local variable.
***************************************************************************/
        pointName = (char *) calloc(1,strlen(*(pointInfo.name+loopCount)) + 1);
        if (pointName == NULL)
        {
            return (CallocErrorMsg(__LINE__));
        }
        strcpy(pointName,*(pointInfo.name+loopCount));

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Begin - convert Point - %s\n",pointName);
        }

/***************************************************************************
  Attach to input file point.
***************************************************************************/
        inPTid = PTattach(inPTFid,pointName);
        if (inPTid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            puts("ERROR:  HDF4 PTattach failure.\n");
            free(pointName);
            break;
        }

/***************************************************************************
  Create a point in the output file.
***************************************************************************/
        outPTid = HE5_PTcreate(outPTFid,pointName);
        if (outPTid == CONVERT_FAIL)
        {
            retVal = CONVERT_FAIL;
            free(pointName);
            puts("ERROR:  HE5_PTcreate fail.\n");
            break;
        }

/***************************************************************************
  Now, convert the point.
***************************************************************************/
        status = ConvertPointLevelData(inPTid,outPTid);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
             printf("Done - Point Level Data - %s\n",pointName);
        }

        status = ConvertHDFEOSAttribute(inPTid,outPTid,CONVERT_POINT_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Point Attributes - %s\n",pointName);
         }

	/* Convert Field attributes written by HDF calls (Note: HDFEOS2 does not have
	   APIs to write/read field attributes */
       status = ConvertFieldAttribute(inPTid,outPTid,CONVERT_POINT_ATTRIBUTE);
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Done - Point Field Attributes - %s\n",pointName);
            printf("End - convert Point - %s\n",pointName);
        }

/***************************************************************************
  Free the name and detach.
***************************************************************************/
        free(pointName);
        PTdetach(inPTid);
        HE5_PTdetach(outPTid);
    }

    return retVal;
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertSWorGDDims()

DESCRIPTION:
	Convert Swath or Grid dimensions from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file object id (HDF4 file)
	outObj		output file object id (HDF5 file)
        typeFlag	Swath dimensions or Grid dimensions flag

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertSWorGDDims(int32 inObj,int outObj,int typeFlag)
{
    char typeString[10];
    int retVal = CONVERT_SUCCESS;       /* return value */
    int32 numDims;                      /* number of dimensions */
    int32 bufSize;                      /* string buffer size */

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        if (typeFlag == CONVERT_SWATH_DIMS)
        {
            strcpy(typeString,"Swath");
        }
        else 
        {
            strcpy(typeString,"Grid");
        }
    }

/***************************************************************************
  Get number of dimensions grid or swath.
***************************************************************************/
    if (typeFlag == CONVERT_SWATH_DIMS)
    {
        numDims = SWnentries(inObj,HDFE_NENTDIM,&bufSize);
    }
    else
    {
        numDims = GDnentries(inObj,HDFE_NENTDIM,&bufSize);
    }

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("Working - Number of %s dimensions = %d\n",typeString,
              (int) numDims);
    }

/***************************************************************************
  If there are dimensions.....convert.
***************************************************************************/
    if (numDims > 0)
    {
        char *dimName;              /* name of dimension */
        char **ptr = NULL;          /* ptr to dimension name(s) */
        int loop;                   /* for loop counter */
        int status;                 /* return status */
        int32 *dims;                /* array of dimensions */
        long count;                 /* number of dimensions */
        size_t *slen = NULL;        /* array of name lengths */
	char *oneDimName, *unlim_oneDimName;

	oneDimName = (char *)malloc(500 * sizeof(char));
	unlim_oneDimName = (char *)malloc(500 * sizeof(char));
/***************************************************************************
  Allocate space to inquire dimensions.
***************************************************************************/
        dims = (int32 *) calloc(numDims,(int) sizeof(int32));
        dimName = (char *) calloc(1,(int) (bufSize+1));
        if ((dims == NULL) || (dimName == NULL))
        {
            return (CallocErrorMsg(__LINE__));
        }

        if (typeFlag == CONVERT_SWATH_DIMS)
        {
            numDims = SWinqdims(inObj,dimName,dims);
        }
        else
        {
            numDims = GDinqdims(inObj,dimName,dims);
        }
       
/***************************************************************************
  Double check - make sure there are dimensions.
***************************************************************************/
        if (numDims > 0)
        {

/***************************************************************************
  Get information about the dimension names.
***************************************************************************/
            count = HE5_EHparsestr(dimName,',',NULL,NULL);
            ptr = (char **) calloc(count,sizeof(char *));
            if (ptr == NULL)
            {
                return (CallocErrorMsg(__LINE__));
            }
            slen = (size_t *) calloc(count,sizeof(long));
            if (slen == NULL)
            {
	      free(ptr);
	      return (CallocErrorMsg(__LINE__));
            }
            count = HE5_EHparsestr(dimName,',',ptr,slen);
   
/***************************************************************************
  Loop over number of dimensions.
***************************************************************************/
            for (loop = 0; loop < count; loop++)
            {
                hsize_t dimSize;         /* dimension size */
		int32 oneDimName_size;

/***************************************************************************
  Get dimension name in local variable.
***************************************************************************/
                memmove(oneDimName,ptr[loop],(int) slen[loop]);
                oneDimName[slen[loop]] = '\0';
		oneDimName_size = SWdiminfo(inObj, oneDimName);

/***************************************************************************
  Set up dimSize to HDF5 type.
***************************************************************************/
                if (dims[loop] == 0)
                {
                    dimSize = (hsize_t) H5S_UNLIMITED;
                }
                else
                {
                    dimSize = (hsize_t) dims[loop];
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Defining %s Dimension %s [%d]\n",
			   typeString,oneDimName, (int32)dimSize);
                }

/***************************************************************************
  Write dimensions to output file.
***************************************************************************/
                if (typeFlag == CONVERT_SWATH_DIMS)
                {
		  if(dimSize == (hsize_t) H5S_UNLIMITED)
		    {
		      
		      strcpy(unlim_oneDimName,"Unlim_");
		      strcat(unlim_oneDimName,oneDimName);
		      strcat(unlim_oneDimName,"_Unlim");
		      
		      status = HE5_SWdefdim(outObj,unlim_oneDimName, H5S_UNLIMITED);
		    }
		  else
		    {
		      status = HE5_SWdefdim(outObj,oneDimName,dimSize);
		    }
                }
                else
                {
                    status = HE5_GDdefdim(outObj,oneDimName,dimSize);
                }

                if (status == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    break;
                }
            }
            free(ptr);
            free(slen);
        }
        free(dims);
        free(dimName);
	free(oneDimName);
	free(unlim_oneDimName);
    }
    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertSwathDimMap()

DESCRIPTION:
	Convert Swath dimension mappings from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file object id (HDF4 file)
	outObj		output file object id (HDF5 file)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertSwathDimMap(int32 inObj,int outObj)
{
    char *dimMap;                   /* dimension mappings */
    int retVal = CONVERT_SUCCESS;   /* return value */
    int32 numMap;                   /* number of dimension mappings */
    int32 bufSize;                  /* string buffer size */
    int32 *offSet;                  /* mapping offsets */
    int32 *increment;               /* mapping increments */

/***************************************************************************
  Get number of mapping entries.
***************************************************************************/
    numMap = SWnentries(inObj,HDFE_NENTMAP,&bufSize);

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("Working - Number of Swath dimension mappings = %d\n",
                  (int) numMap);
    }

/***************************************************************************
  Allocate information to inquire about the geolocation mappings.
***************************************************************************/
    offSet = (int32 *) calloc(numMap,sizeof(int32));
    increment = (int32 *) calloc(numMap,sizeof(int32));
    dimMap = (char *) calloc(1,(int) (bufSize+1));
    if ((offSet == NULL) || (increment == NULL) || (dimMap == NULL))
    {
        return (CallocErrorMsg(__LINE__));
    } 

/***************************************************************************
  Inquire geolocation mappings.
***************************************************************************/
    numMap = SWinqmaps(inObj,dimMap,offSet,increment);

    if (numMap > 0)
    {
        char **ptr;            /* array of pointers for mappings */
        int loop;              /* for looop counter */
        long count;            /* number of dimension mappings */
        size_t *slen;          /* array of lengths */
	char *temBuf;

	temBuf = (char *) malloc(500 * sizeof(char *));

/***************************************************************************
  Setup pointers to allow parsing of dimension names.
***************************************************************************/
        count = HE5_EHparsestr(dimMap,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
	if (ptr == NULL)
        {
            return (CallocErrorMsg(__LINE__));
        }
       slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
        {
	  free(ptr);
	  return (CallocErrorMsg(__LINE__));
        }
        count = HE5_EHparsestr(dimMap,',',ptr,slen);

/***************************************************************************
  Loop over number of dimension mappings.
***************************************************************************/
        for (loop = 0; loop < count; loop++)
        {
            char *ptrN[2];          /* array of pointers for mappings */
            size_t slenN[2];        /* string lengths */
            long countN;            /* number of mappings */
            hsize_t h5Offset;       /* HDF5 offset */
            hsize_t h5Increment;    /* HDF5 increment */
            int status;             /* HDF5 offset */
	    char *geoDim, *dataDim;

	    geoDim = (char *)malloc(500 * sizeof(char *));
	    dataDim = (char *)malloc(500 * sizeof(char *));

/***************************************************************************
  Copy mapping into temp buffer.
***************************************************************************/
            memmove(temBuf,ptr[loop],(int) slen[loop]);
            temBuf[slen[loop]] = '\0';

/***************************************************************************
  Parse geo and data dim info.
***************************************************************************/
            countN = HE5_EHparsestr(temBuf,'/',ptrN,slenN);
            if (countN != 2)
            {
                printf("ERROR:  Parsing data/geo dimensions - %s\n",temBuf);
                retVal = CONVERT_FAIL;
                break;
            }
            memmove(geoDim,ptrN[0],(int) slenN[0]);
            geoDim[slenN[0]] = '\0';
            memmove(dataDim,ptrN[1],(int) slenN[1]);
            dataDim[slenN[1]] = '\0';

/***************************************************************************
  Place offset and increment into HDF5 datatype variables.
***************************************************************************/
            h5Offset = (hsize_t) offSet[loop];
            h5Increment = (hsize_t) increment[loop];

            if (verboseModeGlobal == CONVERT_TRUE)
            {
                printf("Working - Defining Swath dimension mapping\n Geolocation Dim:  %s\nData Dimension:  %s\n",geoDim,dataDim);
            }

/***************************************************************************
  Define dimension mappings in output file.
***************************************************************************/
            status = HE5_SWdefdimmap(outObj,geoDim,dataDim,h5Offset,h5Increment);
            if (status == CONVERT_FAIL)
            {
                puts("ERROR:  HE5_SWdefdimmap failure.\n");
                break;
            }

	    free(geoDim);
	    free(dataDim);
        }
        free(ptr);
        free(slen);
	free(temBuf);
    }
    free(offSet);
    free(increment);
    free(dimMap);

    return retVal;
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertSwathIndexMap()

DESCRIPTION:
	Convert Swath indexed mappings from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file object id (HDF4 file)
	outObj		output file object id (HDF5 file)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertSwathIndexMap(int32 inObj,int outObj)
{
    int retVal = CONVERT_SUCCESS;    /* return value */
    int32 numMap;                    /* number of indexed mappings */
    int32 bufSize;                   /* buffer size */

/***************************************************************************
  Get number of indexed mapping entries.
***************************************************************************/
    numMap = SWnentries(inObj,HDFE_NENTIMAP,&bufSize);

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("Working - Number of Swath indexed mappings = %d\n",
                  (int) numMap);
    }

/***************************************************************************
  If there are any......convert.
***************************************************************************/
    if (numMap > 0)
    {
        char *dimMap;           /* dimension mapping */
        char **ptr;             /* array of mapping pointers */
        int loop;               /* for loop counter */
        long count;             /* number of mappings */
        size_t *slen;           /* array of lengths */
	char *indexMap;

	indexMap = (char *) malloc(500 * sizeof(char *));

/***************************************************************************
  Create space for and get the indexed mapping names.
***************************************************************************/
        dimMap = (char *) calloc(1,(int) (bufSize + 1));
        if (dimMap == NULL)
        {
            return (CallocErrorMsg(__LINE__));
        }
        numMap = SWinqidxmaps(inObj,dimMap,NULL);

/***************************************************************************
  Parse information about indexed mapping names.
***************************************************************************/
        count = HE5_EHparsestr(dimMap,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
        if (ptr == NULL)
        {
            return (CallocErrorMsg(__LINE__));
        }
        slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
        {
	  free(ptr);
	  return (CallocErrorMsg(__LINE__));
        }
        count = HE5_EHparsestr(dimMap,',',ptr,slen);

/***************************************************************************
  Loop over number of indexed mappings.
***************************************************************************/
        for (loop = 0; loop < count; loop++)
        {
            char *ptrN[2];          /* array of mapping pointers */
            int inLoop;             /* for loop counter */
            int32 dimSize;          /* dimension size (HDF4) */
            int32 *index;           /* dimension index (HDF4) */
            int32 numIdx;           /* number indices (HDF4) */
            long countN;            /* number of dimensions */
            size_t slenN[2];        /* array of lengths */
            long *newIndex;         /* indices (HDF5) */
            int status;             /* return status */
	    char *geoDim, *dataDim;
	    geoDim = (char *)malloc(500 * sizeof(char *));
	    dataDim = (char *)malloc(500 * sizeof(char *));

/***************************************************************************
  Load name into buffer.
***************************************************************************/
            memmove(indexMap,ptr[loop],(int) slen[loop]);
            indexMap[slen[loop]] = '\0';

/***************************************************************************
  Get info regarding data and geo dimension.
***************************************************************************/
            countN = HE5_EHparsestr(indexMap,'/',ptrN,slenN);
            memmove(geoDim,ptrN[0],(int) slenN[0]);
            geoDim[slenN[0]] = '\0';
            memmove(dataDim,ptrN[1],(int) slenN[1]);
            dataDim[slenN[1]] = '\0';

/***************************************************************************
  Get dim size.
***************************************************************************/
            dimSize = SWdiminfo(inObj,geoDim);
            index = (int32 *) calloc(dimSize,sizeof(int32));
            if (dimSize == 0)
            {
                retVal = CallocErrorMsg(__LINE__);
                break;
            }

/***************************************************************************
  Get indexed mapping information.
***************************************************************************/
            numIdx = SWidxmapinfo(inObj,geoDim,dataDim,index);
            newIndex = (long *) calloc(numIdx,sizeof(long));
            if (newIndex == NULL)
            {
                retVal = CallocErrorMsg(__LINE__);
                break;
            }

/***************************************************************************
  Copy indices into HDF5 array.
***************************************************************************/
            for (inLoop = 0; inLoop < numIdx; inLoop++)
            {
                newIndex[inLoop] = (long) index[inLoop];
            }

            if (verboseModeGlobal == CONVERT_TRUE)
            {
                printf("Working - Defining Swath indexed mapping\n Geolocation Dim:  %s\nData Dimension:  %s\n",geoDim,dataDim);
            }


/***************************************************************************
  Write indexed mapping info into HDF5 file.
***************************************************************************/
            status = HE5_SWdefidxmap(outObj,geoDim,dataDim,newIndex);
            if (status == CONVERT_FAIL)
            {
                retVal = CONVERT_FAIL;
                puts("ERROR:  HE5_SWdefidxmap failure.\n");
                break;
            }

            free(index);
            free(newIndex);
	    free(geoDim);
	    free(dataDim);
        }
        free(ptr);
        free(slen);
	free(indexMap);
    }

    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertSWorGDFieldInfo()

DESCRIPTION:
	Convert Swath or Grid geolocation/data field information 
	from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inSG            input swath/grid id (HDF4 file)
	outSG           output swath/grid id (HDF5 file)
	outFile		output file id (HDF5 file)
	typeFlag	what to convert 

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure
	GetHDF5TypeFromHDF4Type - match HDF4 data type with HDF5 equivalent

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertSWorGDFieldInfo(int32 inSG,hid_t outSG, int outFile,char *SGname,int typeFlag)
{
    char typeString[30];
    int retVal = CONVERT_SUCCESS;    /* return value */
    int32 nFields;                   /* number of fields */
    int32 bufSize;                   /* buffer length */

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        if (typeFlag == CONVERT_SWATH_GEOFIELD)
        {
            strcpy(typeString,"Swath Geolocation Fields");
        }
        else if (typeFlag == CONVERT_SWATH_DATAFIELD)
        {
            strcpy(typeString,"Swath Data Fields");
        }
        else
        {
            strcpy(typeString,"Grid Data Fields");
        }
    }

/***************************************************************************
  Get number of fields, based on type flag.
***************************************************************************/
    if (typeFlag == CONVERT_SWATH_GEOFIELD)
    {
        nFields = SWnentries(inSG,HDFE_NENTGFLD,&bufSize);
    }
    else if (typeFlag == CONVERT_SWATH_DATAFIELD)
    {
        nFields = SWnentries(inSG,HDFE_NENTDFLD,&bufSize);
    }
    else
    {
        nFields = GDnentries(inSG,HDFE_NENTDFLD,&bufSize);
    }

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("Working - Number of %s = %d\n",typeString,(int) nFields);
    }

/***************************************************************************
  If there are fields......convert.
***************************************************************************/
    if (nFields > 0)
    {
        char *fieldList;           /* field name list */
        int32 *nType;              /* data type (HDF4) */

/***************************************************************************
  Set up to get field name list.
***************************************************************************/
        fieldList = (char *) calloc(1,(int) bufSize+1);
        nType = (int32 *) calloc(nFields,sizeof(int32));
        if ((fieldList == NULL) || (nType == NULL))
        {
            return (CallocErrorMsg(__LINE__));
        }

/***************************************************************************
  Get infomation regarding field.
***************************************************************************/
        if (typeFlag == CONVERT_SWATH_GEOFIELD)
        {
            nFields = SWinqgeofields(inSG,fieldList,NULL,nType);
        }
        else if (typeFlag == CONVERT_SWATH_DATAFIELD)
        {
            nFields = SWinqdatafields(inSG,fieldList,NULL,nType);
        }
        else
        {
            nFields = GDinqfields(inSG,fieldList,NULL,nType);
        }

/***************************************************************************
  Double check, do we have any fields.
***************************************************************************/
        if (nFields > 0)
        {
            char **ptr;          /* array of field pointers */
            int loop;            /* for loop counter */
            long count;          /* number of fields */
            size_t *slen;        /* array of field string lengths */
	    char *fieldName;

	    fieldName = (char *)malloc(500 * sizeof(char *));
/***************************************************************************
  Double check, do we have any fields.
***************************************************************************/
            count = HE5_EHparsestr(fieldList,',',NULL,NULL);
            ptr = (char **) calloc(count,sizeof(char *));
	    if (ptr == NULL)
            {
                return (CallocErrorMsg(__LINE__));
            }
           slen = (size_t *) calloc(count,sizeof(size_t));
            if (slen == NULL)
            {
	      free(ptr);
	      return (CallocErrorMsg(__LINE__));
            }
            count = HE5_EHparsestr(fieldList,',',ptr,slen);

/***************************************************************************
  Loop over number of fields.
***************************************************************************/
            for (loop = 0; loop < nFields; loop++)
            {
                int hdf5Ret;             /* HDF5 function return value */
                hid_t hdf5Type;          /* data type (HDF5) */
                hssize_t hdf5Start[8];   /* start array (HDF5) */
                hsize_t hdf5Edge[8];     /* edge array (HDF5) */
                hsize_t chunkDims[8];    /* chunking dimensions */
                int newLoop;             /* for loop counter */
                int totalSize;           /* size of data in bytes */
                int totalNum;            /* total numbers read */
                int doCompression;       /* compress this field */ 
                int compParm[5] = {0,0,0,0,0};
                                         /* compression parameters */
                intn status;             /* HDF4 function return */
                int32 compCode;          /* compression code */
                int32 rank;              /* number rank (HDF4) */
                int32 dim[8];            /* edge array (HDF4) */
                int32 nt;                /* data type (HDF4) */
                int32 start[8];          /* start array (HDF4) */
                void *inData;            /* field data */
                void *fillValue;         /* fill value */
		int thereIsUnlimDim = 0;
		char *dimList=NULL, *maxdimList=NULL;

		dimList = (char *)malloc(1800 * sizeof(char *));
		maxdimList = (char *)malloc(1800 * sizeof(char *));

/***************************************************************************
  Load field into buffer. 
***************************************************************************/
                memmove(fieldName,ptr[loop],(int) slen[loop]);
                fieldName[slen[loop]] = '\0';

/***************************************************************************
  Determine if we need to do compression.  Currently (10-Jan-01) HDF5 
  only supports the DEFLATE compression method.  So, what we do is if
  the input file has compression we compress using DEFLATE.  As HDF5 
  is upgraded and allows more compression methods we can use the 
  compCode and compParm variables in HE5_SWdefcomchunk() and 
  HE5_GDdefcomchunk().
***************************************************************************/
                doCompression = CONVERT_FALSE;

                if ((compressModeGlobal == CONVERT_TRUE) && 
                                ((typeFlag == CONVERT_SWATH_DATAFIELD) ||
                                 (typeFlag == CONVERT_GRID_DATAFIELD)))
                {
                    if (typeFlag == CONVERT_SWATH_DATAFIELD)
                    {
                        status = SWcompinfo(inSG,fieldName,&compCode,
                                               compParm);
                    }
                    if (typeFlag == CONVERT_GRID_DATAFIELD)
                    {
                        status = GDcompinfo(inSG,fieldName,&compCode,
                                               compParm);
                    }

                    if ((status != FAIL) && 
                        (compCode != HDFE_COMP_NONE))
                    {
                        doCompression = CONVERT_TRUE;
       
                        if (compCode != HDFE_COMP_DEFLATE)
                        {
                            compParm[0] = 5;
                        }
                    }
                }

/***************************************************************************
  Get field information (data type, rank, dimension list).
***************************************************************************/
                if ((typeFlag == CONVERT_SWATH_GEOFIELD) || 
                    (typeFlag == CONVERT_SWATH_DATAFIELD))
                {
                    status = SWfieldinfo(inSG,fieldName,&rank,dim,
                                            &nt,dimList);
                }
                else
                {
                    status = GDfieldinfo(inSG,fieldName,&rank,dim,
                                            &nt,dimList);
                }

                if (status == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    puts("ERROR:  XXfieldinfo failure.\n");
                    break;
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Got %s information for name %s\n",
                              typeString,fieldName);
                }

/***************************************************************************
  Load field information (data type, rank, dimension list) into
  HDF5 data variables.
***************************************************************************/
                for (newLoop = 0, totalNum = 1; newLoop < rank; newLoop++)
                {
                    totalNum *= dim[newLoop];
                    start[newLoop] = 0;
                    hdf5Start[newLoop] = (hssize_t) 0;
                    hdf5Edge[newLoop] = (hsize_t) dim[newLoop];
                }

/***************************************************************************
  Get space to read in field data.
***************************************************************************/
                totalSize = totalNum * DFKNTsize(nt);
		if (verboseModeGlobal == CONVERT_TRUE)
		  {
		    printf("totalNum = %d\n", totalNum);
		    printf("totalSize = %d\n", totalSize);
		  }
                inData = (void *) calloc(1,totalSize);
                if (inData == NULL)
                {
                    retVal = CallocErrorMsg(__LINE__);
                    break;
                }

/***************************************************************************
  Read in the actual data.
***************************************************************************/
                if ((typeFlag == CONVERT_SWATH_GEOFIELD) || 
                    (typeFlag == CONVERT_SWATH_DATAFIELD))
                {
		  printf("----------------------------- Reading swath field %s\n", fieldName);
                    status = SWreadfield(inSG,fieldName,start,NULL,
                                            dim,inData);
                }
                else
                {
		  printf("----------------------------- Reading grid field %s\n", fieldName);
                    status = GDreadfield(inSG,fieldName,start,NULL,
                                            dim,inData);
                }

                if (status == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    puts("ERROR:  XXreadfield failure.\n");
                    break;
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Got %s data for name %s\n",
                              typeString,fieldName);
                }

/***************************************************************************
  Determine HDF5 data type based on HDF4 data type.
***************************************************************************/
                hdf5Type = GetHDF5TypeFromHDF4Type(nt);
                if (hdf5Type == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    puts("ERROR:  GetHDF5TypeFromHDF4Type failure.\n");
		    printf("Type that failed is %d\n", nt);
                    break;
                }

/***************************************************************************
  Define, if not already has done so, any dim that is unlimeted. Also define
                              the maxdimlist
***************************************************************************/
		strcpy(maxdimList,"");

		thereIsUnlimDim = defineUnlimitedDims(inSG, outSG, fieldName,
					     dimList, maxdimList, dim,
					     typeFlag);

/***************************************************************************
  Define and write Swath Geolocation field data.
***************************************************************************/
                if (typeFlag == CONVERT_SWATH_GEOFIELD)
                {

		  if (thereIsUnlimDim == 1)/* there is an unlimited 
					      dim in maxdimlist */
                    {
		      CalculateNumberOfChunks( inSG, fieldName, (int) rank,hdf5Edge,chunkDims);
		      hdf5Ret =  HE5_SWdefchunk(outSG,rank,chunkDims);
                    }

		  /* if dims in maxlimlist does not exist create them */
		  hdf5Ret = createDimsInMaxdimlist(outSG, maxdimList);
		  
		  hdf5Ret = HE5_SWdefgeofield(outSG,fieldName,
					      dimList,maxdimList,hdf5Type,
					      HE5_HDFE_AUTOMERGE);
		  
		  /* JRB - HE5 doc says you should detach and reattach
		     between defining and writing to a field */
		  FlushSwath(outFile, &outSG, SGname);
		  
		  /* update hdfeos5Info structore */

		  hdfeos5Info.total_hdfeos_Used_SDS = hdfeos5Info.total_hdfeos_Used_SDS + 1;
		  hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] = 
		    (char *)malloc((strlen(fieldName) +1)*sizeof(char));
		  if(hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] == NULL) {               
		    return (CallocErrorMsg(__LINE__));
		  }
		  strcpy(hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1],fieldName);

		  /* write data to field */
		  hdf5Ret = HE5_SWwritefield(outSG,fieldName,hdf5Start,
					     NULL,hdf5Edge,inData);
		  if(inData != NULL)
		    {
		      free(inData);
		      inData = NULL;
		    }
                }

/***************************************************************************
  Define and write Swath Data field data.  Swath data fields can
  have a possible fill value, this needs checked and written into
  the HDF5 file.
***************************************************************************/
                else if (typeFlag == CONVERT_SWATH_DATAFIELD)
                {
		    if (thereIsUnlimDim == 1)/* there is an unlimited 
					      dim in maxdimlist */
                    {
		      CalculateNumberOfChunks( inSG, fieldName, (int) rank,hdf5Edge,chunkDims);
		      if(rank > 0) printf("chunkDims[0]=%d\n",(int)chunkDims[0]);
		      if(rank > 1) printf("chunkDims[1]=%d\n",(int)chunkDims[1]);
		      if(rank > 2) printf("chunkDims[2]=%d\n",(int)chunkDims[2]);
		      if(rank > 3) printf("chunkDims[3]=%d\n",(int)chunkDims[3]);
		      hdf5Ret =  HE5_SWdefchunk(outSG,rank,chunkDims);
                    }

                    if (doCompression == CONVERT_TRUE)
                    {
		      CalculateNumberOfChunks( inSG, fieldName, (int) rank,hdf5Edge,chunkDims);
                        /*hdf5Ret = HE5_SWdefcomchunk(outSG,HE5_HDFE_COMP_DEFLATE,
			  compParm,(int) rank,chunkDims);*/
			hdf5Ret = HE5_SWdefcomchunk(outSG,compCode,
                                                compParm,(int) rank,chunkDims);
                    }

                    fillValue = (void *) calloc(1,DFKNTsize(nt));
                    if (fillValue == NULL)
                    {
                        retVal = CallocErrorMsg(__LINE__);
                        break;
                    }
                    status = SWgetfillvalue(inSG,fieldName,fillValue);
                    if (status != CONVERT_FAIL)
                    {
                        HE5_SWsetfillvalue(outSG,fieldName,hdf5Type,fillValue);
                    }

		    /* if dims in maxlimlist does not exist create them */
		    hdf5Ret = createDimsInMaxdimlist(outSG, maxdimList);

                    hdf5Ret = HE5_SWdefdatafield(outSG,fieldName,dimList,
                                            maxdimList,hdf5Type,HE5_HDFE_AUTOMERGE);
		    /* JRB - HE5 doc says you should detach and reattach
		       between defining and writing to a field */
		    FlushSwath(outFile, &outSG, SGname);
		    
		    /* update hdfeos5Info structore */

		    hdfeos5Info.total_hdfeos_Used_SDS = hdfeos5Info.total_hdfeos_Used_SDS + 1;
		    hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] = 
		      (char *)malloc((strlen(fieldName) +1)*sizeof(char));
		    if(hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] == NULL) {               
		      return (CallocErrorMsg(__LINE__));
		    }
		    strcpy(hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1],fieldName);

		    /* write data to field */

                    hdf5Ret = HE5_SWwritefield(outSG,fieldName,hdf5Start,
                                            NULL,hdf5Edge,inData);
		    if(inData != NULL)
		      {
			free(inData);
			inData = NULL;
		      }

                    if (doCompression == CONVERT_TRUE)
                    {
		      hdf5Ret =  HE5_SWdefchunk(outSG,rank,chunkDims);
		      hdf5Ret = HE5_SWdefcomp(outSG,HE5_HDFE_COMP_NONE,
                                            compParm);
                    }

		    /* detach and reattach after writing to a field */
		    FlushSwath(outFile, &outSG, SGname);
		    free(fillValue);
                }

/***************************************************************************
  Define and write Grid Data field data.  Grid data fields can
  have a possible fill value, this needs checked and written into
  the HDF5 file.
***************************************************************************/
                else
                {
		  /* debug
		  if(rank > 0) printf("chunkDims[0]=%d\n",(int)chunkDims[0]);
		  if(rank > 1) printf("chunkDims[1]=%d\n",(int)chunkDims[1]);
		  if(rank > 2) printf("chunkDims[2]=%d\n",(int)chunkDims[2]);
		  if(rank > 3) printf("chunkDims[3]=%d\n",(int)chunkDims[3]);
		  */
		    if (thereIsUnlimDim == 1)/* there is an unlimited 
						dim in maxdimlist */
                    {
		      CalculateNumberOfChunks( inSG, fieldName, (int) rank,hdf5Edge,chunkDims);
		      hdf5Ret =  HE5_GDdeftile(outSG,HE5_HDFE_TILE,rank,chunkDims);
                    }

                    if (doCompression == CONVERT_TRUE)
                    {
                        CalculateNumberOfChunks( inSG, fieldName, (int) rank,hdf5Edge,chunkDims);
			/*
                        hdf5Ret = HE5_GDdefcomtile(outSG,HE5_HDFE_COMP_DEFLATE,
                                               compParm,(int) rank,chunkDims);
                        hdf5Ret = HE5_GDdefcomp(outSG,HE5_HDFE_COMP_DEFLATE,
                                            compParm);
			*/
		      if(rank > 0) printf("chunkDims[0]=%d\n",(int)chunkDims[0]);
		      if(rank > 1) printf("chunkDims[1]=%d\n",(int)chunkDims[1]);
		      if(rank > 2) printf("chunkDims[2]=%d\n",(int)chunkDims[2]);
		      if(rank > 3) printf("chunkDims[3]=%d\n",(int)chunkDims[3]);
			hdf5Ret = HE5_GDdefcomtile(outSG, compCode,
                                               compParm,(int) rank,chunkDims);
                        hdf5Ret = HE5_GDdefcomp(outSG,compCode,
                                            compParm);
                    }

		    fillValue = (void *) calloc(1,DFKNTsize(nt));
                    if (fillValue == NULL)
                    {
                        retVal = CallocErrorMsg(__LINE__);
                        break;
                    }
                    status = GDgetfillvalue(inSG,fieldName,fillValue);
                    if (status != CONVERT_FAIL)
                    {
                        HE5_GDsetfillvalue(outSG,fieldName,hdf5Type,fillValue);
                    }
		    
                    hdf5Ret = HE5_GDdeffield(outSG,fieldName,dimList,
                                            maxdimList,hdf5Type,HE5_HDFE_AUTOMERGE);
		    /* JRB - HE5 doc says you should detach and reattach
		       between defining and writing to a field */
		    FlushGrid(outFile, &outSG, SGname);
		    
		    /* update hdfeos5Info structore */

		    hdfeos5Info.total_hdfeos_Used_SDS = hdfeos5Info.total_hdfeos_Used_SDS + 1;
		    hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] = 
		      (char *)malloc((strlen(fieldName) +1)*sizeof(char));
		    if( hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1] == NULL) {               
		      return (CallocErrorMsg(__LINE__));
		    }
		    strcpy(hdfeos5Info.Used_SDS[hdfeos5Info.total_hdfeos_Used_SDS - 1],fieldName);

		    /* write data to field */

                    hdf5Ret = HE5_GDwritefield(outSG,fieldName,hdf5Start,
                                            NULL,hdf5Edge,inData);
		    if(inData != NULL)
		      {
			free(inData);
			inData = NULL;
		      }

                    if (doCompression == CONVERT_TRUE)
                    {
                        hdf5Ret = HE5_GDdefcomp(outSG,HE5_HDFE_COMP_NONE,
                                            compParm);
                    }
		    if (thereIsUnlimDim == 1)/* there is an unlimited 
						dim in maxdimlist */
                    {
		      hdf5Ret =  HE5_GDdeftile(outSG,HE5_HDFE_NOTILE, rank,chunkDims);
                    }

		    /* detach and reattach after writing to a field */
		    FlushGrid(outFile, &outSG, SGname);
                    free(fillValue);
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Defined and wrote %s data for name %s\n",
                              typeString,fieldName);
                }
		if(inData != NULL)
		  {
		    free(inData);
		    inData = NULL;
		  }

		if(dimList != NULL)
		  {
		    free(dimList);
		    dimList = NULL;
		  }
		
		if(maxdimList != NULL)
		  {
		    free(maxdimList);
		    maxdimList = NULL;
		  }
            }
            free(ptr);
            free(slen);
	    free(fieldName);
        }
        free(fieldList);
        free(nType);
    }

    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertHDFEOSAttribute()

DESCRIPTION:
	Convert Swath, Point, or Grid attribute informaion from HDF4
	to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file Object id (HDF4 file)
	outObj		output file Object id (HDF5 file)
	typeFlag	what to convert (swath or grid or point)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure
	GetHDF5TypeFromHDF4Type - match HDF4 data type with HDF5 equivalent

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertHDFEOSAttribute(int32 inObj,int outObj,int typeFlag)
{
    char typeString[10];
    char *attrNames = NULL;         /* attribute names */
    int retVal = CONVERT_SUCCESS;   /* return value */
    int32 numAttrs;                 /* number of attributes */
    int32 strBufSize;               /* string buffer size */

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        if (typeFlag == CONVERT_SWATH_ATTRIBUTE)
        {
            strcpy(typeString,"Swath");
        }
        else if (typeFlag == CONVERT_GRID_ATTRIBUTE)
        {
            strcpy(typeString,"Grid");
        }
        else
        {
            strcpy(typeString,"Point");
        }
    }

/***************************************************************************
  Inquire attributes based on typeFlag.
***************************************************************************/
    switch (typeFlag)
    {
        case CONVERT_SWATH_ATTRIBUTE:
        {
            numAttrs = SWinqattrs(inObj,NULL,&strBufSize);
            attrNames = (char *) calloc(1,(int) (strBufSize + 1));
            if (attrNames == NULL)
            {
                return (CallocErrorMsg(__LINE__));
            }
            numAttrs = SWinqattrs(inObj,attrNames,&strBufSize);
        }
        break;

        case CONVERT_GRID_ATTRIBUTE:
        {
            numAttrs = GDinqattrs(inObj,NULL,&strBufSize);
            attrNames = (char *) calloc(1,(int) (strBufSize + 1));
            if (attrNames == NULL)
            {
	      return (CallocErrorMsg(__LINE__));
            }
            numAttrs = GDinqattrs(inObj,attrNames,&strBufSize);
        }
        break;

        case CONVERT_POINT_ATTRIBUTE:
        {
            numAttrs = PTinqattrs(inObj,NULL,&strBufSize);
            attrNames = (char *) calloc(1,(int) (strBufSize + 1));
            if (attrNames == NULL)
            {
                return (CallocErrorMsg(__LINE__));
            }
            numAttrs = PTinqattrs(inObj,attrNames,&strBufSize);
        }
        break;
    }

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        printf("Working - Number of %s attributes = %d\n",typeString,
                numAttrs);
    }

/***************************************************************************
  If there are attributes......convert.
***************************************************************************/
    if (numAttrs > 0)
    {
        char **ptr = NULL;        /* array of pointers */
        int loop;                 /* for loop counter */
        long count;               /* number of attributes */
        size_t *slen = NULL;      /* array of lengths */
        char *oneAttrName;

	oneAttrName = (char *)malloc(500 * sizeof(char));

/***************************************************************************
  Parse information on names of attributes.
***************************************************************************/
        count = HE5_EHparsestr(attrNames,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
        if (ptr == NULL)
        {
	  free(attrNames);
	  attrNames = NULL;
	  return (CallocErrorMsg(__LINE__));
        }
        slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
        {
	  free(ptr);
	  free(attrNames);
	  attrNames = NULL;
	  return (CallocErrorMsg(__LINE__));
        }
        count = HE5_EHparsestr(attrNames,',',ptr,slen);
   
/***************************************************************************
  Loop over number of attributes.
***************************************************************************/
        for (loop = 0; loop < count; loop++)
        {
            char *valueString;          /* attribute value */
            int size;                   /* attribute size (HDF4) */
            int realCount;              /* number of values in attribute */
            int hdf5Type;               /* attribute data type (HDF5) */
            int hdf5Ret;                /* function return (HDF5) */
            int32 nType;                /* attribute data type (HDF4) */
            int32 attrCount;            /* number of attributes */
            int32 status;               /* function return (HDF4) */
            hsize_t hdf5Size[1];        /* attribute size (HDF5) */

/***************************************************************************
  Loop over number of attributes.
***************************************************************************/
            memmove(oneAttrName,ptr[loop],(int) slen[loop]);
            oneAttrName[slen[loop]] = '\0';

/***************************************************************************
  Switch based on type of HDF-EOS object.
***************************************************************************/
            switch (typeFlag)
            {
                case CONVERT_SWATH_ATTRIBUTE:
                {

/***************************************************************************
  Get attribute information (data type and count).
***************************************************************************/
                    status = SWattrinfo(inObj,oneAttrName,&nType,
                                           &attrCount);
                    valueString = (char *) calloc(1,attrCount);
                    if (valueString == NULL)
                    {
                        retVal = CallocErrorMsg(__LINE__);
                        break;
                    }

/***************************************************************************
  Read attribute (in valueString).
***************************************************************************/
                    status = SWreadattr(inObj,oneAttrName,valueString);

/***************************************************************************
  Determine number of values in attribute.
***************************************************************************/
                    size = DFKNTsize(nType);
                    realCount = (int) (attrCount/size);

/***************************************************************************
  Set up HDF5 data type and count.
***************************************************************************/
                    hdf5Type = GetHDF5TypeFromHDF4Type(nType);
                    hdf5Size[0] = (hsize_t) realCount;

/***************************************************************************
  Write swath attribute.
***************************************************************************/
                    hdf5Ret = HE5_SWwriteattr(outObj,oneAttrName,hdf5Type,
                                          hdf5Size,(void *) valueString);
                    if ((status == CONVERT_FAIL) || (hdf5Ret== CONVERT_FAIL))
                    {
                        puts("ERROR:  failure during Swath attribute read/write\n");
                        break;
                    }
                }
                break;

                case CONVERT_GRID_ATTRIBUTE:
                {

/***************************************************************************
  Get attribute information (data type and count).
***************************************************************************/
                    status = GDattrinfo(inObj,oneAttrName,&nType,
                                           &attrCount);
                    valueString = (char *) calloc(1,attrCount);
                    if (valueString == NULL)
                    {
                        retVal = CallocErrorMsg(__LINE__);
                        break;
                    }

/***************************************************************************
  Read attribute (in valueString).
***************************************************************************/
                    status = GDreadattr(inObj,oneAttrName,valueString);

/***************************************************************************
  Determine number of values in attribute.
***************************************************************************/
                    size = DFKNTsize(nType);
                    realCount = (int) (attrCount/size);

/***************************************************************************
  Set up HDF5 data type and count.
***************************************************************************/
                    hdf5Type = GetHDF5TypeFromHDF4Type(nType);
                    hdf5Size[0] = (hsize_t) realCount;

/***************************************************************************
  Write grid attribute.
***************************************************************************/
                    hdf5Ret = HE5_GDwriteattr(outObj,oneAttrName,hdf5Type,
                                          hdf5Size,(void *) valueString);
                    if ((status == CONVERT_FAIL) || (hdf5Ret== CONVERT_FAIL))
                    {
                        puts("ERROR:  failure during Grid attribute read/write\n");
                        break;
                    }
                }
                break;

                case CONVERT_POINT_ATTRIBUTE:
                {

/***************************************************************************
  Get attribute information (data type and count).
***************************************************************************/
                    status = PTattrinfo(inObj,oneAttrName,&nType,
                                           &attrCount);
                    valueString = (char *) calloc(1,attrCount);
                    if (valueString == NULL)
                    {
                        retVal = CallocErrorMsg(__LINE__);
                        break;
                    }

/***************************************************************************
  Read attribute (in valueString).
***************************************************************************/
                    status = PTreadattr(inObj,oneAttrName,valueString);

/***************************************************************************
  Determine number of values in attribute.
***************************************************************************/
                    size = DFKNTsize(nType);
                    realCount = (int) (attrCount/size);

/***************************************************************************
  Set up HDF5 data type and count.
***************************************************************************/
                    hdf5Type = GetHDF5TypeFromHDF4Type(nType);
                    hdf5Size[0] = (hsize_t) realCount;

/***************************************************************************
  Write point attribute.
***************************************************************************/
                    hdf5Ret = HE5_PTwriteattr(outObj,oneAttrName,hdf5Type,
                                          hdf5Size,(void *) valueString);
                    if ((status == CONVERT_FAIL) || (hdf5Ret== CONVERT_FAIL))
                    {
                        puts("ERROR:  failure during Grid attribute read/write\n");
                        break;
                    }
                }
                break;
            }  /* end switch */
            free(valueString);

            if (verboseModeGlobal == CONVERT_TRUE)
            {
                printf("Working - Wrote %s attribute named %s\n",typeString,
                        oneAttrName);
            }

        }  /* end for */
        free(ptr);
        free(slen);
    }  /* end if (nummAttrs > 0) */

    free(attrNames);
    attrNames = NULL;
    return retVal;
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertFieldAttribute()

DESCRIPTION:
	Convert Swath, Point, or Grid Field attribute informaion from HDF4
	to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file Object id (HDF4 file)
	outObj		output file Object id (HDF5 file)
	typeFlag	what to convert (swath or grid or point)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:

	CallocErrorMsg		- display uniform message on calloc failure
	GetHDF5TypeFromHDF4Type - match HDF4 data type with HDF5 equivalent

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Mar 11          Abe Taaheri             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertFieldAttribute(int32 inObj,int outObj,int typeFlag)
{
  char  typeString[10];
  char *attrNames = NULL;         /* attribute names */
  int   retVal = CONVERT_SUCCESS; /* return value */
  int32 numAttrs;                 /* number of attributes */
  int32 strBufSize;               /* string buffer size */
  
  char  *fieldlist=NULL;
  int32  nfield;
  int32           rankSDS;	        /* Rank of SDS */
  int32           dum;	        /* Dummy variable */
  int32           dim_sizes[8];
   long count;               /* parsestr() return count */
  size_t *slen = NULL;      /* array of lengths */
  char **fields = NULL;        /* array of pointers */
  int32 bufSize;                   /* buffer length */
  char oneFieldName[100];       /* temporary buffer */
  int32 entrycode = HDFE_NENTDFLD;  /* entry code for data fields */
  int jloop;
  char *sds_name;

  sds_name = (char *)malloc(2048 * sizeof(char));

    /* do the following twice for swath. o
       Once for datafields, and once for geofields */
    for(jloop = 0; jloop < 2; jloop++) {
      if (typeFlag == CONVERT_SWATH_ATTRIBUTE && jloop == 1)
	{
	  entrycode = HDFE_NENTGFLD;
	}
      else if((typeFlag == CONVERT_GRID_ATTRIBUTE || 
	       typeFlag == CONVERT_POINT_ATTRIBUTE ) && jloop == 1)
	{
	  continue;
	}

      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  if (typeFlag == CONVERT_SWATH_ATTRIBUTE)
        {
            strcpy(typeString,"Swath");
        }
        else if (typeFlag == CONVERT_GRID_ATTRIBUTE)
        {
            strcpy(typeString,"Grid");
        }
        else
        {
            strcpy(typeString,"Point");
        }
    }

/***************************************************************************
  Inquire fieldnames in an Object based on typeFlag.
***************************************************************************/
    switch (typeFlag)
    {
        case CONVERT_SWATH_ATTRIBUTE:
        {
	  nfield = SWnentries(inObj, entrycode,&bufSize);
	  fieldlist = (char *) calloc(1,(int) (bufSize + 1));
	  if (fieldlist == NULL)
            {
	      free(sds_name);
	      return (CallocErrorMsg(__LINE__));
            }
	  if(jloop == 0)
	    {
	      nfield=SWinqdatafields(inObj, fieldlist, NULL,NULL);
	    }
	  else if(jloop == 1)
	    {
	      nfield=SWinqgeofields(inObj, fieldlist, NULL,NULL);
	    }
        }
        break;

        case CONVERT_GRID_ATTRIBUTE:
        {
	  nfield = GDnentries(inObj, entrycode,&bufSize);

	  fieldlist = (char *) calloc(1,(int) (bufSize + 1));
	  if (fieldlist == NULL)
            {
	      free(sds_name);
	      return (CallocErrorMsg(__LINE__));
            }
	  nfield=GDinqfields(inObj, fieldlist, NULL,NULL);
        }
        break;
	
        case CONVERT_POINT_ATTRIBUTE:
        {
	  printf("Conversion of Field attributes for Point objects is ignored.\n");
        }
        break;
    }

    if (verboseModeGlobal == CONVERT_TRUE && 
	(typeFlag == CONVERT_SWATH_ATTRIBUTE || 
	 typeFlag == CONVERT_GRID_ATTRIBUTE))
    {
        printf("Working for Native HDF written attributes - Number of %s fields = %d\n",
	       typeString, nfield);
    }

    count = HE5_EHparsestr(fieldlist,',',NULL,NULL);
    fields = (char **) calloc(count,sizeof(char *));
    if (fields == NULL)
      {
	free(fieldlist);
	free(sds_name);
	return (CallocErrorMsg(__LINE__));
      }
    slen = (size_t *) calloc(count,sizeof(long));
    if (slen == NULL)
      {
	free(fields);
	free(fieldlist);
	free(sds_name);
	return (CallocErrorMsg(__LINE__));
      }
    count = HE5_EHparsestr(fieldlist,',',fields,slen);
    nfield = (int32) count;

    if (nfield > 0)
      {
	int j;
	intn status;
	int32 sd_id, sds_id, sds_index;
	intn  status2;
	int32 sdInterfaceID=-1;
	int32           fid;
	int32           dum;	 
	int32 actualdims[HE4_MAX_VAR_DIMS];

	status = SWchkswid(inObj, "ConvertFieldAttribute", &fid, &sdInterfaceID, &dum);
	for (j=0; j<nfield; j++)
	  {
	    memmove(oneFieldName, fields[j],(int) slen[j]);
	    oneFieldName[slen[j]] = '\0';

	    /* get the attributes for each input field 
	       and write to output field */
	    /* first get SDS ID */
	    
	    if (typeFlag == CONVERT_SWATH_ATTRIBUTE)
	      {
		status2 = SWSDfldsrch(inObj, sdInterfaceID, oneFieldName,
				      &sds_id, &dum, &dum, &dum, actualdims,
				      &dum);
		if(status2 == 0 )
		  {
		    printf("\nWorking - got interface ID for swath field %s\n",oneFieldName);
		  }
	      }
	    else if (typeFlag == CONVERT_GRID_ATTRIBUTE)
	      {
		status2 = GDSDfldsrch(inObj, sdInterfaceID, oneFieldName,
				      &sds_id, &dum, &dum, &dum, actualdims,
				      &dum);
		if(status2 == 0 )
		  {
		    printf("\nWorking - got interface ID for grid field %s\n",oneFieldName);
		  }
	      }
	    else
	      {
		/*
		status2 = PTSDfldsrch(inObj, sdInterfaceID, oneFieldName,
				      &sd_id, &dum, &dum, &dum, actualdims,
				      &dum);
	       */
		free(sds_name);
		if(fieldlist != NULL)
		  {
		    free(fieldlist);
		    fieldlist = NULL;
		  }
		free(slen);
		free(fields);
		return;
	      }
	    /*
	    sd_id = SDstart(inNameGlobal,DFACC_RDWR);
	    
	    sds_index=SDnametoindex(sd_id,oneFieldName);
	    sds_id = SDselect(sd_id, sds_index);
	    */
	    if(status2 != 0)
	      {
		printf("No interface ID for field %s\n",oneFieldName);
		continue;
	      }

	    numAttrs = 0;
	    status = SDgetinfo(sds_id, sds_name, &rankSDS, dim_sizes, 
			       &dum, &numAttrs);

/***************************************************************************
  If there are attributes......convert.
***************************************************************************/
	      if (numAttrs > 0)
		{
		  if (sds_convertattrs(outObj,sds_id,oneFieldName,numAttrs,
				      typeFlag)!=CONVERT_SUCCESS) {
		    free(fieldlist);
		    free(slen);
		    free(fields);
		    return CONVERT_FAIL;
		  }
	      }
	  }
      }

    if(fieldlist != NULL)
      {
	free(fieldlist);
	fieldlist = NULL;
      }
    if(slen != NULL)
      {
	free(slen);
	slen = NULL;
      }
    if(fields != NULL)
      {
	free(fields);
	fields = NULL;
      }
    }
    
    free(sds_name);
    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertGridProjInfo()

DESCRIPTION:
	Convert Grid projection information from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file object id (HDF4 file)
	outObj		output file object id (HDF5 file)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertGridProjInfo(int32 inObj,int outObj)
{
    int retVal = CONVERT_SUCCESS;    /* return value */
    int status;                      /* functio return (HDF4) */
    double outParm[16];              /* parameter array (HDF5) */
    float64 inParm[16];              /* parameter array (HDF4) */
    int loop;                        /* for loop counter */
    int outProj;                     /* projection code (HDF5) */
    int outZone;                     /* zone code (HDF5) */
    int outSphere;                   /* sphere code (HDF5) */
    int32 inProj;                    /* projection code (HDF4) */
    int32 inZone;                    /* zone code (HDF4) */
    int32 inSphere;                  /* sphere code (HDF4) */
    int32 pixCode;                   /* pixel registration code */
    int32 origCode;                  /* origin information code */

/***************************************************************************
  Get projection information from input file (projection, zone, sphere 
  codes, parameter array).
***************************************************************************/
    status = GDprojinfo(inObj,&inProj,&inZone,&inSphere,inParm);
    if (status != CONVERT_SUCCESS)
    {
        puts("ERROR:  GDprojinfo failure.\n");
        return CONVERT_FAIL;
    }

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        puts("Working - successfully received Grid Projction Information\n");
    }

/***************************************************************************
  Convert codes and parameter array to HDF5 data types
***************************************************************************/
    outProj = (int) inProj;
    outZone = (int) inZone;
    outSphere = (int) inSphere;
    for (loop = 0; loop < 16; loop++)
    {
        outParm[loop] = (double) inParm[loop];
    }

/***************************************************************************
  Write (define) grid projection.
***************************************************************************/
    status = HE5_GDdefproj(outObj,outProj,outZone,outSphere,outParm);
    if (status != CONVERT_SUCCESS)
    {
        puts("ERROR:  HE5_GDdefproj failure.\n");
        retVal = CONVERT_FAIL;
    }

    if (verboseModeGlobal == CONVERT_TRUE)
    {
        puts("Working - successfully wrote Grid Projction Information\n");
    }

/***************************************************************************
  Get pixel registration from input file and write it out to the
  output file.
***************************************************************************/
    status = GDpixreginfo(inObj,&pixCode);
    if (status != CONVERT_SUCCESS)
    {
        puts("ERROR:  GDpixreginfo failure.\n");
        retVal = CONVERT_FAIL;
    }
    else
    {
        status = HE5_GDdefpixreg(outObj,(int) pixCode); 
        if (status != CONVERT_SUCCESS)
        {
            puts("ERROR:  HE5_GDdefpixreg failure.\n");
            retVal = CONVERT_FAIL;
        }
        else 
        {
            if (verboseModeGlobal == CONVERT_TRUE)
            {
                puts("Working - successfully wrote Grid pixel registration\n");
            }
        }
    }

/***************************************************************************
  Get origin from input file and write it out to the output file.
***************************************************************************/
    status = GDorigininfo(inObj,&origCode);
    if (status != CONVERT_SUCCESS)
    {
        puts("ERROR:  GDorigininfo failure.\n");
        retVal = CONVERT_FAIL;
    }
    else
    {
        status = HE5_GDdeforigin(outObj,(int) origCode); 
        if (status != CONVERT_SUCCESS)
        {
            puts("ERROR:  HE5_GDdeforigin failure.\n");
            retVal = CONVERT_FAIL;
        }
        else 
        {
            if (verboseModeGlobal == CONVERT_TRUE)
            {
                puts("Working - successfully wrote Grid origin information\n");
            }
        }
    }

    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	ConvertPointLevelData()

DESCRIPTION:
	Convert Point leve daa from HDF4 to HDF5.

INPUTS:
        Name            Description                     Units   Min     Max

        inObj		input file id (HDF4 file)
	outObj		output file id (HDF5 file)

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:


RETURNS:
	CONVERT_SUCCESS		- all went well
	CONVERT_FAIL		- problem

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Dec 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int ConvertPointLevelData(int32 inObj,int outObj)
{
    char *prevLevelName = NULL;         /* prev level (for links) */
    int retVal = CONVERT_SUCCESS;       /* return value */
    int loopCount;                      /* level for loop */
    int32 numLevels;                    /* number of levels */

/***************************************************************************
  Get the number of levels in the point.
***************************************************************************/
    numLevels = PTnlevels(inObj);
    if (numLevels <= 0)
    {
        puts("ERROR:  Point has no levels\n");
        retVal = CONVERT_FAIL;
    }
    else
    {
        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Converting %d levels.\n",numLevels);
        }
    }

/***************************************************************************
  Loop over the number of levels.
***************************************************************************/
    for (loopCount = 0; loopCount < numLevels; loopCount++)
    {
        char **ptr = NULL;        /* array of pointers */
        char *fieldList;          /* list of field names */
        char *levelName;          /* name of current level */
        int status;               /* return status (HDF4) */
        int fieldCount;           /* field loop counter */
        int recLoop;              /* record loop counter */
        int32 numFields;          /* number of fields */
        int32 strBufSize;         /* buffer size */
        int32 *fieldType;         /* field data type */
        int32 *fieldOrder;        /* field order */
        int32 *fieldLevel;        /* field level */
        int32 *recs;              /* records array */
        int32 recSize;            /* record size */
        int32 nRecs;              /* number of records */
        herr_t hdf5Ret;           /* return status (HDF5) */
        hsize_t rCount[1];        /* number of records to write */
        long count;               /* parsestr() return count */
        HE5_CmpDTSinfo newPoint;  /* HDF5 point structure */
        off_t offSet;             /* offset of data from start */
        size_t *slen = NULL;      /* array of lengths */
        size_t rSize;             /* size of record to write */
        void *dataBuffer;         /* actual point data */

/***************************************************************************
  Get the number of fields and allocate space.
***************************************************************************/
        numFields = PTnfields(inObj,(int32) loopCount,&strBufSize);

        fieldType = (int32 *) calloc((int) numFields,sizeof(int32));
        fieldOrder = (int32 *) calloc((int) numFields,sizeof(int32));
        fieldList = (char *) calloc(1,(int) (strBufSize + 1));
        if ((fieldType == NULL) || (fieldOrder == NULL) || (fieldList == NULL))
        {
            return (CallocErrorMsg(__LINE__));
        }
 
/***************************************************************************
  Get the level information.
***************************************************************************/
        numFields = PTlevelinfo(inObj,(int32) loopCount,fieldList,
                                fieldType,fieldOrder);
        if (numFields == CONVERT_FAIL)
        {
            puts("ERROR:  PTlevelinfo failure.\n");
            retVal = numFields;
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            break;
        }

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Level %d contains %d fields.\n",loopCount,numFields);
        }

/***************************************************************************
  Get the level name.
***************************************************************************/
        status = PTgetlevelname(inObj,(int32) loopCount,NULL,&strBufSize);
        levelName = (char *) calloc(1,(int) (strBufSize + 1));
        if (levelName == NULL)
        {
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            return (CallocErrorMsg(__LINE__));
        }
        status = PTgetlevelname(inObj,(int32) loopCount,levelName,&strBufSize);
        if (status == CONVERT_FAIL)
        {
            puts("ERROR:  PTlevelinfo failure.\n");
            retVal = status;
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            break;
        }

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Level %d named - %s\n",loopCount,levelName);
        }

/***************************************************************************
  Get space for field level info and then get the size of the 
  record and number of records.
***************************************************************************/
        fieldLevel = (int32 *) calloc((int) numFields,sizeof(int32));
        if (fieldLevel == NULL)
        {
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            return (CallocErrorMsg(__LINE__));
        }
        recSize = PTsizeof(inObj,fieldList,fieldLevel);
        nRecs = PTnrecs(inObj,loopCount);

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Level %d contains %d records of size %d bytes\n",loopCount,
                   (int) nRecs,(int) recSize);
        }

/***************************************************************************
  Now we can start to populate our HDF5 point structure.  Take care
  of numFields and recSize here.  Then break up a field names variable.
***************************************************************************/
        newPoint.nfields = (int) numFields;
        newPoint.datasize = (size_t) recSize;

        count = HE5_EHparsestr(fieldList,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
	if (ptr == NULL)
        {
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            free(fieldLevel);
            return (CallocErrorMsg(__LINE__));
        }
        slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
        {
	    free(ptr);
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            free(fieldLevel);
            return (CallocErrorMsg(__LINE__));
        }
        count = HE5_EHparsestr(fieldList,',',ptr,slen);

/***************************************************************************
  Initialize the offset and loop over the number of fields.
***************************************************************************/
        offSet = (off_t) 0;
        for (fieldCount = 0; fieldCount < newPoint.nfields; fieldCount++)
        {
            char *temString;      /* temporary string */

/***************************************************************************
  Set rank.
***************************************************************************/
            newPoint.rank[fieldCount] = 1;

/***************************************************************************
  Get field name into a string and move that value into our array
  of character pointers (which we need to allocate space for).
***************************************************************************/
            temString = (char *) calloc(1,(int) (slen[fieldCount] + 1));
            newPoint.fieldname[fieldCount] = (char *) calloc(1,(int) (slen[fieldCount] + 1));
            if ((newPoint.fieldname[fieldCount] == NULL) || (temString == NULL))
            {
                free(fieldType);
                free(fieldOrder);
                free(fieldList);
                free(levelName);
                free(fieldLevel);
                return (CallocErrorMsg(__LINE__));
            }
            memmove(temString,ptr[fieldCount],(int) slen[fieldCount]);
            temString[slen[fieldCount]] = '\0';
            strcpy(newPoint.fieldname[fieldCount],temString);
            free(temString);

/***************************************************************************
  Set dims.
***************************************************************************/
            newPoint.dims[fieldCount][0] = (size_t) fieldOrder[fieldCount];

/***************************************************************************
  Need to calculate the offsets of each variable in the record.  Place
  offset in HDF5 point structure.
***************************************************************************/
            if (fieldCount > 0)
            {
                offSet += (off_t) (DFKNTsize(fieldType[fieldCount-1]) * 
                                                    fieldOrder[fieldCount-1]);;
            }
            newPoint.offset[fieldCount] = offSet;

/***************************************************************************
  Set type and class.
***************************************************************************/
            newPoint.dtype[fieldCount] = (hid_t) GetHDF5TypeFromHDF4Type(
                                                          fieldType[fieldCount]);
            newPoint.dclass[fieldCount] = (H5T_class_t) GetHDF5ClassFromHDF4Type(
                                                          fieldType[fieldCount]);
        }
        free(ptr);
        free(slen);

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            puts("Working - HDF5 Point Structure populated\n");
        }

/***************************************************************************
  Since we finally have the structure loaded we can now define a
  new level in the output file.
***************************************************************************/
        hdf5Ret = HE5_PTdeflevel(outObj,levelName,&newPoint);
        if (hdf5Ret == CONVERT_FAIL)
        {
            printf("ERROR:  creating level - name:   %s\n",levelName);
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            free(fieldLevel);
            break;
        }

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Level %s successfully defined in output file\n",
                   levelName);
        }

/***************************************************************************
  We need to see if there is a link here.  I am going to make this
  easy and only check for backward links.  Therefore, if the counter
  is greater than zero (0) we will see if there is a link.  If so,
  define a linkage in our output file.
***************************************************************************/
        if (loopCount > 0)
        {
            char linkField[500];

            status = PTbcklinkinfo(inObj,(int32) loopCount,linkField);
            if (status == CONVERT_SUCCESS)
            {

/***************************************************************************
  There is a link, so we need to define one in our output file.
***************************************************************************/
                hdf5Ret = HE5_PTdeflinkage(outObj,prevLevelName,levelName,linkField);
                if (hdf5Ret == CONVERT_FAIL)
                {
                    printf("ERROR:  creating link with field %s\n",linkField);
                }
                else
                {
                    if (verboseModeGlobal == CONVERT_TRUE)
                    {
                        printf("Working - Defined link with field %s\n",linkField);
                    }
                }
            }
        }

/***************************************************************************
  Save the level name in case we need it for defining links.
***************************************************************************/
        if (prevLevelName != NULL)
        {
            free(prevLevelName);
            prevLevelName = (char *) NULL;
        }
        prevLevelName = (char *) calloc(1,(int) strlen(levelName));
        strcpy(prevLevelName,levelName);

/***************************************************************************
  Set up the records array to read all records.
***************************************************************************/
        recs = (int32 *) calloc((int) nRecs,sizeof(int32));
        for (recLoop = 0; recLoop < nRecs; recLoop++)
        {
            recs[recLoop] = recLoop;
        }

/***************************************************************************
  Create space for the actual level data.
***************************************************************************/
        dataBuffer = (void *) calloc(1,(int) nRecs * recSize);
        if (dataBuffer == NULL)
        {
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            free(fieldLevel);
            return (CallocErrorMsg(__LINE__));
        }

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            puts("Working - Allocated space for level data.\n");
        }

/***************************************************************************
  Read the level data.
***************************************************************************/
        status = PTreadlevel(inObj,loopCount,fieldList,nRecs,recs,dataBuffer);
        if (status == CONVERT_FAIL)
        {
            printf("ERROR:  reading level %s\n",levelName);
            retVal = status;
            free(recs);
            free(dataBuffer);
            free(fieldType);
            free(fieldOrder);
            free(fieldList);
            free(levelName);
            free(fieldLevel);
            break;
        }

        if (verboseModeGlobal == CONVERT_TRUE)
        {
            printf("Working - Successfully read %s level data.\n",levelName);
        }

/***************************************************************************
  Set up size and record count variables.
***************************************************************************/
        rCount[0] = (hsize_t) nRecs;
        rSize = (size_t) recSize;

/***************************************************************************
  Write the level data.
***************************************************************************/
        hdf5Ret = HE5_PTwritelevel(outObj,loopCount,rCount,&rSize,dataBuffer);
        if (hdf5Ret == CONVERT_FAIL)
        {
            printf("ERROR:  writing level data - level name:  %s\n",levelName);
        }
        else
        {
            if (verboseModeGlobal == CONVERT_TRUE)
            {
                printf("Working - Successfully wrote %s level data.\n",levelName);
            }
        }

        free(recs);
        free(dataBuffer);
        free(fieldType);
        free(fieldOrder);
        free(fieldList);
        free(levelName);
        free(fieldLevel);
    }

    if (prevLevelName != NULL)
    {
        free(prevLevelName);
    }

    return retVal;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	GetHDF5TypeFromHDF4Type()

DESCRIPTION:
	Get HDF5 data type from HDF4 equivalent.

INPUTS:
        Name            Description                     Units   Min     Max

        hdf4Type	HDF4 data type

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
        CONVERT_FAIL		- something went wrong
	H5T_NATIVE_FLOAT
	H5T_NATIVE_DOUBLE
	H5T_NATIVE_LDOUBLE
	H5T_NATIVE_CHAR
	H5T_NATIVE_SHORT
	H5T_NATIVE_INT
	H5T_NATIVE_LONG
	H5T_NATIVE_LLONG
	H5T_NATIVE_USHORT
	H5T_NATIVE_UINT
	H5T_NATIVE_ULONG
	H5T_NATIVE_ULLONG

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int GetHDF5TypeFromHDF4Type(int32 hdf4Type)
{
    int hdf5Type = CONVERT_FAIL;
    int sizeToMatch;
    herr_t          status      = FAIL;
/***************************************************************************
  Determine size (in bytes) of HDF4 data type.
***************************************************************************/
    sizeToMatch = DFKNTsize(hdf4Type);

/***************************************************************************
  Switch on HDF4 data type.  This is real straight forward, once we
  have switched on the type of data we just want to know what size 
  (in bytes) of HDF5 data matches what was used with HDF4.
***************************************************************************/
    switch ((int) hdf4Type)
    {

/***************************************************************************
  Float data.
***************************************************************************/
        case DFNT_FLOAT32:
        case DFNT_FLOAT64:
        {
            if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_FLOAT))
            { 
                hdf5Type = H5T_NATIVE_FLOAT;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_DOUBLE))
            { 
                hdf5Type = H5T_NATIVE_DOUBLE;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_LDOUBLE))
            { 
                hdf5Type = H5T_NATIVE_LDOUBLE;
            }
        }
        break;

/***************************************************************************
  Signed integer data.
***************************************************************************/
        case DFNT_INT8:
        case DFNT_INT16:
        case DFNT_INT32:
        {
            if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_CHAR))
            {
	      /* hdf5Type = H5T_NATIVE_CHAR;*/
	      hdf5Type = H5Tcopy(H5T_NATIVE_INT);
	      status = H5Tset_size(hdf5Type, 1);
	      if( status == FAIL)
		{
		  printf("Cannot set the total size for hdf5Type\n\n");
		  return(status);
		}
 
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_SHORT))
            {
                hdf5Type = H5T_NATIVE_SHORT;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_INT))
            { 
                hdf5Type = H5T_NATIVE_INT;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_LONG))
            { 
                hdf5Type = H5T_NATIVE_LONG;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_LLONG))
            { 
                hdf5Type = H5T_NATIVE_LLONG;
            } 
        }
        break;

/***************************************************************************
  Unsigned integer data.
***************************************************************************/
        case DFNT_UINT8:
        case DFNT_UINT16:
        case DFNT_UINT32:
        {
            if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_UCHAR))
            {
	      /* hdf5Type = H5T_NATIVE_UCHAR;*/
	      hdf5Type = H5Tcopy(H5T_NATIVE_UINT);
	      status = H5Tset_size(hdf5Type, 1);
	      if( status == FAIL)
		{
		  printf("Cannot set the total size for hdf5Type\n");
		  return(status);
		}
 
            }
	    else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_USHORT))
            {
                hdf5Type = H5T_NATIVE_USHORT;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_UINT))
            { 
                hdf5Type = H5T_NATIVE_UINT;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_ULONG))
            { 
                hdf5Type = H5T_NATIVE_ULONG;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_ULLONG))
            { 
                hdf5Type = H5T_NATIVE_ULLONG;
            } 
        }
        break;

/***************************************************************************
  Character data.
***************************************************************************/
        case DFNT_CHAR:
        case DFNT_UCHAR:
        {
            if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_CHAR))
            {
                hdf5Type = H5T_NATIVE_CHAR;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_SCHAR))
            { 
                hdf5Type = H5T_NATIVE_SCHAR;
            } 
            else if (sizeToMatch == (int) H5Tget_size(H5T_NATIVE_UCHAR))
            { 
                hdf5Type = H5T_NATIVE_UCHAR;
            } 
        }
        break;
    }  /* end switch */
 
    return hdf5Type;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	GetHDF5ClassFromHDF4Type()

DESCRIPTION:
	Get HDF5 data class from HDF4 data type.

INPUTS:
        Name            Description                     Units   Min     Max

        hdf4Type	HDF4 data type

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
        H5T_NO_CLASS		- something went wrong
	H5T_INTEGER
	H5T_FLOAT
	H5T_STRING

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int GetHDF5ClassFromHDF4Type(int32 hdf4Type)
{
    int hdf5Type;
    H5T_class_t hdf5Class;

    hdf5Type = GetHDF5TypeFromHDF4Type(hdf4Type);

    if ((hdf5Type == H5T_NATIVE_FLOAT) || 
        (hdf5Type == H5T_NATIVE_DOUBLE) || 
        (hdf5Type == H5T_NATIVE_LDOUBLE))
    {
        hdf5Class = H5T_FLOAT;
    }

    else if (hdf5Type == H5T_NATIVE_CHAR)
    {
        hdf5Class = H5T_STRING;
    }

    else if ((hdf5Type == H5T_NATIVE_SHORT) || 
             (hdf5Type == H5T_NATIVE_INT) || 
             (hdf5Type == H5T_NATIVE_LONG) || 
             (hdf5Type == H5T_NATIVE_LLONG) || 
             (hdf5Type == H5T_NATIVE_USHORT) || 
             (hdf5Type == H5T_NATIVE_UINT) || 
             (hdf5Type == H5T_NATIVE_ULONG) || 
             (hdf5Type == H5T_NATIVE_ULLONG))
    {
        hdf5Class = H5T_INTEGER;
    }

    else
    {
        hdf5Class = H5T_NO_CLASS;
    }

    return (int) hdf5Class;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	DisplayHelpAndExit()

DESCRIPTION:
	Display usage message and exit program with normal exit code (0).

INPUTS:
        Name            Description                     Units   Min     Max

        progName	name of program

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
	N/A

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
void DisplayHelpAndExit(char *progName)
{
    puts("**********************************************************************");
    printf("%s converts HDF-EOS Version 2 objects to HDF-EOS Version 5\n",progName);
    puts("**********************************************************************");
    puts("Usage:");
    printf("%s -i <input file> -o <output file> [-v][-h][-c <chunks>]\n",progName);
    printf("%s -version\n",progName);
    puts("    -i - input file name (-i flag must be used)");
    puts("    -o - output file name (-o flag must be used, file name must be a new file)");
    puts("    -v - verbose mode");
    puts("    -c <chunks> - if HDF4 file is compressed, compress (DEFLATE) HDF5 file - chunk number is optional");
    puts("    -version - utility version");
    puts("    -h - help");

    exit(0);
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	DisplayVersionAndExit()

DESCRIPTION:
	Display version of this tool and exit program with 
	normal exit code (0).

INPUTS:
        Name            Description                     Units   Min     Max

        progName	name of program

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
	N/A

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
void DisplayVersionAndExit(char *progName)
{
    printf("%s - %s\n",progName,CONVERT_VERSION);

    exit(0);
}


/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	CallocErrorMsg()

DESCRIPTION:
	Display a generic calloc error message.

INPUTS:
        Name            Description                     Units   Min     Max

        lineNum		line number that error occurred

OUTPUTS:
        Name            Description                     Units   Min     Max

	NONE

FUNCTIONS_CALLED:
	N/A

RETURNS:
	CONVERT_FAIL

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Jun 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
int CallocErrorMsg(int lineNum)
{
    printf("ERROR:  calloc failure at line %d\n",lineNum);

    return CONVERT_FAIL;
}



/***************************************************************************
BEGIN_PROLOG:

FUNCTION:
	CalculateNumberOfChunks(int32 inSG, .... )

DESCRIPTION:
	Calculate the number of chunks to use during compression.

	Algorithm:  If the user supplied a chunk size, use it.  If the 
	user did not supply a chunk size do the following:

	value = sqrt(dimension size)

	if (value is less than or equal to ten (10)) set chunking to one (1).

	Otherwise, modulo value into dimension size and keep subtracting 
	one until the modulo returns zero.  In other words the first value
	less than or equal to the square root of the dimension size that will 
	divide evenly into dimension size is our chunking size.

INPUTS:
        Name            Description                     Units   Min     Max

        numDims		number of dimensions to calculate

	realDims	dimension sizes

OUTPUTS:
        Name            Description                     Units   Min     Max

	chunkDims	chunk sizes corresponding to dim sizes

FUNCTIONS_CALLED:
	N/A

RETURNS:
	N/A

HISTORY:

        Date            Programmer              Description
        ---------       ------------------      -------------------------
        Sep 00          Ray Milburn             Original Programmer

END_PROLOG:
***************************************************************************/
void CalculateNumberOfChunks(int32 inSG, char *fieldname, int numDims, hsize_t *realDims,
                             hsize_t *chunkDims)
{
  int loop, i;
  int startPoint;
  int calcLoop;
  int32           sdInterfaceID;	/* HDF SDS interface ID */
  int32           sdid;	        /* SDS ID */
  
  int32           dum;	        /* Dummy variable */
  int32           rankSDS;	        /* Rank of SDS/Field/chunk */
  int32           chunkFlags=-1;	/* flag to determine if field is chunkd */
  HDF_CHUNK_DEF   chunkDef;	        /* union holding tiling info. */
  int32 * chunkcode=NULL;
  int32 * chunkrank=NULL;
  intn   status;
  int32  use_orig_chunks = 0;
  int32 chunkdims[8];
  int32 chunkcode1;
  int32 chunkrank1;
  int32 fid, sds_index;

  chunkcode = &chunkcode1;
  chunkrank = &chunkrank1;

    status = SWchkswid(inSG, "CalculateNumberOfChunks", &fid, &sdInterfaceID, &dum);
  /* Check whether fieldname is in SDS (multi-dim field) */
  /* --------------------------------------------------- */
      status = SWSDfldsrch(inSG, sdInterfaceID, fieldname, &sdid,
			   &rankSDS, &dum, &dum, &dum, &dum);

  for(i=0; i<8; i++)
    {
      chunkdims[i] = 0;
    }
  *chunkrank = 1;
  status=SWchunkinfo(inSG, fieldname, chunkcode, chunkrank, chunkdims);

  if(*chunkrank > rankSDS) *chunkrank = rankSDS;

  if(status != 0)/* try again using HDF */
    {
      int32 fid, sds_index;
      status = SWchkswid(inSG, "CalculateNumberOfChunks", &fid, &sdInterfaceID, &dum);
      /* Check whether fieldname is in SDS (multi-dim field) */
      /* --------------------------------------------------- */
      status = SWSDfldsrch(inSG, sdInterfaceID, fieldname, &sdid,
			   &rankSDS, &dum, &dum, &dum, &dum);
      if (status != 0) {
	HEpush(DFE_GENAPP, "CalculateNumberOfChunks", __FILE__, __LINE__);
	HEreport("SDS \"%s\" does not exist.\n", fieldname);
	status = -1;
	return;
      }
      /* Query field for tiling information */
      
      sds_index=SDnametoindex(sdInterfaceID,fieldname);
      sdid = SDselect(sdInterfaceID,sds_index);
      status = SDgetchunkinfo(sdid, &chunkDef, &chunkFlags);
      
      
      /* If field is unchunked, return unchunkd flag */
      if (chunkFlags == HDF_NONE)
	{
	  *chunkcode = HDFE_NOTILE;
	}
      /* IF field is chunkd or chunkd with compression */
      else if ((chunkFlags == HDF_CHUNK) ||
	       (chunkFlags == (HDF_CHUNK | HDF_COMP)))
	{
	  if (chunkcode != NULL)
	    {
	      *chunkcode = HDFE_TILE;
	    }
	  if (chunkrank != NULL)
	    {
	      *chunkrank = rankSDS;
	    }
	  if (chunkDims != NULL)
	    {
	      /* Assign size of chunk dimensions */
	      for (i = 0; i < rankSDS; i++)
		{
		  chunkDims[i] = chunkDef.chunk_lengths[i];
		  if(use_orig_chunks == 0 && chunkDims[i] != 0) use_orig_chunks = 1;
		}
	    }
	}
    }
  else
    {
      if(*chunkrank <= 0 || *chunkrank > 8) *chunkrank = 8;
      rankSDS = *chunkrank;

      /* if cunkdims are 1 or less than 0, we are in trouble. Thsi may lead to
	 stack overflow and core dump. So if that is the case we will adjust the chunl size */


      for (i = 0; i < rankSDS; i++)
	{
	  chunkDims[i] = chunkdims[i];
	  /*
	  if ((int) realDims[i] > 10 && (chunkDims[i] == 1)
	    {
	      chunkDims[i] =(int) sqrt((double) realDims[loop]);
	    }
	  */
	  if(use_orig_chunks == 0 && chunkDims[i] != 0) use_orig_chunks = 1;
	}
    }

      /* Use input file chunkDims if user does not want to use different chunck sizes */
      if (numTilesGlobal == CONVERT_NO_TILE && use_orig_chunks == 1) return;
  
 /***************************************************************************
  Loop over the number of dimensions.
 ***************************************************************************/
  for (loop = 0; loop < numDims; loop++)
    {
      
/***************************************************************************
  Set our starting point.  If the user entered a chunk number use that,
  if not then do a square root on the dimension size.
***************************************************************************/
      if (numTilesGlobal != CONVERT_NO_TILE)
        {
	  startPoint = realDims[loop] > numTilesGlobal ? 
	    numTilesGlobal : 
	    (int) (realDims[loop]/2);
        }
      else
        {
	  startPoint = (int) sqrt((double) realDims[loop]);
        }
      
/***************************************************************************
  If the dimension size is 10 or less, just set the chunk size to 1.
***************************************************************************/
      if ((int) realDims[loop] <= 10)
        {
	  chunkDims[loop] = 1;
        }
      else
        {
	  
/***************************************************************************
  Loop in reverse from the starting point until something divides evenly
  into the dimension size.  That is our chunk size.
***************************************************************************/
/*
            for (calcLoop = startPoint; calcLoop > 0; calcLoop--)
            {
                if (((int) realDims[loop] % calcLoop) == 0)
                {
                    break;
                }
            }
            chunkDims[loop] = (hsize_t) calcLoop;
*/
	  chunkDims[loop] = (hsize_t) startPoint;
        }
    }
}

/* JRB - adding stuff to fix bugs */

void FlushSwath(int outFile, hid_t *outSwath, char *SWname)
{
  /*int oldSwath = *outSwath;*/
  hid_t new_outSwath;
  int hdf5Ret;

  hdf5Ret = HE5_SWdetach((hid_t)*outSwath);
  if(hdf5Ret == CONVERT_FAIL)
    fprintf(stderr, "Failed, already?\n");
  new_outSwath = HE5_SWattach(outFile, SWname);
  /*
  if(outSwath != oldSwath)
    fprintf(stderr, "I KNEW this wasn't going to work!\n");
  */
  outSwath = &new_outSwath;

}

void FlushGrid(int outFile, hid_t *outGrid, char *GDname)
{
  /*int oldGrid = outGrid;*/
  hid_t new_outGrid;
  int hdf5Ret;

  hdf5Ret = HE5_GDdetach((hid_t)*outGrid);
  if(hdf5Ret == CONVERT_FAIL)
    fprintf(stderr, "Failed, already?\n");
  new_outGrid = HE5_GDattach(outFile, GDname);
  /*
  if(outGrid != oldGrid)
    fprintf(stderr, "I KNEW this wasn't going to work!\n");
  */
  outGrid = &new_outGrid;;
}


/* define dimensions that are unlimited */


int defineUnlimitedDims(int32 inSG, hid_t outSG, char *fieldName, char *dimList, char *maxdimList, int32 dim[], int typeFlag)
{
  char **ptr = NULL;
  char **ptr2 = NULL;
  char **ptr3 = NULL;
  int loopCount, loop;
  long count;
  size_t *slen = NULL;
  size_t *slen2 = NULL;
  size_t *slen3 = NULL;
  long he5DimsCount;
    hsize_t he5Dims[32];
  char *p;
  int retVal = CONVERT_SUCCESS;
  int thereIsUnlimDim=0;
  int status,  breakFlg = 0;                 /* return status */
  int  isDimUnlimited = 0;                   /* HDF4 checks for unlimited dimension */

  char *temp_maxdimList, *unlim_objName, *objName, *definedObjName, *he5dimList, *temp_str1, *temp_str2;

  temp_maxdimList = (char *)malloc(500 * sizeof(char));
  unlim_objName  = (char *)malloc(500 * sizeof(char));
  objName  = (char *)malloc(500 * sizeof(char));
  definedObjName  = (char *)malloc(500 * sizeof(char));
  he5dimList  = (char *)malloc(1800 * sizeof(char));
  temp_str1 = (char *)malloc(1800 * sizeof(char));
  temp_str2 = (char *)malloc(1800 * sizeof(char));

  /***************************************************************************
  See how many names are in the list.
  ***************************************************************************/
  count = HE5_EHparsestr(dimList,',',NULL,NULL);
  if( count == 0 ) return(retVal);

  strcpy(maxdimList,dimList);

  ptr = (char **) calloc(count,sizeof(char *));
  if (ptr == NULL)
    {
      return (CallocErrorMsg(__LINE__));
    }
  slen = (size_t *) calloc(count,sizeof(long));
  if (slen == NULL)
    {
      free(ptr);
      return (CallocErrorMsg(__LINE__));
    }

  count = HE5_EHparsestr(dimList,',',ptr,slen);

 /***************************************************************************
  Loop over the names and see if they have are unlimited
***************************************************************************/

  if( (typeFlag == CONVERT_SWATH_GEOFIELD) ||
      (typeFlag == CONVERT_SWATH_DATAFIELD))
    {
      he5DimsCount = HE5_SWinqdims(outSG, he5dimList, he5Dims);
    }
  else
    {
      he5DimsCount = HE5_GDinqdims(outSG, he5dimList, he5Dims);
    }


  ptr2 = (char **) calloc(count,sizeof(char *));
  if (ptr2 == NULL)
    {
      return (CallocErrorMsg(__LINE__));
    }
  slen2 = (size_t *) calloc(count,sizeof(long));
  if (slen2 == NULL)
    {
      free(ptr2);
      return (CallocErrorMsg(__LINE__));
    }
  count = HE5_EHparsestr(maxdimList,',',ptr2,slen2);
  strcpy(temp_maxdimList,"");


  ptr3 = (char **) calloc(he5DimsCount,sizeof(char *));
  if (ptr3 == NULL)
    {
      return (CallocErrorMsg(__LINE__));
    }
  slen3 = (size_t *) calloc(he5DimsCount,sizeof(long));
  if (slen3 == NULL)
    {
      free(ptr3);
      return (CallocErrorMsg(__LINE__));
    }

  he5DimsCount = HE5_EHparsestr(he5dimList,',',ptr3,slen3);



  breakFlg = 0;
  for (loopCount = 0; loopCount < count; loopCount++)
    {
      /* 
	 in hdf4 only the first dim can be unlimited, and for unlimited dimension
	 the SDdiminfo() returns 0 for the size, and SDgetinfo() returns actual size
      */

    
      memmove(objName,ptr[loopCount],(int) slen[loopCount]);
      objName[slen[loopCount]] = '\0';
      isDimUnlimited = 0;
      isDimUnlimited = checkobjNameForUlimitedSize(inSG, fieldName);

      strcpy(unlim_objName,"Unlim_");
      strcat(unlim_objName,objName);
      strcat(unlim_objName, "_Unlim");
      
      strcpy(temp_str1,he5dimList);

      if( (typeFlag == CONVERT_SWATH_GEOFIELD) ||
	  (typeFlag == CONVERT_SWATH_DATAFIELD))
	{
	  int32 nstatus=-1;
	  if((p = strstr(he5dimList, unlim_objName)) != NULL || (isDimUnlimited == 1 && loopCount == 0))
	    {
	      breakFlg = 0;
	      for (loop = 0; loop < he5DimsCount; loop++)
		{
		  memmove(definedObjName,ptr3[loop],(int) slen3[loop]);
		  definedObjName[slen3[loop]] = '\0';
		  if((strcmp(objName, definedObjName)) == 0)
		    {
		      breakFlg = 1;
		      break;
		    }
		}

	      /* reconstruct maxdimlist */
	      strcat(temp_maxdimList,unlim_objName);
	      if(loopCount != (count -1)) strcat(temp_maxdimList,",");
	      thereIsUnlimDim = 1;

	      if(breakFlg == 1) continue;

	      nstatus = HE5_SWdiminfo(outSG, unlim_objName);

	      if(nstatus == 0 )
	      status = HE5_SWdefdim(outSG,unlim_objName,H5S_UNLIMITED);
	    }
	  else
	    {
	      memmove(definedObjName,ptr2[loopCount],(int) slen2[loopCount]);
	      definedObjName[slen2[loopCount]] = '\0';
	      /* reconstruct maxdimlist */
	      strcat(temp_maxdimList,definedObjName);
	      if(loopCount != (count -1)) strcat(temp_maxdimList,",");
	      continue;
	    }
	}
      else
	{
	  if((p = strstr(he5dimList, unlim_objName)) != NULL)
	    {
	      breakFlg = 0;
	      for (loop = 0; loop < he5DimsCount; loop++)
		{
		  memmove(definedObjName,ptr3[loop],(int) slen3[loop]);
		  definedObjName[slen3[loop]] = '\0';
		  if((strcmp(objName, definedObjName)) == 0)
		    {
		      breakFlg = 1;
		      break;
		    }
		}

	      /* reconstruct maxdimlist */
	      strcat(temp_maxdimList,unlim_objName);
	      if(loopCount != (count -1)) strcat(temp_maxdimList,",");
	      thereIsUnlimDim = 1;

	      if(breakFlg == 1) continue;
	      status = HE5_GDdefdim(outSG,objName,(hsize_t)dim[loopCount]);
	    }
	  else
	    {
	      memmove(definedObjName,ptr2[loopCount],(int) slen2[loopCount]);
	      definedObjName[slen2[loopCount]] = '\0';
	      /* reconstruct maxdimlist */
	      strcat(temp_maxdimList,definedObjName);
	      if(loopCount != (count -1)) strcat(temp_maxdimList,",");
	      continue;
	    }
	}

      if (status == -1)
	{
	  retVal = CONVERT_FAIL;
	  break;
	}
    }
  strcpy(maxdimList,temp_maxdimList);
  
  if(ptr != NULL) free(ptr);
  if(ptr2 != NULL) free(ptr2);
  if(ptr3 != NULL) free(ptr3);
  if(slen != NULL)free(slen);
  if(slen2 != NULL)free(slen2);
  if(slen3 != NULL)free(slen3);

  free(temp_maxdimList);
  free(unlim_objName);
  free(objName);
  free(definedObjName);
  free(he5dimList);
  free(temp_str1);
  free(temp_str2);

  return(thereIsUnlimDim);
}

/* Convert dimension scales written by hdfeos. Dimension scales witten
   by HDF (such as in MOD02SSH or MOD021KM will be converted somewhere 
   else in this code */

int ConvertDimScales(int32 inObj,int outObj,int typeFlag)
{
  char typeString[30],typeString1[30],typeString2[30];
  int retVal = CONVERT_SUCCESS;            /* return value */
  int32 Geo_nFields=0, Data_nFields=0;     /* number of fields */
  int32 bufSize=1, bufSizeGeo=1;                           /* buffer length */

  if (verboseModeGlobal == CONVERT_TRUE)
    {
      if (typeFlag == CONVERT_SWATH_DIMSCALE)
        {
	  strcpy(typeString1,"Swath Geolocation Fields");
	  strcpy(typeString,"Swath Data Fields");
        }
      else
        {
	  strcpy(typeString,"Grid Data Fields");
        }
    }
  strcpy(typeString2,"Dimesion Scale");

  /***************************************************************************
  Get number of fields, based on type flag.
  ***************************************************************************/
  if (typeFlag == CONVERT_SWATH_DIMSCALE)
    {
      Geo_nFields = SWnentries(inObj,HDFE_NENTGFLD,&bufSizeGeo);
      Data_nFields = SWnentries(inObj,HDFE_NENTDFLD,&bufSize);
    }
  else if (typeFlag == CONVERT_GRID_DIMSCALE)
    {
      Data_nFields = GDnentries(inObj,HDFE_NENTDFLD,&bufSize);
    }
  else
    {
      printf("Error: Conversion of Dimension Scales for objects other than Swath and Grid is not supported.\n");
      return CONVERT_FAIL;
    }

  if (verboseModeGlobal == CONVERT_TRUE)
    {
      if (typeFlag == CONVERT_SWATH_DIMSCALE)
	{
	  printf("Working - Number of %s = %d\n",typeString1,(int) Geo_nFields);
	  printf("Working - Number of %s = %d\n",typeString,(int) Data_nFields);
	}
      else if(typeFlag == CONVERT_GRID_DIMSCALE)
	{
	  printf("Working - Number of %s = %d\n",typeString,(int) Data_nFields);
	}
    }

  /***************************************************************************
  If there are fields......convert.
  ***************************************************************************/
  if (Data_nFields > 0)
    {
      char *fieldList=NULL;           /* field name list */
      int32 *nType=NULL;              /* data type (HDF4) */
      
      /***************************************************************************
       Set up to get field name list.
      ***************************************************************************/
      fieldList = (char *) calloc(1,(int) bufSize+1);
      if (fieldList == NULL)
        {
	  return (CallocErrorMsg(__LINE__));
        }
      nType = (int32 *) calloc(Data_nFields,sizeof(int32));
      if (nType == NULL)
        {
	  free(fieldList);
	  return (CallocErrorMsg(__LINE__));
        }

      /***************************************************************************
       Get infomation regarding field.
      ***************************************************************************/
      if (typeFlag == CONVERT_SWATH_DIMSCALE)
        {
 	  Data_nFields = SWinqdatafields(inObj,fieldList,NULL,nType);
        }
      else
        {
	  Data_nFields = GDinqfields(inObj,fieldList,NULL,nType);
        }

      /***************************************************************************
       Double check, do we have any fields.
      ***************************************************************************/
      if (Data_nFields > 0)
        {
	  char **ptr=NULL;          /* array of field pointers */
	  char **dim_ptr=NULL;      /* array of dim pointers */
	  int loop;            /* for loop counter */
	  long count;          /* number of fields */
	  long dim_count;      /* number of dimensions */
	  size_t *slen=NULL;        /* array of field string lengths */
	  size_t *dim_slen=NULL;    /* array of dim string lengths */
	  char *fieldName=NULL;


	  fieldName = (char *)malloc(500 * sizeof(char));
	  /***************************************************************************
           Double check, do we have any fields.
	  ***************************************************************************/
	  count = HE5_EHparsestr(fieldList,',',NULL,NULL);
	  ptr = (char **) calloc(count,sizeof(char *));
	  if (ptr == NULL)
            {
	      free(fieldList);
	      free(nType);
	      free(fieldName);
	      return (CallocErrorMsg(__LINE__));
            }
	  slen = (size_t *) calloc(count,sizeof(size_t));
	  if (slen == NULL)
            {
	      free(ptr);
	      free(fieldList);
	      free(nType);
	      free(fieldName);
	      return (CallocErrorMsg(__LINE__));
            }
	  count = HE5_EHparsestr(fieldList,',',ptr,slen);

	  /***************************************************************************
           Loop over number of fields.
	  ***************************************************************************/
	  for (loop = 0; loop < Data_nFields; loop++)
            {
	      int hdf5Ret;             /* HDF5 function return value */
	      intn status;             /* HDF4 function return */
	      int32 rank;              /* number rank (HDF4) */
	      int32 dim[8];            /* edge array (HDF4) */
	      int32 nt;                /* data type (HDF4) */
	      void *inData=NULL;            /* field data */
	      int loopCount;
	      char *dimList=NULL;

	      dimList = (char *)malloc(1800 * sizeof(char));
	      /**************************************************************
               Load field into buffer. 
	      **************************************************************/
	      memmove(fieldName,ptr[loop],(int) slen[loop]);
	      fieldName[slen[loop]] = '\0';

	      /**************************************************************
               Get field information (data type, rank, dimension list).
	      ***************************************************************/
                if (typeFlag == CONVERT_SWATH_DIMSCALE)
                {
                    status = SWfieldinfo(inObj,fieldName,&rank,dim,
                                            &nt,dimList);
                }
                else
                {
                    status = GDfieldinfo(inObj,fieldName,&rank,dim,
                                            &nt,dimList);
                }

                if (status == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    puts("ERROR:  XXfieldinfo failure.\n");
                    break;
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Got %s information for name %s\n",
                              typeString,fieldName);
                }

		/***********************************************************
                See how many names are in the list.
		************************************************************/
		dim_count = HE5_EHparsestr(dimList,',',NULL,NULL);
		if( dim_count == 0 )
		  {
		    free(fieldList);
		    fieldList = NULL;
		    continue;
		  }
		dim_ptr = (char **) calloc(dim_count,sizeof(char *));
		if (dim_ptr == NULL)
		  {
		    free(fieldList);
		    free(nType);
		    free(slen);
		    free(ptr);
		    free(fieldName);
		    return (CallocErrorMsg(__LINE__));
		  }
		dim_slen = (size_t *) calloc(dim_count,sizeof(long));
		if (dim_slen == NULL)
		  {
		    free(fieldList);
		    free(nType);
		    free(dim_ptr);
		    free(slen);
		    free(ptr);
		    free(fieldName);
		    return (CallocErrorMsg(__LINE__));
		  }
		dim_count = HE5_EHparsestr(dimList,',',dim_ptr,dim_slen);

		for (loopCount = 0; loopCount < dim_count; loopCount++)
		  {
		    intn    buffsize=1;
		    int32   dimdata_type;
		    int32   dimsize;
		    void   *inData;            /* dim data */
		    hid_t   hdf5dimdata_Type;  /* data type (HDF5) */
		    intn    len=31;
		    char    label[32], unit[32], format[32];
		    hsize_t scount[1];
		    char    *dimName=NULL;

		    dimName = (char *)malloc(500 * sizeof(char));

		    memmove(dimName,dim_ptr[loopCount],(int) dim_slen[loopCount]);
		    dimName[dim_slen[loopCount]] = '\0';
		    if (typeFlag == CONVERT_GRID_DIMSCALE)
		      {
			buffsize = GDgetdimscale(inObj, fieldName, dimName, 
						 &dimsize, &dimdata_type, NULL);
		      }
		    else
		      {
			buffsize = SWgetdimscale(inObj, fieldName, dimName, 
						 &dimsize, &dimdata_type, NULL);
		      }
		    if (buffsize < 0)
		      {
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName=NULL;
			  }
			continue; /* No Dim Scale for this dimName */
		      }
		    inData = (void *) calloc(1,buffsize);
		    if (inData == NULL)
		      {
			retVal = CallocErrorMsg(__LINE__);
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName=NULL;
			  }
			continue;
		      }

		    if (typeFlag == CONVERT_GRID_DIMSCALE)
		      {
			buffsize = GDgetdimscale(inObj, fieldName, dimName, 
					      &dimsize, &dimdata_type, (void *)inData);
		      }
		    else
		      {
			buffsize = SWgetdimscale(inObj, fieldName, dimName, 
						 &dimsize, &dimdata_type,(void *)inData);
		      }

		    if (buffsize < 0)
		      {
			if(inData != NULL)
			  {
			    free(inData);
			    inData=NULL;
			  }
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName=NULL;
			  }
			continue; /* No Dim Scale for this dimName */
		      }

		    /* get dim scale attributes */
		    if (typeFlag == CONVERT_GRID_DIMSCALE)
		      {
			status = GDgetdimstrs(inObj, fieldName, dimName,
					      label, unit, format, len);
		      }
		    else
		      {
			status = SWgetdimstrs(inObj, fieldName, dimName,
					      label, unit, format, len);
		      }


		    if (verboseModeGlobal == CONVERT_TRUE)
		      {
			printf("Working - Got %s data for name %s of field %s\n",
			       typeString2, dimName, fieldName);
		      }
		    /*************************************************************
                      Determine HDF5 data type based on HDF4 data type.
		    **************************************************************/
		    hdf5dimdata_Type = GetHDF5TypeFromHDF4Type(dimdata_type);
		    if (hdf5dimdata_Type == CONVERT_FAIL)
		      {
			retVal = CONVERT_FAIL;
			puts("ERROR:  GetHDF5TypeFromHDF4Type failure.\n");
			printf("Type that failed is %d\n", dimdata_type);
			if(inData != NULL)
			  {
			    free(inData);
			    inData=NULL;
			  }
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName = NULL;
			  }
			continue;
		      }

		    if (typeFlag == CONVERT_GRID_DIMSCALE)
		      {

			if((strcmp(dimName, "XDim") == 0) ||
			   (strcmp(dimName, "YDim") == 0))
			  {
			    hdf5Ret = HE5_GDsetdimscale(outObj, fieldName, dimName, 
						    (hsize_t)dimsize, hdf5dimdata_Type, 
						     NULL);
			  }
			else
			  {
			    hdf5Ret = HE5_GDsetdimscale(outObj, fieldName, dimName, 
						   (hsize_t)dimsize, hdf5dimdata_Type, 
						    (void *)inData);
			  }

			scount[0]= strlen(label);
			status = HE5_GDwritedscaleattr(outObj, dimName, "label", 
						       H5T_NATIVE_CHAR, scount, label);
			scount[0]= strlen(unit);
			status = HE5_GDwritedscaleattr(outObj, dimName, "unit", 
						       H5T_NATIVE_CHAR, scount, unit);
			scount[0]= strlen(format);
			status = HE5_GDwritedscaleattr(outObj, dimName, "format", 
						       H5T_NATIVE_CHAR, scount, format);
		      }
		    else
		      {
			hdf5Ret = HE5_SWsetdimscale(outObj, fieldName, dimName, 
						   (hsize_t)dimsize, hdf5dimdata_Type, 
						   (void *)inData);
			scount[0]= strlen(label);
			status = HE5_SWwritedscaleattr(outObj, dimName, "label", 
						       H5T_NATIVE_CHAR, scount, label);
			scount[0]= strlen(unit);
			status = HE5_SWwritedscaleattr(outObj, dimName, "unit", 
						       H5T_NATIVE_CHAR, scount, unit);
			scount[0]= strlen(format);
			status = HE5_SWwritedscaleattr(outObj, dimName, "format", 
						       H5T_NATIVE_CHAR, scount, format);

		      }

		    if (inData != NULL)
		      {
			free (inData);
			inData = NULL;
		      }

		    if(dimName != NULL)
		      {
			free(dimName);
			dimName = NULL;
		      }
		  }

		if(dimList != NULL)
		  {
		    free(dimList);
		    dimList = NULL;
		  }
		if(dim_ptr !=NULL)
		  {
		    free(dim_ptr);
		    dim_ptr = NULL;
		  }
		if(dim_slen !=NULL)
		  {
		    free(dim_slen);
		    dim_slen = NULL;
		  }
	    }

	  if(fieldName != NULL)
	    {
	      free(fieldName);
	      fieldName = NULL;
	    }
	  if(ptr !=NULL)
	    {
	      free(ptr);
	      ptr = NULL;
	    }
	  if(slen != NULL)
	    {
	      free(slen);
	      slen = NULL;
	    }
	}
      if(fieldList != NULL)
	{
	  free(fieldList);
	  fieldList = NULL;
	}
      if(nType != NULL)
	{
	  free(nType);
	  nType = NULL;
	}

    }
  else if (Geo_nFields > 0)
    {
      char *fieldList=NULL;           /* field name list */
      int32 *nType=NULL;              /* data type (HDF4) */

      /***************************************************************************
       Set up to get field name list.
      ***************************************************************************/
      fieldList = (char *) calloc(1,(int) bufSizeGeo+1);
      if (fieldList == NULL)
        {
	  return (CallocErrorMsg(__LINE__));
        }
      nType = (int32 *) calloc(Geo_nFields,sizeof(int32));
      if (nType == NULL)
        {
	  free(fieldList);
	  return (CallocErrorMsg(__LINE__));
        }

      /***************************************************************************
       Get infomation regarding field.
      ***************************************************************************/
      if (typeFlag == CONVERT_SWATH_DIMSCALE)
        {
	  Geo_nFields = SWinqgeofields(inObj,fieldList,NULL,nType);
         }

      /***************************************************************************
       Double check, do we have any fields.
      ***************************************************************************/
      if (Geo_nFields > 0)
        {
	  char **ptr=NULL;          /* array of field pointers */
	  char **dim_ptr=NULL;      /* array of dim pointers */
	  int loop;            /* for loop counter */
	  long count;          /* number of fields */
	  long dim_count;      /* number of dimensions */
	  size_t *slen=NULL;        /* array of field string lengths */
	  size_t *dim_slen=NULL;    /* array of dim string lengths */
	  char *fieldName=NULL;

	  fieldName = (char *)malloc(500 * sizeof(char));
	  /***************************************************************************
           Double check, do we have any fields.
	  ***************************************************************************/
	  count = HE5_EHparsestr(fieldList,',',NULL,NULL);
	  ptr = (char **) calloc(count,sizeof(char *));
	  if (ptr == NULL)
            {
	      free(fieldList);
	      free(nType);
	      free(fieldName);
	      return (CallocErrorMsg(__LINE__));
            }
	  slen = (size_t *) calloc(count,sizeof(size_t));
	  if (slen == NULL)
            {
	      free(ptr);
	      free(fieldList);
	      free(nType);
	      free(fieldName);
	      return (CallocErrorMsg(__LINE__));
            }
	  count = HE5_EHparsestr(fieldList,',',ptr,slen);

	  /***************************************************************************
           Loop over number of fields.
	  ***************************************************************************/
	  for (loop = 0; loop < Geo_nFields; loop++)
            {
	      int hdf5Ret;             /* HDF5 function return value */
	      int doCompression;       /* compress this field */ 
	      intn status;             /* HDF4 function return */
	      int32 rank;              /* number rank (HDF4) */
	      int32 dim[8];            /* edge array (HDF4) */
	      int32 nt;                /* data type (HDF4) */
	      void *inData;            /* field data */
	      int loopCount;
	      char *dimList=NULL;

	      dimList = (char *)malloc(1800 * sizeof(char));
	      /**************************************************************
               Load field into buffer. 
	      **************************************************************/
	      memmove(fieldName,ptr[loop],(int) slen[loop]);
	      fieldName[slen[loop]] = '\0';


	      /**************************************************************
               Get field information (data type, rank, dimension list).
	      ***************************************************************/
                if (typeFlag == CONVERT_SWATH_DIMSCALE)
                {
                    status = SWfieldinfo(inObj,fieldName,&rank,dim,
                                            &nt,dimList);
                }
 
                if (status == CONVERT_FAIL)
                {
                    retVal = CONVERT_FAIL;
                    puts("ERROR:  SWfieldinfo failure.\n");
                    break;
                }

                if (verboseModeGlobal == CONVERT_TRUE)
                {
                    printf("Working - Got %s information for name %s\n",
                              typeString,fieldName);
                }

		/***********************************************************
                See how many names are in the list.
		************************************************************/
		dim_count = HE5_EHparsestr(dimList,',',NULL,NULL);
		if( dim_count == 0 ) continue;
		dim_ptr = (char **) calloc(dim_count,sizeof(char *));
		if (dim_ptr == NULL)
		  {
		    return (CallocErrorMsg(__LINE__));
		  }
		dim_slen = (size_t *) calloc(dim_count,sizeof(long));
		if (dim_slen == NULL)
		  {
		    free(dim_ptr);
		    return (CallocErrorMsg(__LINE__));
		  }

		dim_count = HE5_EHparsestr(dimList,',',dim_ptr,dim_slen);

		for (loopCount = 0; loopCount < dim_count; loopCount++)
		  {
		    intn    buffsize=1;
		    int32   dimdata_type;
		    int32   dimsize;
		    void   *inData=NULL;            /* dim data */
		    hid_t   hdf5dimdata_Type;  /* data type (HDF5) */
		    intn    len=31;
		    char    label[32], unit[32], format[32];
		    hsize_t scount[1];
		    char    *dimName=NULL;

		    dimName = (char *)malloc(500 * sizeof(char));

		    memmove(dimName,dim_ptr[loopCount],(int)dim_slen[loopCount]);
		    dimName[dim_slen[loopCount]] = '\0';
		    if (typeFlag == CONVERT_SWATH_DIMSCALE)
		      {
			buffsize = SWgetdimscale(inObj, fieldName, dimName, 
						 &dimsize, &dimdata_type, NULL);
		      }

		    if (buffsize < 0)
		      {
			free(dimName);
			dimName = NULL;
			continue; /* No Dim Scale for this dimName */
		      }
		    inData = (void *) calloc(1,buffsize);
		    if (inData == NULL)
		      {
			free(dimName);
			dimName = NULL;
			retVal = CallocErrorMsg(__LINE__);
			continue;
		      }
		    if (typeFlag == CONVERT_SWATH_DIMSCALE)
		      {
			buffsize = SWgetdimscale(inObj, fieldName, dimName, 
						 &dimsize, &dimdata_type,(void *)inData);
		      }

		    if (buffsize < 0) 
		      {
			if (inData != NULL)
			  {
			    free (inData);
			    inData = NULL;
			  }
			
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName = NULL;
			  }
			continue; /* No Dim Scale for this dimName */
		      }
		    status = SWgetdimstrs(inObj, fieldName, dimName,
					  label, unit, format, len);

		    strcpy(typeString1,"Dimesion Scale");
		    if (verboseModeGlobal == CONVERT_TRUE)
		      {
			printf("Working - Got %s data for name %s of field %s\n",
			       typeString2, dimName, fieldName);
		      }
		    /*************************************************************
                      Determine HDF5 data type based on HDF4 data type.
		    **************************************************************/
		    hdf5dimdata_Type = GetHDF5TypeFromHDF4Type(dimdata_type);
		    if (hdf5dimdata_Type == CONVERT_FAIL)
		      {
			if(dimName != NULL)
			  {
			    free(dimName);
			    dimName = NULL;
			  }
			if(inData != NULL)
			  {
			    free(inData);
			    inData = NULL;
			  }
			retVal = CONVERT_FAIL;
			puts("ERROR:  GetHDF5TypeFromHDF4Type failure.\n");
			printf("Type that failed is %d\n", dimdata_type);
			continue;
		      }

		    if (typeFlag == CONVERT_SWATH_DIMSCALE)
		      {
			hdf5Ret = HE5_SWsetdimscale(outObj, fieldName, dimName, 
						   (hsize_t)dimsize, hdf5dimdata_Type, 
						   (void *)inData);
			scount[0]= strlen(label);
			status = HE5_SWwritedscaleattr(outObj, dimName, "label", 
						       H5T_NATIVE_CHAR, scount, label);
			scount[0]= strlen(unit);
			status = HE5_SWwritedscaleattr(outObj, dimName, "unit", 
						       H5T_NATIVE_CHAR, scount, unit);
			scount[0]= strlen(format);
			status = HE5_SWwritedscaleattr(outObj, dimName, "format", 
						       H5T_NATIVE_CHAR, scount, format);
		      }

		    if (inData != NULL)
		      {
			free (inData);
			inData = NULL;
		      }
		    if(dimName!= NULL)
		      {
			free(dimName);
			dimName= NULL;
		      }
		  }
		if(dimList!= NULL)
		  {
		    free(dimList);
		    dimList = NULL;
		  }
		if(dim_ptr !=NULL)
		  {
		    free(dim_ptr);
		    dim_ptr = NULL;
		  }
		if(dim_slen !=NULL)
		  {
		    free(dim_slen);
		    dim_slen = NULL;
		  }
	    }
	  if(fieldName != NULL)
	    {
	      free(fieldName);
	      fieldName = NULL;
	    }
	  if(ptr !=NULL)
	    {
	      free(ptr);
	      ptr = NULL;
	    }
	  if(slen != NULL)
	    {
	      free(slen);
	      slen = NULL;
	    }
	}
      if(fieldList != NULL)
	{
	  free(fieldList);
	  fieldList = NULL;
	}
      if(nType != NULL)
	{
	  free(nType);
	  nType = NULL;
	}
    }
  return CONVERT_SUCCESS;
}


/* Convert dimension scales written by HDF (such as in MOD02SSH or MOD021KM ). 
   Dimension scales witten by HDF-EOS be converted somewhere else in this code */

int ConvertHDF4DimScales(char *outFileName, Hdf4DimScaleInfo *DimscaleInfo)
{
  char typeString[30],typeString1[30],typeString2[30];
  int      retVal = CONVERT_SUCCESS;            /* return value */
  int32   Geo_nFields=0, Data_nFields=0;     /* number of fields */
  long    bufSize;                           /* buffer length */
  int     typeFlag;
  hid_t   outSWid;                  /* HDF5 swath id (output) */
  hid_t   outGDid;                      /* output grid id (HDF5) */
  hid_t   outGDFid;                           /* output file HDF5 file id */
  hid_t   outSWFid;                           /* output file HDF5 file id */
  int     iobj;
  hid_t   outObj;

  HdfInfo infoPtr_Dimscale;
  char   *nameList;                /* list of object names */
  int     numObjs = 0;               /* number of objects found */
  long    strBufSize;              /* length of object name list */
  int32   objects;                 /* number of objects */
  int     iobject;
  int     loopCount;

  char fieldlist[100];
  int frank[8];
  hid_t fntype[8];


  
  for(iobj=0; iobj<2; iobj++) {
    
    if(iobj ==0 ){
      outGDFid = HE5_GDopen(outFileName,H5F_ACC_RDWR);
      if (outGDFid == CONVERT_FAIL) 
	{
	  printf("ERROR:  Unable to open Grid %s\n",outFileName);
	  return CONVERT_FAIL;
	}
      typeFlag = CONVERT_GRID_DIMSCALE;
    }
    else {
      outSWFid = HE5_SWopen(outFileName,H5F_ACC_RDWR);
      if (outSWFid == CONVERT_FAIL) 
	{
	  printf("ERROR:  Unable to open Grid %s\n",outFileName);
	  return CONVERT_FAIL;
	}
      typeFlag = CONVERT_SWATH_DIMSCALE;
    }

    if (verboseModeGlobal == CONVERT_TRUE)
      {
	if (typeFlag == CONVERT_SWATH_DIMSCALE)
	  {
	    strcpy(typeString1,"Swath Geolocation Fields");
	    strcpy(typeString,"Swath Data Fields");
	  }
	else
	  {
	    strcpy(typeString,"Grid Data Fields");
	  }
      }
    
    /* get swaths or grids, and look at them one-by-one */
    
    switch (typeFlag)
      {
      case CONVERT_SWATH_DIMSCALE:
        {
	  objects = (int32)HE5_SWinqswath(outFileName,NULL,&strBufSize);
	  if (objects > 0)
            {
	      nameList = (char *) calloc(1,(int) strBufSize + 1);
	      if (nameList == NULL)
                {
		  objects = CallocErrorMsg(__LINE__);
                }
	      else
                {
		  objects = (int32)HE5_SWinqswath(outFileName,nameList,&strBufSize);
                }
            }
        }
        break;
	
	/***************************************************************************
         Determine how many grid objects there are and get the list of 
         grid names.
	***************************************************************************/
      case CONVERT_GRID_DIMSCALE:
        {
	  objects = (int32)HE5_GDinqgrid(outFileName,NULL,&strBufSize);
	  if (objects > 0)
            {
	      nameList = (char *) calloc(1,(int) strBufSize + 1);
	      if (nameList == NULL)
                {
		  objects = CallocErrorMsg(__LINE__);
                }
	      else
                {
		  objects = (int32)HE5_GDinqgrid(outFileName,nameList,&strBufSize);
                }
            }
        }
        break;
      }
    
    /***************************************************************************
     Set up structure to contain an array of object names.
    ***************************************************************************/
    numObjs = (int) objects;
    infoPtr_Dimscale.total = numObjs;
    if (numObjs > 0)
      {
        char **ptr = NULL;
        char objName[500];
        int loopCount;
        long count;
        size_t *slen = NULL;

        infoPtr_Dimscale.name = (char **) calloc(numObjs,sizeof(char *));
        if (infoPtr_Dimscale.name == NULL)
	  {
	    free(nameList);
            return (CallocErrorMsg(__LINE__));
	  }
	
	/***************************************************************************
         See how many names are in the list.
	***************************************************************************/
        count = HE5_EHparsestr(nameList,',',NULL,NULL);
        ptr = (char **) calloc(count,sizeof(char *));
        if (ptr == NULL)
	  {
	    free(nameList);
            return (CallocErrorMsg(__LINE__));
	  }
       slen = (size_t *) calloc(count,sizeof(long));
        if (slen == NULL)
	  {
	    free(ptr);
	    free(nameList);
            return (CallocErrorMsg(__LINE__));
	  }
        count = HE5_EHparsestr(nameList,',',ptr,slen);
	
	/***************************************************************************
         Loop over the names and place them in the structure.
	***************************************************************************/
        for (loopCount = 0; loopCount < infoPtr_Dimscale.total; loopCount++)
	  {
            memmove(objName,ptr[loopCount],(int) slen[loopCount]);
            objName[slen[loopCount]] = '\0';
	    
            infoPtr_Dimscale.name[loopCount] = (char *) calloc(1,strlen(objName) + 1);
            if (infoPtr_Dimscale.name[loopCount] == NULL)
	      {
                numObjs = CallocErrorMsg(__LINE__);
                break;
	      }
            strcpy(infoPtr_Dimscale.name[loopCount],objName);
	  }
        free(ptr);
        free(slen);
      }
    else
      {
	if(iobj == 1) 
	  {
	    printf("ERROR:  No Grid or Swath found in %s\n",outFileName);
	    return CONVERT_FAIL;
	  }
	else
	  {
	    continue;
	  }
      }
    
    for(iobject = 0; iobject < infoPtr_Dimscale.total; iobject++)
      {
	if (typeFlag == CONVERT_SWATH_DIMSCALE)
	  {
	    outSWid = HE5_SWattach(outGDFid,infoPtr_Dimscale.name[iobject] );
	    outObj = outSWid;
	  }
	else
	  {
	    outGDid = HE5_GDattach(outSWFid,infoPtr_Dimscale.name[iobject] );
	    outObj = outGDid;
	  }
	
	/***************************************************************************
         Get number of fields, based on type flag.
	***************************************************************************/
	if (typeFlag == CONVERT_SWATH_DIMSCALE)
	  {
	    Geo_nFields =  (int32)HE5_SWnentries(outObj,HE5_HDFE_NENTGFLD,&bufSize);
	    Data_nFields = (int32)HE5_SWnentries(outObj,HE5_HDFE_NENTDFLD,&bufSize);
	  }
	else if (typeFlag == CONVERT_GRID_DIMSCALE)
	  {
	    Data_nFields = (int32)HE5_GDnentries(outObj,HE5_HDFE_NENTDFLD,&bufSize);
	  }
	else
	  {
	    printf("Error: Conversion of Dimension Scales for objects other than Swath and Grid is not supported.\n");
	  }

	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    if (typeFlag == CONVERT_SWATH_DIMSCALE)
	      {
		printf("Working - Number of %s = %d\n",typeString1,(int) Geo_nFields);
		printf("Working - Number of %s = %d\n",typeString,(int) Data_nFields);
	      }
	    else if(typeFlag == CONVERT_GRID_DIMSCALE)
	      {
		printf("Working - Number of %s = %d\n",typeString,(int) Data_nFields);
	      }
	  }
	
	/***************************************************************************
        If there are fields......convert.
	***************************************************************************/
	if (Data_nFields > 0)
	  {
	    char *fieldList=NULL;           /* field name list */
	    int32 *nType=NULL;              /* data type (HDF4) */
	    
	    /***************************************************************************
             Set up to get field name list.
	    ***************************************************************************/
	    fieldList = (char *) calloc(1,(int) bufSize+1);
	    if (fieldList == NULL)
	      {
		free(nameList);
		return (CallocErrorMsg(__LINE__));
	      }

	    nType = (int32 *) calloc(Geo_nFields,sizeof(int32));
	    if (nType == NULL)
	      {
		free(fieldList);
		free(nameList);
		return (CallocErrorMsg(__LINE__));
	      }
 
	    /***************************************************************************
            Get infomation regarding field.
	    ***************************************************************************/
	    if (typeFlag == CONVERT_SWATH_DIMSCALE)
	      {
		Data_nFields =  (int32)HE5_SWinqdatafields(outObj,fieldList,NULL,NULL);
	      }
	    else
	      {
		Data_nFields =  (int32)HE5_GDinqfields(outObj,fieldList,NULL,NULL);
	      }
	    
	    /***************************************************************************
             Double check, do we have any fields.
	    ***************************************************************************/
	    if (Data_nFields > 0)
	      {
		char **ptr;          /* array of field pointers */
		char **dim_ptr;      /* array of dim pointers */
		int loop;            /* for loop counter */
		long count;          /* number of fields */
		long dim_count;      /* number of dimensions */
		size_t *slen;        /* array of field string lengths */
		size_t *dim_slen;    /* array of dim string lengths */
		char *fieldName=NULL;

		fieldName = (char *)malloc(500 * sizeof(char *));
		/***************************************************************************
                Double check, do we have any fields.
		***************************************************************************/
		count = HE5_EHparsestr(fieldList,',',NULL,NULL);
		ptr = (char **) calloc(count,sizeof(char *));
		if (ptr == NULL)
		  {
		    free(fieldName);
		    free(fieldList);
		    free(nType);
		    free(nameList);
		    return (CallocErrorMsg(__LINE__));
		  }
		slen = (size_t *) calloc(count,sizeof(size_t));
		if (slen == NULL)
		  {
		    free(ptr);
		    free(fieldName);
		    free(fieldList);
		    free(nType);
		    free(nameList);
		    return (CallocErrorMsg(__LINE__));
		  }
		count = HE5_EHparsestr(fieldList,',',ptr,slen);
		
		/***************************************************************************
                 Loop over number of fields.
		***************************************************************************/
		for (loop = 0; loop < Data_nFields; loop++)
		  {
		    int hdf5Ret;             /* HDF5 function return value */
		    intn status;             /* HDF4 function return */
		    int rank;              /* number rank (HDF4) */
		    hsize_t dim[8];            /* edge array (HDF4) */
		    hid_t nt[8];                /* data type (HDF4) */
		    void *inData;            /* field data */
		    int loopCount;
		    int found_dimension;
		    char    dimName[100];
		    char *dimList=NULL, *maxdimList=NULL;


		    dimList = (char *)malloc(1800 * sizeof(char *));
		    maxdimList = (char *)malloc(1800 * sizeof(char *));
		    /**************************************************************
                    Load field into buffer. 
		    **************************************************************/
		    memmove(fieldName,ptr[loop],(int) slen[loop]);
		    fieldName[slen[loop]] = '\0';
		    
		    /* skip the field that has the same name as the dimension scale. This field is the
		       dimension scale itself */
		    if(strcmp(fieldName, DimscaleInfo->name) == 0)
		      {
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			continue; 
		      }
		    
		    /**************************************************************
                    Get field information (data type, rank, dimension list).
		    ***************************************************************/
		    if (typeFlag == CONVERT_SWATH_DIMSCALE)
		      {
			status =  HE5_SWfieldinfo(outObj,fieldName,&rank,dim,
						  nt,dimList, maxdimList);
		      }
		    else
		      {
			status =  HE5_GDfieldinfo(outObj,fieldName,&rank,dim,
						  nt,dimList, maxdimList);
		      }
		    
		    if (status == CONVERT_FAIL)
		      {
			retVal = CONVERT_FAIL;
			puts("ERROR:  XXfieldinfo failure.\n");
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			break;
		      }
		    
		    if (verboseModeGlobal == CONVERT_TRUE)
		      {
			printf("Working - Got %s information for name %s\n",
			       typeString,fieldName);
		      }
		    
		    /***********************************************************
                    See how many names are in the list.
		    ************************************************************/

		    dim_count = HE5_EHparsestr(dimList,',',NULL,NULL);
		    if( dim_count == 0 ) continue;
		    dim_ptr = (char **) calloc((dim_count),sizeof(char *));
		    if (dim_ptr == NULL)
		      {
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			free(nameList);
			return (CallocErrorMsg(__LINE__));
		      }
		    dim_slen = (size_t *) calloc(dim_count,sizeof(long));
		    if (dim_slen == NULL)
		      {
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			free(dim_ptr);
			free(nameList);
			return (CallocErrorMsg(__LINE__));
		      }
		    dim_count = HE5_EHparsestr(dimList,',',dim_ptr,dim_slen);
		    		    
		    
		    /* ======================================================== */
		    
		    found_dimension = 0;

		    for (loopCount = 0; loopCount < dim_count ; loopCount++)
		      {
			memmove(dimName,dim_ptr[loopCount],(int) dim_slen[loopCount]);
			dimName[dim_slen[loopCount]] = '\0';

			if(strcmp(dimName, DimscaleInfo->name) == 0)
			  {
			    found_dimension = 1;
			    break;
			  }
		      }

		    if(found_dimension == 1) printf("....... found dimension in the dimension list. Will set dimscale.....\n");
		    if(found_dimension == 0)
		      {
			free(dim_ptr);
			dim_ptr = NULL;
			free(dim_slen);
			dim_slen  = NULL;
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			continue; /* continue to next field in swath or grid */
		      }
		    
		    if(found_dimension == 1) /* one of the field dimesions agrees with dimscale SDS name */
		      {
			intn    buffsize;
			int32   dimdata_type;
			int32   dimsize;
			void   *inData;            
			hid_t   hdf5dimdata_Type;
			intn    len=31;
			char    label[32], unit[32], format[32];
			hsize_t scount[1];

			buffsize = DimscaleInfo->sds_buffsize;
			strcpy(dimName, DimscaleInfo->name);
			//inData = (void *) calloc(1,buffsize);
			inData = (void*) DimscaleInfo->sds_data;
			dimdata_type = DimscaleInfo->sds_dtype;
			dimsize = DimscaleInfo->dimsize;
			
			strcpy(typeString1,"Dimesion Scale");
			if (verboseModeGlobal == CONVERT_TRUE)
			  {
			    printf("Working - Got %s data for name %s of field %s\n",
				   typeString1, dimName, fieldName);
			  }

			/*************************************************************
                         Determine HDF5 data type based on HDF4 data type.
			**************************************************************/
			hdf5dimdata_Type = GetHDF5TypeFromHDF4Type(dimdata_type);
			if (hdf5dimdata_Type == CONVERT_FAIL)
			  {
			    retVal = CONVERT_FAIL;
			    puts("ERROR:  GetHDF5TypeFromHDF4Type failure.\n");
			    printf("Type that failed is %d\n", dimdata_type);
			    if(dimList != NULL)
			      {
				free(dimList);
				dimList =  NULL;
			      }
			    if(maxdimList != NULL)
			      {
				free(maxdimList);
				maxdimList=  NULL;
			      }
			    continue;
			  }

			if (typeFlag == CONVERT_GRID_DIMSCALE)
			  {
			    if((strcmp(dimName, "XDim") == 0) ||
			       (strcmp(dimName, "YDim") == 0))
			      {
				hdf5Ret = HE5_GDsetdimscale(outObj, fieldName, dimName, 
							    (hsize_t)dimsize, hdf5dimdata_Type, 
							    NULL);
			      }
			    else
			      {
				hdf5Ret = HE5_GDsetdimscale(outObj, fieldName, dimName, 
							    (hsize_t)dimsize, hdf5dimdata_Type, 
							    (void *)inData);
			      }
			    
			    for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
			      {
				scount[0]= strlen(DimscaleInfo->sds_attdata[loopCount]);
				status = HE5_GDwritedscaleattr(outObj, dimName, 
							       DimscaleInfo->sdsattr_name[loopCount], 
							       DimscaleInfo->he5_atype[loopCount], scount, 
							       DimscaleInfo->sds_attdata[loopCount]);
			      }
			    
			  }
			else
			  {
			    hdf5Ret = HE5_SWsetdimscale(outObj, fieldName, dimName, 
							(hsize_t)dimsize, hdf5dimdata_Type, 
							(void *)inData);

			    for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
			      {
				scount[0]= strlen(DimscaleInfo->sds_attdata[loopCount]);
				status = HE5_SWwritedscaleattr(outObj, dimName, 
							       DimscaleInfo->sdsattr_name[loopCount], 
							       DimscaleInfo->he5_atype[loopCount], scount, 
							       DimscaleInfo->sds_attdata[loopCount]);
			      }
			  }
		      }

		    if(dimList != NULL)
		      {
			free(dimList);
			dimList =  NULL;
		      }
		    if(maxdimList != NULL)
		      {
			free(maxdimList);
			maxdimList =  NULL;
		      }
		    if (dim_ptr != NULL)
		      {
			free(dim_ptr);
			dim_ptr = NULL;
		      }
		    if (dim_slen != NULL)
		      {
			free(dim_slen);
			dim_slen = NULL;
		      }
		  }
		/* free allocated memory for DimscaleInfo members */
		for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
		  {
		    if(DimscaleInfo->sds_attdata[loopCount] != NULL)
		      {
			free(DimscaleInfo->sds_attdata[loopCount] );
			DimscaleInfo->sds_attdata[loopCount] = NULL;
		      }
		    if(DimscaleInfo->sdsattr_name[loopCount] != NULL)
		      {
			free(DimscaleInfo->sdsattr_name[loopCount] );
			DimscaleInfo->sdsattr_name[loopCount] = NULL;
		      }
		  }
		if(fieldName != NULL)
		  {
		    free(fieldName);
		    fieldName = NULL;
		  }
		if (ptr != NULL)
		  {
		    free(ptr);
		    ptr = NULL;
		  }
		if (slen != NULL)
		  {
		    free(slen);
		    slen = NULL;
		  }
	      }

	    if(fieldList != NULL)
	      {
		free(fieldList);
		fieldList = NULL;
	      }
	    if(nType != NULL)
	      {
		free(nType);
		nType = NULL;
	      }
	  }
	else if (Geo_nFields > 0)
	  {
	    char *fieldList=NULL;           /* field name list */
	    int32 *nType=NULL;              /* data type (HDF4) */
	    
	    /***************************************************************************
            Set up to get field name list.
	    ***************************************************************************/
	    fieldList = (char *) calloc(1,(int) bufSize+1);
	    if (fieldList == NULL)
	      {
		free(nameList);
		return (CallocErrorMsg(__LINE__));
	      }

	    nType = (int32 *) calloc(Geo_nFields,sizeof(int32));
	    if (nType == NULL)
	      {
		free(fieldList);
		free(nameList);
		return (CallocErrorMsg(__LINE__));
	      }
 
	    /***************************************************************************
            Get infomation regarding field.
	    ***************************************************************************/
	    if (typeFlag == CONVERT_SWATH_DIMSCALE)
	      {
		Geo_nFields = HE5_SWinqgeofields(outObj,fieldList,NULL,NULL);
	      }
	    
	    /***************************************************************************
            Double check, do we have any fields.
	    ***************************************************************************/
	    if (Geo_nFields > 0)
	      {
		char **ptr;          /* array of field pointers */
		char **dim_ptr;      /* array of dim pointers */
		int loop;            /* for loop counter */
		long count;          /* number of fields */
		long dim_count;      /* number of dimensions */
		size_t *slen;        /* array of field string lengths */
		size_t *dim_slen;    /* array of dim string lengths */
		char *fieldName=NULL;
		
		fieldName = (char *)malloc(500 * sizeof(char *));
		/***************************************************************************
                 Double check, do we have any fields.
		***************************************************************************/
		count = HE5_EHparsestr(fieldList,',',NULL,NULL);
		ptr = (char **) calloc(count,sizeof(char *));
		if (ptr == NULL)
		  {
		    free(fieldName);
		    free(fieldList);
		    free(nType);
		    free(nameList);
		    return (CallocErrorMsg(__LINE__));
		  }
		slen = (size_t *) calloc(count,sizeof(size_t));
		if (slen == NULL)
		  {
		    free(ptr);
		    free(fieldName);
		    free(fieldList);
		    free(nType);
		    free(nameList);
		    return (CallocErrorMsg(__LINE__));
		  }
		count = HE5_EHparsestr(fieldList,',',ptr,slen);
		
		/***************************************************************************
                 Loop over number of fields.
		***************************************************************************/
		for (loop = 0; loop < Geo_nFields; loop++)
		  {
		    int hdf5Ret;             /* HDF5 function return value */
		    int doCompression;       /* compress this field */ 
		    herr_t status;             /* HDF4 function return */
		    int rank;              /* number rank (HDF4) */
		    hsize_t dim[8];            /* edge array (HDF4) */
		    hid_t nt[8];                /* data type (HDF4) */
		    void *inData;            /* field data */
		    int loopCount;
		    int found_dimension;
		    char    dimName[100];
		    char *dimList=NULL, *maxdimList=NULL;
		    
		    dimList = (char *)malloc(1800 * sizeof(char *));
		    maxdimList = (char *)malloc(1800 * sizeof(char *));
		    /**************************************************************
                     Load field into buffer. 
		    **************************************************************/
		    memmove(fieldName,ptr[loop],(int) slen[loop]);
		    fieldName[slen[loop]] = '\0';
		    
		    
		    /**************************************************************
                     Get field information (data type, rank, dimension list).
		    ***************************************************************/
		    if (typeFlag == CONVERT_SWATH_DIMSCALE)
		      {
			status = HE5_SWfieldinfo(outObj,fieldName,&rank,dim,
						 nt,dimList,maxdimList);
		      }
		    
		    if (status == CONVERT_FAIL)
		      {
			retVal = CONVERT_FAIL;
			puts("ERROR:  XXfieldinfo failure.\n");
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			break;
		      }
		    
		    if (verboseModeGlobal == CONVERT_TRUE)
		      {
			printf("Working - Got %s information for name %s\n",
			       typeString,fieldName);
		      }
		    
		    /***********************************************************
                     See how many names are in the list.
		    ************************************************************/
		    dim_count = HE5_EHparsestr(dimList,',',NULL,NULL);
		    if( dim_count == 0 ) continue;
		    dim_ptr = (char **) calloc(dim_count,sizeof(char *));
		    if (dim_ptr == NULL)
		      {
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			free(nameList);
			return (CallocErrorMsg(__LINE__));
		      }
		    dim_slen = (size_t *) calloc(dim_count,sizeof(long));
		    if (dim_slen == NULL)
		      {
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			free(nameList);
			free(dim_ptr);
			return (CallocErrorMsg(__LINE__));
		      }
		    dim_count = HE5_EHparsestr(dimList,',',dim_ptr,dim_slen);
		    

		    /* ========================================================= */
		    
		    found_dimension = 0;
		    for (loopCount = 0; loopCount < dim_count; loopCount++)
		      {
			memmove(dimName,ptr[loopCount],(int) slen[loopCount]);
			dimName[slen[loopCount]] = '\0';
			if(strcmp(dimName, DimscaleInfo->name) == 0)
			  {
			    found_dimension = 1;
			    break;
			  }
		      }
		    
		    if(found_dimension == 0)
		      {
			free(dim_ptr);
			dim_ptr = NULL;
			free(dim_slen);
			dim_slen  = NULL;
			if(dimList != NULL)
			  {
			    free(dimList);
			    dimList =  NULL;
			  }
			if(maxdimList != NULL)
			  {
			    free(maxdimList);
			    maxdimList=  NULL;
			  }
			continue; /* continue to next field in swath or grid */
		      }
		    if(found_dimension == 1) /* one of the field dimesions agrees with dimscale SDS name */
		      {
			intn    buffsize;
			int32   dimdata_type;
			int32   dimsize;
			void   *inData;            /* dim data */
			hid_t   hdf5dimdata_Type;  /* data type (HDF5) */
			intn    len=31;
			char    label[32], unit[32], format[32];
			hsize_t scount[1];
			buffsize=DimscaleInfo->sds_buffsize;
			strcpy(dimName, DimscaleInfo->name);
			//inData = (void *) calloc(1,buffsize);
			inData = (void*) DimscaleInfo->sds_data;
			dimdata_type = DimscaleInfo->sds_dtype;
			dimsize = DimscaleInfo->dimsize;
			
			strcpy(typeString1,"Dimesion Scale");
			if (verboseModeGlobal == CONVERT_TRUE)
			  {
			    printf("Working - Got %s data for name %s of field %s\n",
				   typeString2, dimName, fieldName);
			  }
			/*************************************************************
                         Determine HDF5 data type based on HDF4 data type.
			**************************************************************/
			hdf5dimdata_Type = GetHDF5TypeFromHDF4Type(dimdata_type);
			if (hdf5dimdata_Type == CONVERT_FAIL)
			  {
			    retVal = CONVERT_FAIL;
			    puts("ERROR:  GetHDF5TypeFromHDF4Type failure.\n");
			    printf("Type that failed is %d\n", dimdata_type);
			    continue;
			  }
			
			if (typeFlag == CONVERT_SWATH_DIMSCALE)
			  {
			    hdf5Ret = HE5_SWsetdimscale(outObj, fieldName, dimName, 
							(hsize_t)dimsize, hdf5dimdata_Type, 
							(void *)inData);
			    
			    for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
			      {
				scount[0]= strlen(DimscaleInfo->sds_attdata[loopCount]);
				status = HE5_SWwritedscaleattr(outObj, dimName, 
							       DimscaleInfo->sdsattr_name[loopCount], 
							       DimscaleInfo->he5_atype[loopCount], scount, 
							       DimscaleInfo->sds_attdata[loopCount]);
			      }
			  }
		      }
		    if (dim_ptr != NULL)
		      {
			free(dim_ptr);
			dim_ptr = NULL;
		      }
		    if (dim_slen != NULL)
		      {
			free(dim_slen);
			dim_slen = NULL;
		      }

		    if (dimList != NULL)
		      {
			free(dimList);
			dimList = NULL;
		      }
		    if (maxdimList != NULL)
		      {
			free(maxdimList);
			maxdimList = NULL;
		      }
		  }
		/* free allocated memory for DimscaleInfo members */
		for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
		  {
		    if(DimscaleInfo->sds_attdata[loopCount] != NULL)
		      {
			free(DimscaleInfo->sds_attdata[loopCount] );
			DimscaleInfo->sds_attdata[loopCount] = NULL;
		      }
		    if(DimscaleInfo->sdsattr_name[loopCount] != NULL)
		      {
			free(DimscaleInfo->sdsattr_name[loopCount] );
			DimscaleInfo->sdsattr_name[loopCount] = NULL;
		      }
		  }
		if(fieldName != NULL)
		  {
		    free(fieldName);
		    fieldName = NULL;
		  }
		if (ptr != NULL)
		  {
		    free(ptr);
		    ptr = NULL;
		  }
		if (slen != NULL)
		  {
		    free(slen);
		    slen = NULL;
		  }
	      }

	    if(fieldList != NULL)
	      {
		free(fieldList);
		fieldList = NULL;
	      }
	    if(nType != NULL)
	      {
		free(nType);
		nType = NULL;
	      }
	  }
      }
    
    if(nameList != NULL)
      {
	free(nameList);
	nameList = NULL;
      } 
  }
return CONVERT_SUCCESS;
}
 

  /*-------------------------------------------------------------------
   * Function:	ConvertOtherGlobalAttr()
   *
   * Purpose:     converts global attributes of HDF4 file (possibly written by native HDF4)
   *              into hdf-eos5 attributes in HDFEOS INFORMATION group
   *              
   * Return:	FAIL if failed, SUCCEED if successful.
   *
   * In :	        
   sd_id:        SD identifier for HDF4 file
   fid_out:      out file id
 
   *------------------------------------------------------------------
   */	

intn ConvertOtherGlobalAttr(int32 sd_id, int fid_out, char *inFileName, char *outFileName)
{
  int         check_glosds = 1;
  hid_t       HDFfid;
  int32       num_glattrs; /* number of SD interface attributes */
  int32       num_sds;     /* number of SDS objects */
  static int  OnetimeFlage = 0;
  int       i;         /* temporary variable. */
  char      sdattr_name[2*HE4HE5_MAX_NC_NAME]; /* SDS attribute name*/
  int32     count_sdadata; /* number of elements of SD attributes.*/
  int32     sd_atype; /* SD attribute datatype */
  void*     sd_adata = NULL; /* attribute data*/
  size_t    h4_amemsize;/* size of attribute data type in memory. */
  size_t    h4_asize;  /* size of attribute data type in disk. */

  char      indstr[32]; /* index string format converted from
			   decimal index */
  char*     sdtempattr_name; /* a temporary variable to store SD attribute name */
  hsize_t    temp_countadata[1]; /* temporary variable to describe the number of
				   SD attributes */
  hid_t     h5str_type; /* HDF5 attribute data type in disk when the data type
			   is string. */
  hid_t     h5a_sid; /* HDF5 attribute space id */
  hid_t     h5a_id; /* HDF5 attribute id */
  hid_t     h5_atype; /* HDF5 attribute data type in disk */
  hid_t     h5_amemtype; /* HDF5 attribute data type in memory */
  hid_t     he5_atype;        /* HDF5 attribute data type in disk */
  char      temp_Att_Name[8];
  hid_t     dummy_ntype;
  hsize_t   dummy_count;
  uint32    temp_countatdata;/* temporary variable to describe the number of
				 SDS attributes */
  herr_t    ret=CONVERT_SUCCESS;  /* the temporary variable to check 
				     the "return" status  of
				     HDF5 APIs */
 
  if (verboseModeGlobal == CONVERT_TRUE && OnetimeFlage == 0)
    {
      printf("\n Searching for global attributes other than Core and Archive Metadata.\n");
      OnetimeFlage = 1;
    }

  HDFfid =(hid_t) fid_out;

  /* 
     temporarily close output and open it with hdfeos call. 
     At the end of this routine or before returning from this
     routine we will recover HDFfid with H5Fopen
  */

  H5Fclose(HDFfid);
	  
  HDFfid = HE5_EHopen(outFileName, H5F_ACC_RDWR, H5P_DEFAULT);

  /* obtain number of SD attributes. */
  num_sds = 0;
  num_glattrs = 0;
  if(SDfileinfo(sd_id,&num_sds,&num_glattrs) == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("Cannot start SD interface for getting attribute information\n");
	}
      RecoverOutputFid(HDFfid, &fid_out, outFileName);
      return CONVERT_FAIL;
  }

  printf("number of SDS =%d\n",(int)num_sds);
  printf("number of Global Attrinutes =%d\n",num_glattrs);

  /***************************************************************************
  If there are attributes......convert.
  ***************************************************************************/
  if (num_glattrs > 0)
    {
      for (i = 0;i < num_glattrs; i++) {
	
	if (SDattrinfo(sd_id,i,sdattr_name,&sd_atype,
		       &count_sdadata) == CONVERT_FAIL){
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("\n Cannot start SD attribute information\n");
	    }
	  RecoverOutputFid(HDFfid, &fid_out, outFileName);
	  return CONVERT_FAIL;
	}
	/* if attribute is coreMetadata, ArchiveMeadta, structMetadata, 
	   Product specific Metadata, or HDFEOSVersion,
	   skip it
	*/
	strncpy(temp_Att_Name, sdattr_name,7);
	temp_Att_Name[7]='\0';
	(void)strupr_S(temp_Att_Name);

	if((strcmp(temp_Att_Name,"COREMET") == 0) ||
	   (strcmp(temp_Att_Name,"ARCHIVE") == 0) ||
	   (strcmp(temp_Att_Name,"STRUCTM") == 0) ||
	   (strcmp(temp_Att_Name,"PRODUCT") == 0) ||
	   (strcmp(temp_Att_Name,"HDFEOSV") == 0) )
	  {
	    printf("Skiping ....... %s\n",sdattr_name);
	    continue;
	  }

	/* convert HDF4 attribute data type to HDF5 attribute data type. */ 
	if(h4type_to_h5type(sd_atype,&h5_amemtype,&h4_amemsize,
			    &h4_asize,&h5_atype)== CONVERT_FAIL) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("\n Cannot convert SD attribute data type from H4 to H5\n");
	    }
	  RecoverOutputFid(HDFfid, &fid_out, outFileName);
	  return CONVERT_FAIL;
	}

	if(count_sdadata != 0) {
	  /* allocating memory for SD attribute data. */
	  sd_adata = calloc((h4_amemsize * count_sdadata + 1),sizeof(char *));
	  if(sd_adata == NULL) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("Cannot allocate memory for SD attribute data\n");
		RecoverOutputFid(HDFfid, &fid_out, outFileName);
		return (CallocErrorMsg(__LINE__));
	      }
	  }
	  
	  if(SDreadattr(sd_id,i,(VOIDP)sd_adata)== CONVERT_FAIL) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("\n Cannot read SD attribute\n");
	      }
	    free(sd_adata);
	    RecoverOutputFid(HDFfid, &fid_out, outFileName);
	    return CONVERT_FAIL;
	  }	  
	  temp_countadata[0] = (hsize_t)count_sdadata;
	}
	else {/* treat the number of data as 1 for empty attribute*/
	  temp_countatdata = 1;
	  temp_countadata[0] = (hsize_t)temp_countatdata;
	  strcpy(sd_adata,"");
	  he5_atype = GetHDF5TypeFromHDF4Type(DFNT_CHAR);
	  
	  /* if an attribute doesn't have a name, a default name is set. */
	  if(sdattr_name[0] == '\0') {
	    sdtempattr_name = (char *)calloc(strlen("HDF4_SD")+strlen("ATTR")+8, sizeof(char));
	    
	    if(sdtempattr_name == NULL) {
	      if (verboseModeGlobal == CONVERT_TRUE)
		{
		  printf("cannot allocate memory for object name\n");
		}
	      free(sd_adata);
	      RecoverOutputFid(HDFfid, &fid_out, outFileName);
	      return CONVERT_FAIL;
	    }
	    
	    sprintf(indstr, "%d",i);
	    strcpy(sdtempattr_name,"HDF4_SD");
	    strcat(sdtempattr_name,"_");
	    strcat(sdtempattr_name,"ATTR");
	    strcat(sdtempattr_name,"_");
	    strcat(sdtempattr_name,indstr);
	    strncpy(sdattr_name,sdtempattr_name,strlen(sdtempattr_name));
	    free(sdtempattr_name);
	  }
	}
	
	/* convert HDF4 attribute data type to HDF-EOS5 attribute data type. */
	if((he5_atype = GetHDF5TypeFromHDF4Type(sd_atype))== CONVERT_FAIL) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("cannot convert SD attribute data type from H4 to He5 for attr. %d = %s", 
		     i,sdattr_name);
	    }
	  free(sd_adata);
	  RecoverOutputFid(HDFfid, &fid_out, outFileName);
	  return CONVERT_FAIL;
	}
	
	ret=-1;
	if (ret != CONVERT_SUCCESS) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf(" Setting global attribute %s\n",sdattr_name); 
	    }
	  
	  ret = 0;
	  ret = HE5_EHwriteglbattr(HDFfid,
				   sdattr_name, he5_atype, 
				   temp_countadata, sd_adata);
	  
	  if(ret != CONVERT_SUCCESS)
	    {
	      free(sd_adata);
	      RecoverOutputFid(HDFfid, &fid_out, outFileName);
	      return CONVERT_FAIL;
	    }
	  free(sd_adata);
	  sd_adata = NULL;
	  HE5_EHset_error_on(2, 0);/* reset both HDFEOS5 and HDF5 error writing */
	}
	if(sd_adata != NULL) 
	  {
	    free(sd_adata); 
	    sd_adata = NULL;
	  }
      }
      RecoverOutputFid(HDFfid, &fid_out, outFileName);
      return ret;
     }
  else
    {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("No other global attributes to write\n");
	}
      RecoverOutputFid(HDFfid, &fid_out, outFileName);
      return CONVERT_SUCCESS;
    }
}


/*********************************************************************** 

   H4toH5datatype and h4type_to_h5type used from H4 to H5 converter of 
   HDF Group

************************************************************************/
		  
int  H4toH5datatype(const   int32 h4type, 
		    hid_t*  h5typeptr,
		    hid_t*  h5memtypeptr,
		    size_t* h4sizeptr,
		    size_t* h4memsizeptr){
  /* will wait for user's response to add more options*/
  return h4type_to_h5type(h4type,h5memtypeptr,h4memsizeptr,
			  h4sizeptr,h5typeptr);
}
/*-------------------------------------------------------------------
 * Function:	h4type_to_h5type  free(corrected);

 *
 * Purpose:	this function will convert HDF4 datatype into HDF5 
                datatype.
                The converter includes file to file datatype and 
		datasize conversion, file to memory datatype and 
		datasize conversion. 
		Check the mapping document for details.

 * Return:	false, FAIL. otherwise,SUCCEED.
  *
 * In :	        h4type: HDF4 datatype
                h4toh5id:  h4toh5 identifier

 * Out:         h4size: the file(disk) size of h4type.
                h4memsize: the real memory size of h4type.
 *              h5memtype: pointer of which value should be h5memtype
                (the real type stored at the memory).
                h5type:    pointer of which value should be h5type
		(the HDF5 type that is stored at the disk).
 *	     
 *
 *-----------------------------------------------------------------
 */	
		  
int  h4type_to_h5type(const int32 h4type, 
		      hid_t* h5memtype,
		      size_t* h4memsize,
		      size_t* h4size, 
		      hid_t *h5type)
{

   switch (h4type) {

  case DFNT_CHAR8:

    *h4size = 1;
    *h4memsize = sizeof(int8);
    /* assume DFNT_CHAR8 C type character. */
    *h5memtype = H5T_STRING;
    *h5type =  H5T_STRING;
    break;

  case DFNT_UCHAR8:

    *h4size = 1;
    *h4memsize = sizeof(int8);
    *h5memtype = H5T_STRING;
    *h5type = H5T_STRING;
    break;

  case DFNT_INT8:

    *h4size = 1;
    *h5type = H5T_STD_I8BE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype = H5T_NATIVE_SCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed INT8\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_UINT8:

    *h4size =1;
    *h5type = H5T_STD_U8BE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsigned INT8\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NINT8:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 1;
    *h5type = H5T_NATIVE_SCHAR;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_SCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
       if (verboseModeGlobal == CONVERT_TRUE)
	 {
	   printf("cannot convert native INT8\n");
	 }
       return CONVERT_FAIL;
    }
    break;

  case DFNT_NUINT8:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 1;
    *h5type = H5T_NATIVE_UCHAR;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsighed naive INT8 \n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LINT8:
    *h4size = 1;
    *h5type = H5T_STD_I8LE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian INT8\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LUINT8:
    *h4size = 1;
    *h5type = H5T_STD_U8LE;
    *h4memsize = sizeof(int8);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian unsigned INT8\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_INT16:
    *h4size = 2;
    *h5type = H5T_STD_I16BE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_UINT16:
    *h4size = 2;
    *h5type = H5T_STD_U16BE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsigned int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NINT16:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 2;
    *h5type = H5T_NATIVE_SHORT;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert native int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NUINT16:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 2;
    *h5type = H5T_NATIVE_USHORT;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsigned native int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LINT16:
    *h4size = 2;
    *h5type = H5T_STD_I16LE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LUINT16:
    *h4size = 2;
    *h5type = H5T_STD_U16LE;
    *h4memsize = sizeof(int16);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian unsigned int16\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_INT32:
    *h4size = 4;
    *h5type = H5T_STD_I32BE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed int32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_UINT32:
    *h4size = 4;
    *h5type = H5T_STD_U32BE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsigned int32\n");

	}
       return CONVERT_FAIL;
    }
    break;

  case DFNT_NINT32:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 4;
    *h5type = H5T_NATIVE_INT;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert native signed int32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NUINT32:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting results may not be correct.\n");*/
    *h4size =4;
    *h5type = H5T_NATIVE_UINT;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed int32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LINT32:
    *h4size =4;
    *h5type = H5T_STD_I32LE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian signed int32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LUINT32:
    *h4size =4;
    *h5type = H5T_STD_U32LE;
    *h4memsize = sizeof(int32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian unsigned int32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_INT64:
    *h4size = 8;
    *h5type = H5T_STD_I64BE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_LLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_UINT64:
    *h4size = 8;
    *h5type = H5T_STD_U64BE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_ULLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert unsigned int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NINT64:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 8;
    *h5type = H5T_NATIVE_INT;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_LLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert native signed int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NUINT64:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting results may not be correct.\n");*/
    *h4size =8;
    *h5type = H5T_NATIVE_UINT;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_ULLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert signed int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LINT64:
    *h4size =8;
    *h5type = H5T_STD_I64LE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_CHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
     else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_LLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian signed int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LUINT64:
    *h4size =8;
    *h5type = H5T_STD_U64LE;
    *h4memsize = sizeof(long);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
     else if(*h4memsize == H5Tget_size(H5T_NATIVE_LLONG))
      *h5memtype = H5T_NATIVE_ULLONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian unsigned int64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_FLOAT32:
    *h4size =4;
    *h5type = H5T_IEEE_F32BE;
    *h4memsize = sizeof(float32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert float32\n");
	}
      return CONVERT_FAIL;
    }
    break;
  
  case DFNT_FLOAT64:
    *h4size = 8;
    *h5type = H5T_IEEE_F64BE;
    *h4memsize = sizeof(float64);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LDOUBLE))
      *h5memtype = H5T_NATIVE_LDOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert float64\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NFLOAT32:
    /*printf("warning, Native HDF datatype is encountered\n");
    printf(" the converting results may not be correct.\n");*/
    *h4size = 4;
    *h5type = H5T_NATIVE_FLOAT;
    *h4memsize = sizeof(float32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert native float32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_NFLOAT64:
/*    printf("warning, Native HDF datatype is encountered");
    printf(" the converting result may not be correct.\n");*/
    *h4size = 8;
    *h5type = H5T_NATIVE_DOUBLE;
    *h4memsize = sizeof(float64);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LDOUBLE))
      *h5memtype = H5T_NATIVE_LDOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert native float64\n");
	}
      return CONVERT_FAIL;	
    }
    break;

  case DFNT_LFLOAT32:
    *h4size = 4;
    *h5type = H5T_IEEE_F32LE;
    *h4memsize = sizeof(float32);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian float32\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_LFLOAT64:
    *h4size = 8;
    *h5type = H5T_IEEE_F64LE;
    *h4memsize = sizeof(float64);
    if(*h4memsize == H5Tget_size(H5T_NATIVE_FLOAT))
      *h5memtype = H5T_NATIVE_FLOAT;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_DOUBLE))
      *h5memtype = H5T_NATIVE_DOUBLE;
    else if(*h4memsize == H5Tget_size(H5T_NATIVE_LDOUBLE))
      *h5memtype = H5T_NATIVE_LDOUBLE;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert little-endian float64\n");
	}
      return CONVERT_FAIL;
    }
    break;

    default: {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot find the corresponding datatype in HDF5\n");
    
	}
      return CONVERT_FAIL;
    }
  }
  return CONVERT_TRUE;
}


/*------------------------------------------------------------
 * Function:	mkstr
 *
 * Purpose:     make HDF5 string type
 *              
 * Return:	type
 *
 * In :	        
                 size:      String Size
		H5T_str_t: pad
 
 *------------------------------------------------------------
 */	

hid_t mkstr(size_t size, 
	    H5T_str_t pad) {

  hid_t type; /* HDF5 data type id. */
 
  if((type=H5Tcopy(H5T_C_S1))<0) {
    printf("cannot make H5 String type\n");
    return -1;
  }
  if((H5Tset_size(type,size))<0) {
    printf("cannot set size for H5 String \n");
    return -1;
  }
  if((H5Tset_strpad(type,pad))<0){
    printf("cannot set H5 String Pad\n");
    return -1;
  }
  return type;
}


hid_t H5Acreate_safe(hid_t loc_id, const char *name, hid_t type_id, hid_t space_id, hid_t acpl_id)
{
  char *corrected = NULL;
  const char *safename = name;
  hid_t aid = -1;

  /* attribute name matters only if netCDF-4 matters */
  /*
#ifdef HAVE_LIBHDFEOS
  if (H4toH5config_use_netcdf4_hack()) {
    if ((corrected = H4toH5correct_name_netcdf4(h4toh5id, name)) == NULL) {
      H4TOH5ERROR_NOMEM1("escape name");
      return aid;
    }
    safename = corrected;
    free(corrected);
  }
#endif
  */
  /* The second chance is allowed. */
  H5E_BEGIN_TRY {
    aid = H5Acreate(loc_id, safename, type_id, space_id, acpl_id);
  } H5E_END_TRY;

  return aid;
}


void strlwr_S
(
    char *str		/* I/O:  string to convert */
)
{
    while ( *str )
    {
	*str = tolower( *str );
	str++;
    }
}

void strupr_S
(
    char *str		/* I/O:  string to convert */
)
{
    while ( *str )
    {
	*str = toupper( *str );
	str++;
    }
}


int RecoverOutputFid(hid_t HDFfid, int *fid_out, char *outFileName)
{
  herr_t   status = FAIL;/* routine return status variable */
  int retVal = CONVERT_SUCCESS;           /* return value */
  /* Call HE5_EHclose to perform file close */
  /* -------------------------------------- */
  status = HE5_EHclose(HDFfid);
  if(status == CONVERT_FAIL)
    {
      retVal = CONVERT_FAIL;
      printf("ERROR: Unable to close output file %s\n",outFileName);
      return (retVal);
    }
  else
    {
      *fid_out = (int) H5Fopen(outFileName, H5F_ACC_RDWR,H5P_DEFAULT);
      if (*fid_out == CONVERT_FAIL) 
	{
	  retVal = CONVERT_FAIL; 
	  printf("ERROR: Unable to open output file %s\n",outFileName);
	  return retVal;
	}
    }
  retVal = CONVERT_SUCCESS;
  return retVal;
}



/*************************************************************************

MODULE: ConvertLoneVdatas

PURPOSE: 
    This module converts all Vdata and their attributes that are not
    written by hdfeos2. Basically these are Vdatas that are under root group in 
    hdf4 file are created using HDF4 calls
 

RETURN VALUE:

Value    	Description
-----		-----------
CONVERT_SUCCESS	Returns CONVERT_SUCCESS if successful.
                Returns CONVERT_FAIL otherwise

HISTORY:
Version    Date     Programmer      Code     Reason
-------    ----     ----------      ----     ------
           04/12    Abe Taaheri              

NOTES:

**************************************************************************/

int ConvertLoneVdatas(char *inFileName, char *outFileName)
{
  int32  file_id; /* HDF4 file id */
  int32  vdata_id; /* HDF4 Vdata id */
  int32  *ref_vdata_array; /* array to store Vdata reference number */
  int32  istat; /* check the "return" status  of HDF4 APIs */
  int    lone_vd_index; /* index of lone Vdata */
  int    num_lonevd; /* number of lone Vdata */
  int   attr_flag;
  char *vdata_name, *h5groupname, *vdata_class;

  vdata_name = (char *)malloc(VGNAMELENMAX  * sizeof(char *));
  h5groupname = (char *)malloc(VGNAMELENMAX  * sizeof(char *));
  vdata_class = (char *)malloc( VSNAMELENMAX  * sizeof(char *));

  attr_flag  = H425_ALLATTRS;

  file_id = Hopen(inFileName, DFACC_READ, 0);
  if(file_id == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("Cannot start/open hdf4 file %s for Vdata conversion\n",inFileName);
      }
    free(vdata_name);
    free(h5groupname);
    free(vdata_class);
    return CONVERT_FAIL;
  }

  Vstart(file_id);

  num_lonevd = VSlone(file_id,NULL,0);
  if (num_lonevd == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain amount of Vdata ");
      }
    free(vdata_name);
    free(h5groupname);
    free(vdata_class);
    return CONVERT_FAIL;
  }

  printf("Total number of lone vdata = %d\n",num_lonevd);

  if (num_lonevd > 0) {
    
    ref_vdata_array = (int32 *)malloc(sizeof(int32) *(num_lonevd));
    
    num_lonevd = VSlone(file_id,ref_vdata_array,num_lonevd);
    
    if(num_lonevd == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to obtain amount of Vdata ");
	}
      free(ref_vdata_array);
      free(vdata_name);
      free(h5groupname);
      free(vdata_class);
      return CONVERT_FAIL;
    }

    /* walk through all lone vdatas. */
    
    for(lone_vd_index = 0; lone_vd_index < num_lonevd;lone_vd_index++) 
      {
	vdata_id = VSattach(file_id,ref_vdata_array[lone_vd_index],"r");
	
	if(vdata_id == CONVERT_FAIL) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("unable to obtain Vdata id");
	    }
	  free(ref_vdata_array);
	  free(vdata_name);
	  free(h5groupname);
	  free(vdata_class);
	  return CONVERT_FAIL;
	}

	ZeroMemory(vdata_name,VGNAMELENMAX);
	VSQueryname(vdata_id,vdata_name);

	/* Make sure this vdata is not an attribute of other hdf4 objects.*/
	printf("\n Converting vadata vdata_name = %s\n",vdata_name);
	if(!VSisattr(vdata_id)) {
	  
	  ZeroMemory(vdata_class,VGNAMELENMAX);
	  istat = VSgetclass(vdata_id,vdata_class);
	  if(istat == CONVERT_FAIL) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("unable to get Vdata class name");
	      }
	    free(ref_vdata_array);
	    VSdetach(vdata_id);
	    free(vdata_name);
	    free(h5groupname);
	    free(vdata_class);
	    return CONVERT_FAIL;
	  }
	  
	  if(!strncmp(vdata_class,_HDF_CHK_TBL_CLASS,strlen(_HDF_CHK_TBL_CLASS))){
	    
             VSdetach(vdata_id);     
	     continue;
	  }
          
	  strcpy(h5groupname,"HDFEOS INFORMATION");
	  if(He4toHe5vdata(file_id,vdata_id, outFileName,h5groupname,NULL,attr_flag)== CONVERT_FAIL) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("cannot translate independent vdata to HDF5 datasets.");
	      }
	    free(ref_vdata_array);
	    VSdetach(vdata_id);
	    free(vdata_name);
	    free(h5groupname);
	    free(vdata_class);
	    return CONVERT_FAIL;
	  }
	}
	istat = VSdetach(vdata_id);
	if(istat == CONVERT_FAIL) {
	    if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("cannot detach Vdata interface");
	      }
	  free(ref_vdata_array);
	  free(vdata_name);
	  free(h5groupname);
	  free(vdata_class);
	  return CONVERT_FAIL;
	}
      }

    free(ref_vdata_array);
  }
  free(vdata_name);
  free(h5groupname);
  free(vdata_class);
  Hclose(file_id);
  return CONVERT_SUCCESS;   
}


/*-------------------------------------------------------------------------
 * Function:	He4toHe5vdata
 *
 * Purpose:     translate Vdata object into hdf5 dataset 
 *              
 * Return:	FAIL if failed, SUCCEED if successful.
 *
 * In :	        file_id: hdf4 file id
                vdata_id: RI identifier
		h5filename: hdf5 file name
		h5groupname: hdf5 group name
		h5dsetname: hdf5 dataset name
		attr_flag: flag for attribute
		
   Out:

   Modification:
 *-------------------------------------------------------------------------
 */	

int He4toHe5vdata(int32 file_id,
		  int32 vdata_id,
		  char* outFileName,
		  char* h5groupname, 
		  char* h5dsetname,
		  int attr_flag)
{
  /* define variables for HDF4. */
  int32     istat; /* the temporary variable to check the return status of
		      HDF4 interfaces. */
  int32     n_records; /* number of records */
  int32     vdata_ref; /* Vdata reference number */
  int32     vdata_tag; /* Vdata tag */
  int32     interlace_mode; /* interlace mode of Vdata */
  int32     vdata_size; /* size of vdata "struct" in disk */
  int32     vdatamem_size; /* size of vdata "struct" in memory */
  int32     field_index; /* the index of the field */
  int32     fieldorder; /* the order of the field */
  int32     fieldtype; /* the data type of the field */
  int       i,temp; /* temporary variables */
  int32     nfields; /* number of field */
  int       num_vd_attrs; /* number of Vdata attributes */
  int       num_vd_field_attrs; /* number of Vdata field attributes */

  VOIDP     vd_data; /* data buffer to store Vdata */

  char      vdlabel[10]; /* the label of Vdata */
 
  char *vdata_name=NULL, *fieldname=NULL, *vdata_class=NULL, *field_name_list=NULL;

  /* define varibles for hdf5. */

  hid_t     h5d_sid; /* The data space id of HDF5 */
  hid_t     h5dset; /* The HDF5 dataset id */
  hid_t     h5_ctype; /* The HDF5 file compound data type converted from
			 Vdata data type */
  hid_t     h5_cmemtype; /* The HDF5 memory compound data type converted from
			    Vdata data type */
  hid_t*    h5memtype = NULL; /* The atomic HDF5 memory data type */
  hid_t*    h5type = NULL; /* The atomic HDF5 file data type */
  size_t*   h4memsize = NULL; /* The memory size of HDF4 data type */
  size_t*   h4size = NULL; /* The file size of HDF4 data type */
  hsize_t   h5_vddims[1]; /* The size of the dimension of the HDF5 dataset
			     converted from Vdata record */
  hsize_t   h5_maxvddims[1]; /* The maximum size of the dimension of the 
				HDF5 dataset converted from Vdata record */
  hsize_t   h5_chunkdims[1]; /* The chunking size of the the dimension of the 
				HDF5 dataset converted from Vdata record */
  hid_t     create_plist; /* the creation property list of HDF5 dataset */
  char*     h5cvdata_name; /* The absolute path of HDF5 dataset converted from
			      Vdata */
  char*     h5cvdata_name1;
  char*     ori_h5cvdata_name; /* The absolute path of the first touched HDF5 
				  dataset converted from Vdata with this 
				  reference number */
  int       check_vdata=0; /* a flag to check the status of the Vdata */
  int*      check_lookupptr; /* the pointer that passes the value of the status
				of the Vdata */
  hid_t     h5_group; /* HDF5 group id */
  herr_t    ret; /* the flag to check the "return" status of HDF5 APIs. */

  int	    empty_vdata;  /* whether or not vdata ie empty */
  int check_name;   /* a temporary flag to check name*/
  int set_name=CONVERT_SUCCESS;

  char* h5tablefieldname;
  char* h5tablefield="FIELD_";
  char* h5tablename="_NAME";
  char  findex_str[MAXREF_LENGTH];
  hid_t outFid;
  int check_lookup_updatename;

  vdata_name = (char *)malloc(HE4HE5_MAX_NC_NAME  * sizeof(char *));
  fieldname = (char *)malloc(HE4HE5_MAX_NC_NAME  * sizeof(char *));
  vdata_class = (char *)malloc( VSNAMELENMAX  * sizeof(char *));
  field_name_list = (char *)malloc(VSFIELDMAX*FIELDNAMELENMAX  * sizeof(char *));

  temp = -1;
  check_name = 0;  
  check_lookup_updatename = 0;

  check_lookupptr=&temp;
/* Zeroing out memory for vdlabel,vdata_class,vdata_name */
  ZeroMemory(vdata_name,HE4HE5_MAX_NC_NAME);
  ZeroMemory(fieldname,HE4HE5_MAX_NC_NAME);
  ZeroMemory(vdata_class,VSNAMELENMAX);
  ZeroMemory(field_name_list,VSFIELDMAX*FIELDNAMELENMAX);
  ZeroMemory(vdlabel,10);
  if (VSQueryvsize(vdata_id,&vdata_size) == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to query vdata size");
      }
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  } 

  if(vdata_size == 0) return CONVERT_SUCCESS;
  
  /* obtain number of records, field name list, vdata name. */
  if(VSinquire(vdata_id,&n_records,&interlace_mode,
	       field_name_list,&vdata_size,vdata_name) == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata imformation");
      }
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  }
  
  printf(" ...Number of records in this vadat = %d\n ...List of fields in this vdata = %s\n",  n_records, field_name_list); 

  empty_vdata = n_records == 0;
  
  /* get absolute path of hdf5 vdata name. */
  
  vdata_ref =  VSQueryref(vdata_id);
  if (vdata_ref == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata reference number");
      }
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  }
  
  /* obtain HDF5 group id */
  
  outFid  = H5Fopen(outFileName,H5F_ACC_RDWR,H5P_DEFAULT);

  /*outFid = HE5_EHopen(outFileName, H5F_ACC_RDWR, H5P_DEFAULT);*/
  if(outFid  < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain HDF5 file id");
      }
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  }
  
  h5_group    =  get_h5groupid(h5groupname,outFid);
  if(h5_group < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain HDF5 group id");
      }
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  }
  
  if(h5dsetname == NULL) {
    /* Here we have to be careful:
       1) We assume the user will not use the string
       "HDF4_VD_<ref>" where <ref> is the reference number
       of the Vdata as their own Vdata name. 
       2) We will check whether this Vdata has been visited
       or not. If this Vdata has been visited, it must be assigned a
       name "HDF4_VD_<ref>". We should check whether "HDF4_VD_<ref>"
       has been used. 
       3) So the procedure we are using is:
       i) obtain the Vdata name by using get_h5datasetname.
       ii) check whether this name has been visited,
       due to the possible name clashing among Vdatas(two Vdatas may
       share the same name), we will use routine "lookup_name" to assure
       the non-confliction use of this Vdata name.
    */
    
    h5cvdata_name = get_h5datasetname(h5groupname,vdata_name,
				      vdata_ref,HDF4_VDATA,OBJECT_HASHSIZE,check_lookupptr);
    
    if(check_name/*=lookup_name(h5cvdata_name,OBJECT_HASHSIZE,
		   temph4toh5id->name_hashtab)*/==1) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("this object name has been used");
	}
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      return CONVERT_FAIL;
    }
  }
  else {
    h5cvdata_name = get_h5datasetname(h5groupname,h5dsetname,
				      vdata_ref,HDF4_VDATA,OBJECT_HASHSIZE,check_lookupptr);
    if(*check_lookupptr == 1) {/* the user-input vdata name is used.*/
      free(h5cvdata_name);
      H5Gclose(h5_group);
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("this dataset name is used by another object");
	}
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      return CONVERT_FAIL;
    }
  }
  /*Update the HDF5 dataset table.*/
    if(check_lookup_updatename /*= lookup_updatename(h5cvdata_name,OBJECT_HASHSIZE,temph4toh5id->name_hashtab)*/==-1) {
    H5Gclose(h5_group);
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("this vdata name has been used");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    return CONVERT_FAIL;
  }
  /* check whether this vdata is touched, if not; put this vdata name into 
     the vdata table; this will be used to create hardlink when multiple 
     groups include one vdata object.*/
 
  if(check_vdata == 0) {
    /* put the absolute path of vdata into "hashing table".*/
    if(set_name/*=set_name(vdata_ref,VD_HASHSIZE,temph4toh5id->vd_hashtab,
		h5cvdata_name,-1)*/==CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot set object name");
	}
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }
  else {/* create a hard link and return back.*/
    /*ori_h5cvdata_name = get_name(vdata_ref,VD_HASHSIZE,
      temph4toh5id->vd_hashtab,&temp);*/
    ori_h5cvdata_name = vdata_name;
    if(ori_h5cvdata_name == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot obtain the name of HDF5 dataset converted from Vdata");
	}
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    if(H5Glink(h5_group,H5G_LINK_HARD,ori_h5cvdata_name,h5cvdata_name) <0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot set up hard link to original dataset name ");
	}
      free(h5cvdata_name);
      free(ori_h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    free(ori_h5cvdata_name);
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    ret = H5Gclose(h5_group);
    if (ret < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot set up hard link to original dataset name ");
	}     
      return CONVERT_FAIL;
    }
    return CONVERT_SUCCESS;
  }
  
  vdata_tag = VSQuerytag(vdata_id);
  if (vdata_tag == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("fail to obtain Vdata tag ");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  /* get the class name */
  
  if(VSgetclass(vdata_id,vdata_class) == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain Vdata class name");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  /* get number of record,field_name,Size of a record and
     Name of the vdata*/ 
  
  if(VSQueryvsize(vdata_id,&vdata_size)==FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot query Vdata size ");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(vdata_size == 0) {/* empty vdata set. */
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_SUCCESS;
  }
  
  vdatamem_size = 0;
  vdata_size = 0;
  nfields = VFnfields(vdata_id);
  
  if (nfields == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain number of Vdata fields");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  h5memtype = calloc((size_t)nfields,sizeof(hid_t));
  h5type    = calloc((size_t)nfields,sizeof(hid_t));
  h4memsize = calloc((size_t)nfields,sizeof(size_t));
  h4size    = calloc((size_t)nfields,sizeof(size_t));
  
  if(h5memtype == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf( "cannot allocate HDF5 memory type");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(h5type == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate HDF5  type");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(h5memtype == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate HDF5 memory type");
      }
    free(h5cvdata_name);
    free(h5type);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(h4memsize == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate HDF5 memory size");
      }
    free(h5cvdata_name);
    free(h5memtype);
    free(h5type);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(h4size == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate HDF5  size");
      }
    free(h5cvdata_name);
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  for (i=0;i<nfields;i++) {
    
    /* obtain field type. */
    fieldtype = VFfieldtype(vdata_id,i);
    if(fieldtype == CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot obtain Vdata field type");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    
    /* obtain field order.*/
    fieldorder = VFfieldorder(vdata_id,i);  
    
    if(fieldorder == CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot obtain Vdata field order");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }

    /* datatype conversion from hdf4 to hdf5.
       the corresponding memory data type is also converted.*/
    if(h4type_to_h5type(fieldtype,&h5memtype[i],&h4memsize[i],
			&h4size[i],&h5type[i])== CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert HDF4 data type to HDF5 data type");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }

    vdatamem_size +=fieldorder*h4memsize[i];
    vdata_size +=fieldorder*h4size[i];
    
  }
  
  vd_data = malloc(vdatamem_size*n_records);   

  if(vd_data == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate memory for Vdata");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if (!empty_vdata) {
    /* The following function should not be called when the vdata is empty; otherwise, this function returns FAIL */
    istat   = VSsetfields(vdata_id,field_name_list); 
    
    if(istat == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot set Vdata fields");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    
    istat = VSread(vdata_id,(uint8*)vd_data,n_records,FULL_INTERLACE);
    
    if(istat == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot read Vdata");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }

  for (i=0;i<nfields;i++) {

    /* obtain field order.*/
    fieldorder = VFfieldorder(vdata_id,i);  
    if(fieldorder == CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot obtain field order");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }

  /*  create HDF5 compound datatype for both memory and file.*/
  
  h5_ctype       = H5Tcreate(H5T_COMPOUND,(size_t)vdata_size);
  h5_cmemtype    = H5Tcreate(H5T_COMPOUND,(size_t)vdatamem_size);
  
  if(gen_h5comptype(vdata_id,nfields,h4size,h4memsize,h5type,
		    h5memtype,h5_ctype,h5_cmemtype)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain hdf5 compound data type");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(vd_data);
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  /* this block was added while implementing netCDF4 hack,
   * but it seems that it was actually an unintentional bug fix.
   */
  if (1) {
    char *h5cvdata_type_name = malloc(strlen(h5cvdata_name) + 2 + 1);
    if (h5cvdata_type_name == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot allocate memory for type name");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    h5cvdata_name1 = (char *)malloc(strlen(h5cvdata_name)*sizeof(char));
    h5cvdata_name1 = correct_name2(h5cvdata_name);
    h5cvdata_name = correct_name(h5cvdata_name1);

    free(h5cvdata_name1);

    sprintf(h5cvdata_type_name, "%s_t", h5cvdata_name);

    if (H5Tcommit(h5_group, h5cvdata_type_name, h5_ctype) < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot commit hdf5 named type %s \n", h5cvdata_name);
	}
      free(h5cvdata_type_name);
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    free(h5cvdata_type_name);
  }

  h5_vddims[0] = n_records;
  h5_maxvddims[0] = H5S_UNLIMITED;
  h5d_sid = H5Screate_simple(1,h5_vddims,h5_maxvddims);
  if(h5d_sid <0){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create hdf5 data space");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(vd_data);
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  h5_chunkdims[0] = h5_vddims[0];
  if (h5_chunkdims[0] == 0)
    h5_chunkdims[0] = HE4HE5_DEFAULT_NUMELEM_IN_CHUNK;
  create_plist = H5Pcreate(H5P_DATASET_CREATE);
  if(create_plist < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("failed to create property list");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(vd_data);       
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    H5Sclose(h5d_sid);
    return CONVERT_FAIL;	
  }
  if (H5Pset_chunk(create_plist,1,h5_chunkdims)<0){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("failed to set up chunking property list");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(vd_data);       
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Gclose(h5_group);
    H5Pclose(create_plist);
    H5Sclose(h5d_sid);
    return CONVERT_FAIL;	
  }
  h5dset = H5Dcreate(h5_group,h5cvdata_name,h5_ctype,h5d_sid,create_plist);
  
  if(h5dset <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create HDF5 dataset");
      }
    free(h5memtype);
    free(h5type);
    free(h4memsize);
    free(h4size);
    free(vd_data);       
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Sclose(h5d_sid);
    H5Gclose(h5_group);
    H5Pclose(create_plist);
    return CONVERT_FAIL;
  }

  if (!empty_vdata) {
    if(H5Dwrite(h5dset,h5_cmemtype,H5S_ALL,H5S_ALL,H5P_DEFAULT,vd_data)<0){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot write HDF5 dataset converted from Vdata");
	}
      free(h5memtype);
      free(h5type);
      free(h4memsize);
      free(h4size);
      free(vd_data);
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Sclose(h5d_sid);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      H5Pclose(create_plist);
      return CONVERT_FAIL;
    }
  }

  free(h5memtype);
  free(h5type);
  free(h4memsize);
  free(h4size);
  free(vd_data);

  /* handle vdata attributes and vdata field attributes. */

  num_vd_attrs = VSfnattrs(vdata_id,_HDF_VDATA);

  if (num_vd_attrs == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain number of attributes of Vdata");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Sclose(h5d_sid);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
    
  /* when field_index = -1, only transfer vdata attribute.*/

  field_index = -1;
  if(attr_flag <0 || attr_flag >2) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("attribute flag is set wrong");
      }
    free(h5cvdata_name);
    free(vdata_name);
    free(fieldname);
    free(vdata_class);
    free(field_name_list);
    H5Sclose(h5d_sid);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;	
  }

  if(attr_flag == H425_ALLATTRS || attr_flag == H425_NONEWATTRS) {
    if(vdata_transattrs(vdata_id,h5dset,num_vd_attrs,field_index,
			NULL)==FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert vdata attributes");
	}
      free(h5cvdata_name);
      free(vdata_name);
      free(fieldname);
      free(vdata_class);
      free(field_name_list);
      H5Sclose(h5d_sid);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }

    /* The following three attributes are required by HDF5 table specification,
       They HDF5 table title, HDF5 table class and HDF5 table version.*/
    if(vdata_name[0] != '\0') {
      
      
      if(he4_transpredattrs(h5dset,HDF5_TABLE_TITLE,vdata_name)==FAIL){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to convert Vdata name to HDF5 table title");
	  }
	free(h5cvdata_name);
	free(vdata_name);
	free(fieldname);
	free(vdata_class);
	free(field_name_list);
	H5Sclose(h5d_sid);
	H5Dclose(h5dset);
	H5Gclose(h5_group);
	return	CONVERT_FAIL;
      }
    }

     if(he4_transpredattrs(h5dset,HDF4_VDATA_CLASS,VDATA_CLASS)==FAIL){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to convert generate to HDF5 table class");
	  }
	free(h5cvdata_name);
	free(vdata_name);
	free(fieldname);
	free(vdata_class);
	free(field_name_list);
	H5Sclose(h5d_sid);
	H5Dclose(h5dset);
	H5Gclose(h5_group);
	return CONVERT_FAIL;
      }
     
     if(he4_transpredattrs(h5dset,HDF4_VDATA_VERSION_ATTR,HDF5TABLE_SPE_VERSION)==FAIL){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to generate HDF5 table version attribute");
	  }
	free(h5cvdata_name);
	free(vdata_name);
	free(fieldname);
	free(vdata_class);
	free(field_name_list);
	H5Sclose(h5d_sid);
	H5Dclose(h5dset);
	H5Gclose(h5_group);
	return CONVERT_FAIL;
     }

     for (i =0;i< nfields;i++) {
       
       if(conv_int_str((uint32)i,findex_str)==FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("cannot convert int to string");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
       h5tablefieldname=malloc(strlen(h5tablefield)+strlen(h5tablename)+
			       strlen(findex_str)+1);
       strcpy(h5tablefieldname,h5tablefield);
       strcat(h5tablefieldname,findex_str);
       strcat(h5tablefieldname,h5tablename);
       if(VFfieldname(vdata_id,i)== NULL) {
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("cannot obtain Vdata field name");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
      
       strcpy(fieldname,VFfieldname(vdata_id,i));
       
       if(he4_transpredattrs(h5dset,h5tablefieldname,fieldname)==FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("unable to convert generate to HDF5 table field name attribute");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
       
       free(h5tablefieldname);
       num_vd_field_attrs = VSfnattrs(vdata_id,i);
       if(num_vd_field_attrs == CONVERT_FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("cannot obtain number of Vdata field attribute");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
       
       if(vdata_transattrs(vdata_id,h5dset,num_vd_field_attrs,i,
			   fieldname)==FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("cannot convert Vdata field attribute");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
     }
    
     /* converting annotations of vdata into corresponding hdf5 attribute.*/
     /* 
     if(He4toHe5anno_obj_all_labels(h5groupname,h5cvdata_name,(uint16)vdata_ref,
				  vdata_tag)==FAIL){
       if (verboseModeGlobal == CONVERT_TRUE)
	 {
	   printf("Vdata annotation object label conversion is wrong ");
	 }
       free(h5cvdata_name);
       H5Sclose(h5d_sid);
       H5Dclose(h5dset);
       H5Gclose(h5_group);
       return CONVERT_FAIL;
     }
     
     if(He4toHe5anno_obj_all_descs(h5groupname,h5cvdata_name,(uint16)vdata_ref,
				 vdata_tag)==FAIL){
       if (verboseModeGlobal == CONVERT_TRUE)
	 {
	   printf("Vdata annotation object description conversion is wrong ");
	 }
       free(h5cvdata_name);
       H5Sclose(h5d_sid);
       H5Dclose(h5dset);
       H5Gclose(h5_group);
       return CONVERT_FAIL;
     }
     */
     /* converting predefined attributes. */
     
     if(attr_flag == H425_ALLATTRS) {
       strcpy(vdlabel,VDATALABEL);
       if(he4_transpredattrs(h5dset,HDF4_OBJECT_TYPE,vdlabel)==FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("unable to generate HDF4 OBJECT TYPE attribute");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
       
       
       if(vdata_name[0] != '\0') {
	 if(he4_transpredattrs(h5dset,HDF4_OBJECT_NAME,vdata_name)==FAIL){
	   if (verboseModeGlobal == CONVERT_TRUE)
	     {
	       printf("unable to convert Vdata name to an attribute");
	     }
	   free(h5cvdata_name);
	   free(vdata_name);
	   free(fieldname);
	   free(vdata_class);
	   free(field_name_list);
	   H5Sclose(h5d_sid);
	   H5Dclose(h5dset);
	   H5Gclose(h5_group);
	   return CONVERT_FAIL;
	 }
       }
       
       if(h4_transnumattr(h5dset,HDF4_REF_NUM,(uint16)vdata_ref)==FAIL){
	 if (verboseModeGlobal == CONVERT_TRUE)
	   {
	     printf("unable to convert Vdata reference to an attribute");
	   }
	 free(h5cvdata_name);
	 free(vdata_name);
	 free(fieldname);
	 free(vdata_class);
	 free(field_name_list);
	 H5Sclose(h5d_sid);
	 H5Dclose(h5dset);
	 H5Gclose(h5_group);
	 return CONVERT_FAIL;
       }
     }
  }

  if(vdata_name != NULL)
    {
      free(vdata_name);
      vdata_name = NULL;
    }
  if(fieldname != NULL)
    {
      free(fieldname);
      fieldname = NULL;
    }
  if(vdata_class != NULL)
    {
      free(vdata_class);
      vdata_class = NULL;
    }
  if(field_name_list != NULL)
    {
      free(field_name_list);
      field_name_list = NULL;
    }

  ret = H5Sclose(h5d_sid);
  if(ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 data space");
      }
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  ret = H5Dclose(h5dset);
  if(ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 dataset interface ");
      }
    free(h5cvdata_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  ret = H5Gclose(h5_group);
  if(ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 group interface ");
      }
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  free(h5cvdata_name);
  return CONVERT_SUCCESS;
}
   
/*-------------------------------------------------------------------------
 * Function:	He4toHe5vdata_attr_name
 *
 * Purpose:     translate one Vdata attribute into an attribute of the
                corresponding hdf5 dataset 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id:        vdata identifier
                h5groupfullpath: HDF5 group absolute path(must be provided)
		h5dsetname:      hdf5 dataset(can be set to NULL.If set to NULL,
		                 HDF4 default vdata name will be used.)
		attr_name:       name of vata attribute
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */	

int   He4toHe5vdata_attr_name(
			  int32 vdata_id,
			  char* h5groupfullpath,
			  char* h5dsetname,
			  char*   attr_name)
{
   
  int attr_index; /* index of H4toH5 vdata attribute */

  if((attr_index=VSfindattr(vdata_id,_HDF_VDATA,attr_name))==FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate memory for vdata name");
      }
    return CONVERT_FAIL;
  }
  if(He4toHe5vdata_attr_index(vdata_id,h5groupfullpath,h5dsetname,attr_index)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot convert vdata attribute");
      }
    return CONVERT_FAIL;
  }
  return CONVERT_SUCCESS;
}
      
/*-------------------------------------------------------------------------
 * Function:	He4toHe5vdata_attr_index
 *
 * Purpose:     translate one Vdata attribute into an attribute of the
                corresponding hdf5 dataset 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id: vdata identifier
                h5groupfullpath: HDF5 group absolute path(must be provided)
		h5dsetname: hdf5 dataset(can be set to NULL.If set to NULL,
		            HDF4 default vdata name will be used.)
		attr_index: index of vata attribute
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */	

int  He4toHe5vdata_attr_index(
			  int32 vdata_id,
			  char* h5groupfullpath,
			  char* h5dsetname,
			  int   attr_index)
{
   
  char      vdata_name[HE4HE5_MAX_NC_NAME]; /* Vdata name */
  char*     cor_vdata_name;
  char      svdattr_name[2*HE4HE5_MAX_NC_NAME]; /* Vdata attribute name */
  char*     svdrepattr_name; /* temporary string to store Vdata attribute name */
  char      refstr[MAXREF_LENGTH]; /* reference number of Vdata interface with
					string format */
  char*     h5crepvdata_name; /* temporary string to store the absolute path of
				 HDF5 dataset name converted from Vdata */
  char*     h5cvdata_name; /* The absolute path of HDF5 dataset name converted
			      from Vdata */

  int32     count_svdadata; /* number of elements of the Vdata attribute */
  uint32    temp_countadata; /* temporary count of the number of elements of
				Vdata attribute */
  int32     svd_atype; /* Vdata attribute type */
  int       istat; /* check the "return" status  of HDF4 APIs */
  int32     vdata_ref; /* Vdata reference number */

  size_t    sh4_amemsize; /* The memory size of Vdata attribute data type */
  size_t    sh4_asize; /* The file size of Vdata attribute data type */

  hid_t     sh5a_sid; /* The data space id of HDF5 attribute */
  hid_t     sh5a_id; /* The HDF5 attribute id */
  hid_t     sh5_atype; /* The file data type id of HDF5 attribute */
  hid_t     sh5_amemtype;/* The memory data type id of HDF5 attribute */
  hid_t     sh5str_type; /* The file data type id of HDF5 attribute with string
			    format */
  hid_t     sh5str_memtype; /* The memory data type id of HDF5 attribute with 
			       string format */

  hid_t     h5_group; /* The HDF5 group id */
  hid_t     h5dset; /* The HDF5 dataset id */
  hsize_t   sh5dims[1]; /* The size of the dimension of the HDF5 attribute
			     converted from Vdata attribute */
  void*     svd_adata; /* Vdata buffer */
  herr_t    sret; /* the flag to check the "return" status of HDF5 APIs. */
  char* cor_h5dsetname;
  int check_name=0;

  /* zeroing out memory for svdattr_name and refstr */                         
  ZeroMemory(svdattr_name,2*HE4HE5_MAX_NC_NAME); 
  ZeroMemory(refstr,MAXREF_LENGTH);          

   /* The following block obtains group name and dataset name. */
  /* obtain h5_group id */
  h5_group    =  get_h5groupid(h5groupfullpath);
  if(h5_group < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain HDF5 group ID");
      }
    return CONVERT_FAIL;
  }
  
  /*obtain default vdata name */
  istat = VSQueryname(vdata_id,vdata_name);
  if(istat == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain the Vdata name");
      }
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  vdata_ref = VSQueryref(vdata_id);
  if(vdata_ref < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain the reference number of Vdata");
      }
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(h5dsetname == NULL) {
    /* change format of reference number. */
    if(conv_int_str((uint32)vdata_ref,refstr)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert reference number into string type");
	}
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    h5crepvdata_name=make_objname_no(refstr,h5groupfullpath,
				     HDF4_VDATA);
    
    if(check_name/*=lookup_name(h5crepvdata_name,VD_HASHSIZE,
		   temph4toh5id->name_hashtab)*/==1){
      h5cvdata_name = malloc(strlen(h5crepvdata_name)+1);
      if(h5cvdata_name == NULL) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot allocate memory for vdata name");
	  }
	H5Gclose(h5_group);
	return CONVERT_FAIL;
      }
      strncpy(h5cvdata_name,h5crepvdata_name,strlen(h5crepvdata_name));
      free(h5crepvdata_name);
    }
    else {
      cor_vdata_name=correct_name(vdata_name);
      if(cor_vdata_name== NULL) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot allocate memory for corrected vdata name");
	  }
	H5Gclose(h5_group);
	return CONVERT_FAIL;
      }
      h5cvdata_name = make_objname_yes(cor_vdata_name,h5groupfullpath);
      free(cor_vdata_name);
    }	
  }
  else {
    
    cor_h5dsetname= correct_name(h5dsetname);
    if(cor_h5dsetname== NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot allocate memory for corrected vdata name");
	}
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    h5cvdata_name = make_objname_yes(cor_h5dsetname,h5groupfullpath);
    free(cor_h5dsetname);
  }
  h5dset = H5Dopen(h5_group,h5cvdata_name);
  if(h5dset < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot open HDF5 data set");
      }
    H5Gclose(h5_group);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
   /* if the field_index is -1, no field attribute exists, only
     VDATA attributes are converted.*/
  if (VSattrinfo(vdata_id,_HDF_VDATA,attr_index,svdattr_name,&svd_atype,
		 &count_svdadata,NULL)== CONVERT_FAIL){  
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata attribute information");
      }
    H5Gclose(h5_group);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
  if(svdattr_name[0] == '\0') {
    svdrepattr_name = trans_obj_name(DFTAG_VG,attr_index);
    strcpy(svdattr_name,svdrepattr_name);
    free(svdrepattr_name);
  }
  
  /* converting attribute data type into the corresponding HDF5 data type */
  
  if(h4type_to_h5type(svd_atype,&sh5_amemtype,&sh4_amemsize,
		      &sh4_asize,&sh5_atype)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to convert HDF4 data type into HDF5 data type");
      }
    H5Gclose(h5_group);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
  svd_adata = malloc(sh4_amemsize * count_svdadata);
  if(svd_adata == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to allocate memory for vdata attribute data");
      }
    H5Gclose(h5_group);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
  if(VSgetattr(vdata_id,_HDF_VDATA,attr_index,(VOIDP)svd_adata)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata attribute");
      }
    H5Gclose(h5_group);
    free(svd_adata);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
  /* now do attribute-transferring:
     1. deal with string data type
     2. set attribute space
     3. get attribute name  */
  
  if (sh5_atype == H5T_STRING) {
    
    temp_countadata = count_svdadata*sh4_asize;
    H425_CHECK_OVERFLOW(temp_countadata,uint32,size_t);
    if ((sh5str_type = mkstr((size_t)count_svdadata,
			     H5T_STR_SPACEPAD))<0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to make HDF5 string data type");
	}
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
    if ((sh5str_memtype = mkstr((size_t)(count_svdadata*sh4_amemsize),
				H5T_STR_SPACEPAD))<0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to make HDF5 string data type for memory");
	}
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
    sh5a_sid = H5Screate(H5S_SCALAR);
    
    if (sh5a_sid < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to create HDF5 data space");
	}
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
    sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5str_type,
			     sh5a_sid,H5P_DEFAULT);
    if (sh5a_id <0) {
      /*if (transattrs_split(h5dset, svdattr_name, sh5a_sid, svd_adata, count_svdadata) < 0) {*/
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to CREATE HDF5 attribute");
	  }
	H5Sclose(sh5a_sid);
	H5Gclose(h5_group);
	free(h5cvdata_name);
	free(svd_adata);
	return CONVERT_FAIL;
    }
    else if (H5Awrite(sh5a_id,sh5str_memtype,(void *)svd_adata) < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to write attribute");
	}
      H5Sclose(sh5a_sid);
      H5Aclose(sh5a_id);
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
  }
  
  else {
    
    if(count_svdadata == 1) {
      sh5a_sid = H5Screate(H5S_SCALAR);
      
      if (sh5a_sid < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 space id for attribute");
	  }
	H5Gclose(h5_group);
	free(h5cvdata_name);
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    else {
      sh5dims[0] = count_svdadata;
      sh5a_sid =  H5Screate_simple(1,sh5dims,NULL);
      
      if (sh5a_sid < 0)  {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 space id for attribute");
	  }
	H5Gclose(h5_group);
	free(h5cvdata_name);
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    
    sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5_atype,
			     sh5a_sid,H5P_DEFAULT);
    
    if(sh5a_id <0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to create HDF5 attribute ID");
	}
      H5Sclose(sh5a_sid);
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
    sret = H5Awrite(sh5a_id,sh5_amemtype,(void *)svd_adata);
    
    if(sret <0) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to write HDF5 attribute data");
	}
      H5Sclose(sh5a_sid);
      H5Aclose(sh5a_id);
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
  }
  free(h5cvdata_name);
  free(svd_adata);
  sret = H5Sclose(sh5a_sid);
  if( sret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 attribute data space");
      }
    H5Aclose(sh5a_id);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  if (sh5a_id >= 0) {
    sret = H5Aclose(sh5a_id);
    if( sret < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to close HDF5 attribute interface");
	}
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }
  sret = H5Dclose(h5dset);
  if(sret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 dataset");
      }
    H5Gclose(h5_group);
  }
  sret = H5Gclose(h5_group);
  if(sret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 attribute interface");
      }
    return CONVERT_FAIL;
  }
  return CONVERT_SUCCESS;
}


/*-------------------------------------------------------------------------
 * Function:	He4toHe5vdata_field_attr_index
 *
 * Purpose:     translate one Vdata field attribute into an attribute of the
                corresponding hdf5 dataset 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id: vdata identifier
                h5groupfullpath: HDF5 group absolute path(must be provided)
		h5dsetname: hdf5 dataset(can be set to NULL.If set to NULL,
		            HDF4 default vdata name will be used.)
		field_index: index of vata field
                field_attrindex: index of vdata field attribute
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */	

int  He4toHe5vdata_field_attr_index(
			       int32 vdata_id,
			       char* h5groupfullpath,
			       char* h5dsetname,
			       int   field_index,
			       int   field_attrindex)
{
   
  char      vdata_name[HE4HE5_MAX_NC_NAME]; /* Vdata name */
  char*     cor_vdata_name;
  char      svdattr_name[2*HE4HE5_MAX_NC_NAME]; /* Vdata attribute name */
  char      fieldname[HE4HE5_MAX_NC_NAME]; /* Vdata field name */
  char      refstr[MAXREF_LENGTH]; /* reference number of Vdata interface with
					string format */
  char*     h5crepvdata_name; /* temporary string to store the absolute path of
				 HDF5 dataset name converted from Vdata */
  char*     h5cvdata_name; /* The absolute path of HDF5 dataset name converted
			      from Vdata */

  int32     count_svdadata; /* number of elements of the Vdata attribute */
  uint32    temp_countadata; /* temporary count of the number of elements of
				Vdata attribute */
  int32     svd_atype;  /* Vdata attribute type */
  int       istat;  /* check the "return" status  of HDF4 APIs */
  int32     vdata_ref;  /* Vdata reference number */

  size_t    sh4_amemsize; /* The memory size of Vdata attribute data type */
  size_t    sh4_asize; /* The file size of Vdata attribute data type */

  hid_t     sh5a_sid; /* The data space id of HDF5 attribute */
  hid_t     sh5a_id; /* The HDF5 attribute id */
  hid_t     sh5_atype; /* The file data type id of HDF5 attribute */
  hid_t     sh5_amemtype; /* The memory data type id of HDF5 attribute */
  hid_t     sh5str_type; /* The file data type id of HDF5 attribute with string
			    format */
  hid_t     sh5str_memtype; /* The memory data type id of HDF5 attribute with 
			       string format */

  hid_t     h5_group; /* The HDF5 group id */
  hid_t     h5dset; /* The HDF5 dataset id */
  hsize_t   sh5dims[1]; /* The size of the dimension of the HDF5 attribute
			     converted from Vdata attribute */
  void*     svd_adata; /* Vdata buffer */
  herr_t    sret; /* the flag to check the "return" status of HDF5 APIs. */
  char*    cor_h5dsetname;
  int check_name=0;

  /* zeroing out memory for svdattr_name and refstr */
  ZeroMemory(svdattr_name,2*HE4HE5_MAX_NC_NAME); 
  ZeroMemory(refstr,MAXREF_LENGTH);          
  
  /* The following block obtains group name and dataset name. */
  /* obtain h5_group id */
  h5_group    =  get_h5groupid(h5groupfullpath);
  if(h5_group < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain HDF5 group ID");
      }
    return CONVERT_FAIL;
  }
  
  /*obtain default vdata name */
  istat = VSQueryname(vdata_id,vdata_name);
  if(istat == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain the Vdata name");
      }
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  vdata_ref = VSQueryref(vdata_id);
  if(vdata_ref < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain the reference number of Vdata");
      }
    H5Gclose(h5_group);
   return  CONVERT_FAIL;
  }
  
  if(h5dsetname == NULL) {
    /* change format of reference number. */
    if(conv_int_str((uint32)vdata_ref,refstr)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert reference number into string type");
	}
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    h5crepvdata_name=make_objname_no(refstr,h5groupfullpath,
				     HDF4_VDATA);
    
    if(check_name/*=lookup_name(h5crepvdata_name,VD_HASHSIZE,
		   temph4toh5id->name_hashtab)*/==1){
      h5cvdata_name = malloc(strlen(h5crepvdata_name)+1);
      if(h5cvdata_name == NULL) { 
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot allocate memory for vdata name");
	  }
	H5Gclose(h5_group);
	return CONVERT_FAIL;
      }
      strncpy(h5cvdata_name,h5crepvdata_name,strlen(h5crepvdata_name));
      free(h5crepvdata_name);
    }
    else {
      cor_vdata_name= correct_name(vdata_name);
      if(cor_vdata_name == NULL) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("cannot allocate memory for corrected vdata name");
	  }
	H5Gclose(h5_group);
	return CONVERT_FAIL;
      }
      h5cvdata_name = make_objname_yes(cor_vdata_name,h5groupfullpath);
      free(cor_vdata_name);
    }
  }
  else {
    cor_h5dsetname = correct_name(h5dsetname);
    if(cor_h5dsetname == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot allocate memory for corrected vdata name");
	}
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    h5cvdata_name = make_objname_yes(cor_h5dsetname,h5groupfullpath);
    free(cor_h5dsetname);
  }	
  h5dset = H5Dopen(h5_group,h5cvdata_name);
  if(h5dset < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot open HDF5 data set");
      }
    H5Gclose(h5_group);
    free(h5cvdata_name);
    return CONVERT_FAIL;
  }
  
  if(VFfieldname(vdata_id,field_index)== NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot obtain Vdata field name");
      }
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  strcpy(fieldname,VFfieldname(vdata_id,field_index));
  
  /* if the field_index is -1, no field attribute exists, only
     VDATA attributes are converted.*/
  if (VSattrinfo(vdata_id,field_index,field_attrindex,svdattr_name,&svd_atype,
		 &count_svdadata,NULL)== CONVERT_FAIL){  
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata attribute information");
      }
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  if (field_index != -1 && svdattr_name != NULL) {
    strcat(svdattr_name,":");
    strcat(svdattr_name,fieldname);
  }
  
  else {
    strcat(svdattr_name,":");
    strcat(svdattr_name,"HDF4_VDATA_ATTR_");
    if(conv_int_str((uint32)field_index,refstr)==FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to convert vdata field index to string");
	}
      free(h5cvdata_name);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    strcat(svdattr_name,refstr);
    
  }
  
  /* converting attribute data type into the corresponding HDF5 data type */
  
  if(h4type_to_h5type(svd_atype,&sh5_amemtype,&sh4_amemsize,
		      &sh4_asize,&sh5_atype)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to convert HDF4 data type into HDF5 data type");
      }
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  svd_adata = malloc(sh4_amemsize * count_svdadata);
  if(svd_adata == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to allocate memory for vdata attribute data");
      }
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if(VSgetattr(vdata_id,field_index,field_attrindex,(VOIDP)svd_adata)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to obtain Vdata attribute");
      }
    free(svd_adata);
    free(h5cvdata_name);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  /* now do attribute-transferring:
     1. deal with string data type
     2. set attribute space
     3. get attribute name  */
  
  if (sh5_atype == H5T_STRING) {
    
    temp_countadata = count_svdadata*sh4_asize;
    H425_CHECK_OVERFLOW(temp_countadata,uint32,size_t);
    if ((sh5str_type = mkstr((size_t)temp_countadata,
			     H5T_STR_SPACEPAD))<0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to make HDF5 string data type");
	}
      free(svd_adata);
      free(h5cvdata_name);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    
    if ((sh5str_memtype = mkstr(count_svdadata*sh4_amemsize,
				H5T_STR_SPACEPAD))<0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to make HDF5 string data type for memory");
	}
      free(svd_adata);
      free(h5cvdata_name);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    
    sh5a_sid = H5Screate(H5S_SCALAR);
    
    if (sh5a_sid < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to create HDF5 data space");
	}
      free(svd_adata);
      free(h5cvdata_name);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
    
    sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5str_type,
			     sh5a_sid,H5P_DEFAULT);
    if (sh5a_id <0) {
      /*if (transattrs_split(h5dset, svdattr_name, sh5a_sid, svd_adata, count_svdadata) < 0) {*/
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to CREATE HDF5 attribute");
	  }
	H5Sclose(sh5a_sid);
	free(svd_adata);
	free(h5cvdata_name);
	H5Dclose(h5dset);
	H5Gclose(h5_group);
	return CONVERT_FAIL;
    }
    else if (H5Awrite(sh5a_id,sh5str_memtype,(void *)svd_adata) < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to write attribute");
	}
      H5Sclose(sh5a_sid);
      H5Aclose(sh5a_id);
      free(h5cvdata_name);
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      free(svd_adata);
      return CONVERT_FAIL;
    }
  }
  
  else {
    
    if(count_svdadata == 1) {
      sh5a_sid = H5Screate(H5S_SCALAR);
      
      if (sh5a_sid < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 space id for attribute");
	  }
	H5Gclose(h5_group);
	free(h5cvdata_name);
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    else {
      sh5dims[0] = count_svdadata;
      sh5a_sid =  H5Screate_simple(1,sh5dims,NULL);
      
      if (sh5a_sid < 0)  {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 space id for attribute");
	  }
	H5Gclose(h5_group);
	free(h5cvdata_name);
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    
    sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5_atype,
			     sh5a_sid,H5P_DEFAULT);
    
    if(sh5a_id <0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to create HDF5 attribute ID");
	}
      H5Sclose(sh5a_sid);
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
    sret = H5Awrite(sh5a_id,sh5_amemtype,(void *)svd_adata);
    
    if(sret <0) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to write HDF5 attribute data");
	}
      H5Sclose(sh5a_sid);
      H5Aclose(sh5a_id);
      H5Gclose(h5_group);
      free(h5cvdata_name);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    
  }
  
  free(h5cvdata_name);
  free(svd_adata);
  sret = H5Sclose(sh5a_sid);
  if( sret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 attribute data space");
      }
    H5Aclose(sh5a_id);
    H5Dclose(h5dset);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  
  if (sh5a_id >= 0) {
    sret = H5Aclose(sh5a_id);
    if( sret < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to close HDF5 attribute interface");
	}
      H5Dclose(h5dset);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }
  
  sret = H5Dclose(h5dset);
  if( sret <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 dataset interface");
      }  
    H5Gclose(h5_group);
  }
  sret = H5Gclose(h5_group);
  if(sret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to close HDF5 attribute interface");
      }
    return CONVERT_FAIL;
  }
  return CONVERT_SUCCESS;
}


/*-------------------------------------------------------------------------
 * Function:	He4toHe5vdata_field_attr_name
 *
 * Purpose:     translate one Vdata field attribute into an attribute of the
                corresponding hdf5 dataset 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id: vdata identifier
                h5groupfullpath: HDF5 group absolute path(must be provided)
		h5dsetname: hdf5 dataset(can be set to NULL.If set to NULL,
		            HDF4 default vdata name will be used.)
		field_index: index of vata field
                field_attrindex: index of vdata field attribute
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */	

int  He4toHe5vdata_field_attr_name (
			       int32 vdata_id,
			       char* h5groupfullpath,
			       char* h5dsetname,
			       char* fieldname,
			       char* field_attrname)
{
  
  int32 field_attrindex; /* field attribute index */
  int32 field_index; /* field index */
  
  field_index = -1;

  
  /* obtain field and field attribute index */
  if(VSfindex(vdata_id,fieldname,&field_index)==FAIL){
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot retrieve field attribute index");
      }
    return CONVERT_FAIL;
  }

  field_attrindex = VSfindattr(vdata_id,field_index,field_attrname);
  if(field_attrindex == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot retrieve field attribute index");
      }
    return CONVERT_FAIL;
  }
     
  if(He4toHe5vdata_field_attr_index(vdata_id,h5groupfullpath,h5dsetname,field_index,field_attrindex) == CONVERT_FAIL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot convert VDATA field attribute");
      }
    return CONVERT_FAIL;
  }
  return CONVERT_SUCCESS;
}

/*-------------------------------------------------------------------------
 * Function:	vdata_transattrs
 *
 * Purpose:     translate Vdata attributes into attributes of the
                corresponding hdf5 dataset 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id: vdata identifier
		h5dset: hdf5 dataset
		snum_vdattrs: number of vd attributes
		field_index:  index of vdata fields
		attr_name: vdata(or vdata field) attribute name
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */
	

int  vdata_transattrs(
		      int32 vdata_id,
		      hid_t h5dset,
		      int snum_vdattrs,
		      int field_index,
		      char* attr_name)
{
   
  char      svdattr_name[2*HE4HE5_MAX_NC_NAME]; /* Vdata attribute name */
  char*     svdrepattr_name; /* temporary string to store Vdata attribute
				name */
  char      refstr[MAXREF_LENGTH]; /* reference number of Vdata interface with
					string format */
  int32     count_svdadata; /* number of elements of the Vdata attribute */
  uint32    temp_countadata; /* temporary count of the number of elements of
				Vdata attribute */
  int32     svd_atype; /* Vdata attribute type */

  size_t    sh4_amemsize; /* The memory size of Vdata attribute data type */
  size_t    sh4_asize; /* The file size of Vdata attribute data type */

  hid_t     sh5a_sid; /* The data space id of HDF5 attribute */
  hid_t     sh5a_id; /* The HDF5 attribute id */
  hid_t     sh5_atype; /* The file data type id of HDF5 attribute */
  hid_t     sh5_amemtype; /* The memory data type id of HDF5 attribute */
  hid_t     sh5str_type; /* The file data type id of HDF5 attribute with string
			    format */
  hid_t     sh5str_memtype;  /* The memory data type id of HDF5 attribute with 
			       string format */

  hsize_t   sh5dims[1]; /* The size of the dimension of the HDF5 attribute
			     converted from Vdata attribute */
  void*     svd_adata; /* Vdata buffer */
  herr_t    sret; /* the flag to check the "return" status of HDF5 APIs. */
  int       i;
  
  /* zeroing out memory for svdattr_name and refstr */                                  
  ZeroMemory(svdattr_name,2*HE4HE5_MAX_NC_NAME); 
  ZeroMemory(refstr,MAXREF_LENGTH);          

  /* separate vdata attribute from vdata field attributes. */

  if (field_index < -1) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf(
	       "Vdata field index should not be less than -1");
      }
    return CONVERT_FAIL;
  }

  for (i = 0;i < snum_vdattrs; i++) {

    /* if the field_index is 0, no field attribute exists, only
       VDATA attributes are converted.*/

    if (VSattrinfo(vdata_id,field_index,i,svdattr_name,&svd_atype,
		   &count_svdadata,NULL)== CONVERT_FAIL){  
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to obtain Vdata attribute information");
	}
      return CONVERT_FAIL;
    }
         
    if(svdattr_name[0] == '\0') {
      svdrepattr_name = trans_obj_name(DFTAG_VG,i);
      strcpy(svdattr_name,svdrepattr_name);
      free(svdrepattr_name);
    }

    if (field_index == -1);

    else if (field_index != -1 && attr_name != NULL) {
      strcat(svdattr_name,":");
      strcat(svdattr_name,attr_name);
    }

    else {
      strcat(svdattr_name,":");
      strcat(svdattr_name,"HDF4_VDATA_ATTR_");
      if(conv_int_str((uint32)field_index,refstr)==CONVERT_FAIL) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to convert vdata field index to string");
	  }
	return CONVERT_FAIL;
      }
      strcat(svdattr_name,refstr);
      
    }

    /* converting attribute data type into the corresponding HDF5 data type */

    if(h4type_to_h5type(svd_atype,&sh5_amemtype,&sh4_amemsize,
			&sh4_asize,&sh5_atype)==CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to convert HDF4 data type into HDF5 data type");
	}
      return CONVERT_FAIL;
    }

    svd_adata = malloc(sh4_amemsize * count_svdadata);

    if(svd_adata == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to allocate memory for vdata attribute data");
	}
      return CONVERT_FAIL;
    }

    if(VSgetattr(vdata_id,field_index,i,(VOIDP)svd_adata)==FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to obtain Vdata attribute");
	}
      free(svd_adata);
      return CONVERT_FAIL;
    }
	
    /* now do attribute-transferring:
       1. deal with string data type
       2. set attribute space
       3. get attribute name  */
           
    if (sh5_atype == H5T_STRING) {

       temp_countadata = count_svdadata*sh4_asize;
      H425_CHECK_OVERFLOW(temp_countadata,uint32,size_t);
      if ((sh5str_type = mkstr((size_t)temp_countadata,
			       H5T_STR_SPACEPAD))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to make HDF5 string data type");
	  }
	free(svd_adata);
	return CONVERT_FAIL;
      }

      if ((sh5str_memtype = mkstr(count_svdadata*sh4_amemsize,
				  H5T_STR_SPACEPAD))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to make HDF5 string data type for memory");
	  }
	free(svd_adata);
	return CONVERT_FAIL;
      }

      sh5a_sid = H5Screate(H5S_SCALAR);

      if (sh5a_sid < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 data space");
	  }
	free(svd_adata);
	return CONVERT_FAIL;
      }

      sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5str_type,
			  sh5a_sid,H5P_DEFAULT);
      if (sh5a_id <0) {
	/*if (transattrs_split(h5dset, svdattr_name, sh5a_sid, svd_adata, count_svdadata) < 0) {*/
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("unable to CREATE HDF5 attribute");
	    }
	  H5Sclose(sh5a_sid);
	  free(svd_adata);
	  return CONVERT_FAIL;
      }
      else if (H5Awrite(sh5a_id,sh5str_memtype,(void *)svd_adata) < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to write attribute");
	  }
	H5Sclose(sh5a_sid);
	H5Aclose(sh5a_id);
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    
    else {
      /* netCDF4 cannot read numeric values of scalar space.
       * Force to use Simple type even if count_svdadata == 1
       */
      if(count_svdadata == 1) {
        sh5a_sid = H5Screate(H5S_SCALAR);

        if (sh5a_sid < 0) {
	  if (verboseModeGlobal == CONVERT_TRUE)
	      {
		printf("unable to create HDF5 space id for attribute");
	      }
	  free(svd_adata);
	  return CONVERT_FAIL;
	}
      }
      else {
        sh5dims[0] = count_svdadata;
        sh5a_sid =  H5Screate_simple(1,sh5dims,NULL);
          
        if (sh5a_sid < 0)  {
	  if (verboseModeGlobal == CONVERT_TRUE)
	    {
	      printf("unable to create HDF5 space id for attribute");
	    }
	  free(svd_adata);
	  return CONVERT_FAIL;
	}
      }

      sh5a_id = H5Acreate_safe(h5dset,svdattr_name,sh5_atype,
			  sh5a_sid,H5P_DEFAULT);
    
      if(sh5a_id <0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 attribute ID");
	  }
	H5Sclose(sh5a_sid);
	free(svd_adata);
	return CONVERT_FAIL;
      }

      sret = H5Awrite(sh5a_id,sh5_amemtype,(void *)svd_adata);

      if(sret <0) {

	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to write HDF5 attribute data");
	  }
	H5Sclose(sh5a_sid);
	H5Aclose(sh5a_id);
	free(svd_adata);
	return CONVERT_FAIL;
      }

     
    }
    sret = H5Sclose(sh5a_sid);
    if(sret < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to close HDF5 data space");
	}
      H5Aclose(sh5a_id);
      free(svd_adata);
      return CONVERT_FAIL;
    }
    if (sh5a_id >= 0) {
      sret = H5Aclose(sh5a_id);
      if(sret < 0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to close HDF5 attribute id");
	  }
	free(svd_adata);
	return CONVERT_FAIL;
      }
    }
    free(svd_adata);
  }
  return CONVERT_SUCCESS;
}
/*-------------------------------------------------------------------------
 * Function:	gen_h5comptype
 *
 * Purpose:     generate hdf5 compound data type
                 
 *              
 * Return:	FAIL if failed, CONVERT_SUCCESS if successful.
 *
 * In :	        
                vdata_id: vdata identifier
		nfields: number of fields
		sh4size: pointer to datatype size in memory
		sh4memsize: pointer to datatype size in memory
		sh5type: pointer to hdf5 datatype
		sh5memtype: pointer to actual hdf5 datatype in memory
		h5_ctype:  hdf5 compound datatype
		h5_cmemtype: hdf5 compound datatype in memory
   Out:
   Modifications:

 *-------------------------------------------------------------------------
 */	

int gen_h5comptype(
		   int32 vdata_id,
		   int32 nfields,
		   size_t* sh4size,
		   size_t* sh4memsize,
		   hid_t* sh5type,
		   hid_t* sh5memtype,
		   hid_t h5_ctype,
		   hid_t h5_cmemtype)
{

  char*   fieldname; /* Vdata field name */
  int32   fieldorder; /* Vdata field order */
  size_t  fil_offset; /* offset of atomic data type in a file compound data type */
  size_t  mem_offset; /* offset of atomic data type in a memory compound data type */
  hsize_t fielddim[1]; /* size of the dimension of the field */
  hid_t   h5str_type; /* HDF5 string data type */
  int     check_ifstr;/* flag to check if the h5 type is string.*/
  int     i;
 
  check_ifstr    = 0;
  fil_offset     = 0;
  mem_offset     = 0;

  for (i =0;i< nfields;i++) {
   
    fieldname = NULL;
    fieldorder = VFfieldorder(vdata_id,i);
   
    if(fieldorder == CONVERT_FAIL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to obtain Vdata field order");
	}
      return CONVERT_FAIL;
    }

    fieldname = VFfieldname(vdata_id,i);
    if(fieldname == NULL){
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("unable to obtain Vdata field name"); 
	}
      return CONVERT_FAIL;
    }
   

    /* when vdata is a character array, we will write the whole
       array as one hdf5 type string. */

    if(sh5type[i] == H5T_STRING) {

      if ((h5str_type = mkstr(sh4size[i]*fieldorder,H5T_STR_SPACEPAD))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to make HDF5 string");
	  }
	return CONVERT_FAIL;
      }
      sh5type[i] = h5str_type;
      check_ifstr  = 1;
    }

    if (sh5memtype[i] == H5T_STRING) {

      if((h5str_type = mkstr(sh4memsize[i]*fieldorder,H5T_STR_SPACEPAD))<0){ 
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to make HDF5 string for memeory");
	  }
	return CONVERT_FAIL;
      }
      sh5memtype[i] = h5str_type;

    }

    fielddim[0] = fieldorder;
        
    /* if field type is an array, use H5Tinsert_array. 
       When the data type is H5T_STRING, 
       we will treat the the vdata  as a HDF5 scalar type.*/

    if (fielddim[0] == 1 || check_ifstr == 1) {

      if(H5Tinsert(h5_ctype,fieldname,fil_offset,sh5type[i])<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to insert HDF5 compound data type");
	  }
	return CONVERT_FAIL;
      }

      if(H5Tinsert(h5_cmemtype,fieldname,mem_offset,sh5memtype[i])<0){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to insert HDF5 memory compound data type");
	  }
	return CONVERT_FAIL;
      }
    }
      
    else {
      hid_t arr_type;    /* Array datatype for inserting fields */

      /* Create array datatype */
      if((arr_type=H5Tarray_create(sh5type[i],1,fielddim,NULL))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 array data type");
	  }
	return CONVERT_FAIL;
      }

      if(H5Tinsert(h5_ctype,fieldname,fil_offset,arr_type)<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to insert HDF5 array data type");
	  }
	return CONVERT_FAIL;
      }

      /* Close array datatype */
      if(H5Tclose(arr_type)<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to close HDF5 data type");
	  }
	return CONVERT_FAIL;
      }

      /* Create array datatype */
      if((arr_type=H5Tarray_create(sh5memtype[i],1,fielddim,NULL))<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to create HDF5 array data type");
	  }
	return CONVERT_FAIL;
      }

      if(H5Tinsert(h5_cmemtype,fieldname,mem_offset,arr_type)<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to insert HDF5 array data type");
	  }
	return CONVERT_FAIL;
      }

      /* Close array datatype */
      if(H5Tclose(arr_type)<0) {
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("unable to close HDF5 array data type");
	  }
	return CONVERT_FAIL;

      }
    }

    if( check_ifstr == 1) {
      fil_offset = fil_offset + sh4size[i]*fieldorder;
      mem_offset = mem_offset + sh4memsize[i]*fieldorder;
      check_ifstr = 0;
    }
    else { 
      fil_offset = fil_offset + sh4size[i]*fieldorder;
      mem_offset = mem_offset + sh4memsize[i]*fieldorder;
    }

  }

  return CONVERT_SUCCESS;
}

/*************************************************************************

MODULE: ConvertLoneSDSes

PURPOSE: 
    This module converts all SDSes and their attributes that are not
    written by hdfeos2. Basically these are SDSs that are under root group in 
    hdf4 file are created using HDF4 calls
 

RETURN VALUE:

Value    	Description
-----		-----------
CONVERT_SUCCESS	Returns CONVERT_SUCCESS if successful.
                Returns CONVERT_FAIL otherwise

HISTORY:
Version    Date     Programmer      Code     Reason
-------    ----     ----------      ----     ------
           04/12    Abe Taaheri              

NOTES:

**************************************************************************/

int ConvertLoneSDSes(char *inFileName, char *outFileName)
{
  int        retVal = CONVERT_SUCCESS;           /* return value */
  int32      sd_id;
  int        outFid;
  int32      sds_id; /* SDS id */
  int32      sds_ref; /* SDS reference number */
  int32      num_sds; /* number of SDS objects */
  int32      num_glsdsattrs; /* number of global SD interface attributes */
  char       h5groupname[19];
  char       sdsname[HE4HE5_MAX_NC_NAME];/* SDS name */
  hid_t      HDFfid;
  int        i, sdsCheckLoop;
  int32      sds_rank; /* rank of SDS */
  int32      sds_dimsizes[HE4HE5_MAX_VAR_DIMS]; /* array to describe dimensional size of SDS */
  int32      sds_dtype; /* SDS data type */
  int        check_sds;
  int32      num_sdsattrs; /* number of SDS attributes */

  strcpy(h5groupname,"HDFEOS INFORMATION");

  sd_id = SDstart(inFileName, DFACC_RDWR);
  outFid = (int) H5Fopen(outFileName,H5F_ACC_RDWR,H5P_DEFAULT);
  if (outFid == CONVERT_FAIL) 
    {
      SDend(sd_id);
      retVal = CONVERT_FAIL; 
      printf("ERROR: Unable to open output file %s\n",outFileName);
      return retVal;
    }

  /* transfer HDF4 written SDS and attributes from input HDF file to output file */
  /* obtain number of SD attributes. */
  num_sds = 0;
  num_glsdsattrs = 0;
  if(SDfileinfo(sd_id,&num_sds,&num_glsdsattrs) == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("Cannot start SD interface for getting attribute information\n");
	}
      return CONVERT_FAIL;
  }

  printf("Total number of SDS Iincluding HDF-EOS fields) =%d\n",(int)num_sds);

  for (i=0;i<num_sds;i++){

    sds_id  = SDselect(sd_id,i);
    if (sds_id == CONVERT_FAIL ) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("error in obtaining sd id\n");
	  SDendaccess(sds_id );
	  SDend(sd_id);
	  H5Fclose((hid_t)outFid);
	  return CONVERT_FAIL;
	}
    }

    /* DO NOT convert SDS dimensional scale. Will be done elsewhere. */
    /*  if(SDiscoordvar(sds_id)) continue;*/

    /* using SDS "getinfo" API to get all the information of this SDS object.*/
    if (SDgetinfo(sds_id,sdsname,&sds_rank,sds_dimsizes,&sds_dtype,
		  &num_sdsattrs)==FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("failed to get SDS information\n");
	  SDendaccess(sds_id );
	  SDend(sd_id);
	  H5Fclose((hid_t)outFid);
	  return CONVERT_FAIL;
	}
    }

  /* Check to see if the name is swath/grid/point data/geo field */
    check_sds = 0;
    for(sdsCheckLoop = 0; sdsCheckLoop < hdfeos5Info.total_hdfeos_Used_SDS; sdsCheckLoop++)
      {
	if(strcmp(sdsname, hdfeos5Info.Used_SDS[sdsCheckLoop]) == 0) check_sds = 1;
      }

    if(check_sds == 1 && !(SDiscoordvar(sds_id))) continue; /* skip conversion for this sds */

    /* convert the HDF written dimension scale SDS objects */
    if(SDiscoordvar(sds_id)) {

      if(CopyDimscalesdsToHe5(outFileName, sds_id) == CONVERT_FAIL){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
 	    printf("error in converting SDS objects %s.\n",sdsname);
	    SDendaccess(sds_id );
	    SDend(sd_id);
	    H5Fclose((hid_t)outFid);
	    return CONVERT_FAIL;
	  }
      }
    }
    else {
      /* convert the "lone" SDS objects that are not dimension scales*/
      if(CopysdsToHe5((hid_t)outFid, sds_id, h5groupname) == CONVERT_FAIL){
	if (verboseModeGlobal == CONVERT_TRUE)
	  {
	    printf("error in converting SDS objects %s.\n",sdsname);
	    SDendaccess(sds_id );
	    SDend(sd_id);
	    H5Fclose((hid_t)outFid);
	    return CONVERT_FAIL;
	  }
       }
    }

    SDendaccess(sds_id );
  }

  SDend(sd_id);
  H5Fclose((hid_t)outFid);
  
  return CONVERT_SUCCESS;
}



/* Function ZeroMemory
   Purpose: Zero out memory
   return:  None
   In: size_t n(DWORD in windows)
      void* s(PVOID in windows)
*/


void ZeroMemory(void*s,size_t n) {
#ifdef WIN32
  ZeroMemory(s,n);
#else
  bzero(s,n);
#endif
}


/*------------------------------------------------------------------
 * Function:    get_h5groupid(char*,hid_t)
 *
 * Purpose:     get HDF5 group id
 *              
 * Return:      HDF5 group id
 *
 * In :	        HDF5 group name
                handler of outFid        
               
   Out:
   Modification:         
 *-------------------------------------------------------------------
 */
hid_t get_h5groupid(const char* h5groupname,hid_t outFid) {

  hid_t h5_file; /* HDF5 file id */
  hid_t h5_group = -1; /* HDF5 group id */
  int check_groupname; /* flag to check absolute path of the group */


  h5_file = outFid;

  h5_group = H5Gopen(h5_file,h5groupname); 
  if(h5_group < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("failed to open HDF5 group %s\n", h5groupname);
	}
      return CONVERT_FAIL;	
    }
  return h5_group;
}


/*-------------------------------------------------------------------------
 * Function:	CopyDimscalesdsToHe5
 *
 * Purpose:     translate a dimension scale SDS object into an HDF-EOS5 dimscale (with attributes)
 *              
 * Return:	CONVERT_FAIL if failed, CONVERT_CONVERT_SUCCESS if successful.
 *
 * In :	        
 outFid:      h5 identifier
 sds_id:      SDS identifier
 h5groupname: hdf5 group name(must be provided)
 
 *-------------------------------------------------------------------------
 */	 

int CopyDimscalesdsToHe5(char *outFileName, int32 sds_id)
{
  int status = CONVERT_SUCCESS; 
  /* HDF4 variables. */
  int32   sds_dtype; /* SDS data type */
  int32   sds_rank; /* rank of SDS */
  int32   sds_dimsizes[HE4HE5_MAX_VAR_DIMS]; /* array to describe dimensional 
					 size of SDS */
  int32*  sds_start; /* The starting point of SDS array to convert */
  int32*  sds_edge; /* The ending point of SDS array to convert */
  int32*  sds_stride; /* The stride of SDS array to convert */
  int32   count_sdsdata; /* The count of SDS array */
  int32   sds_ref; /* SDS reference number */
  intn    sds_empty; /* the variable to check whether SDS is empty. */
  int32   istat; /* check the "return" status  of HDF4 APIs */
  int     i; /* temporary variable to represent the index of for loop */
  int     check_sds; /* temporary variable to check SDS */
  int32   num_sdsattrs; /* number of SDS attributes */
  void*   sds_data; /* buffer to store SDS data */

  int     check_sdsname; /* temporary variable to check the confliction 
			    of SDS name */
  int     check_gloattr; /* temporary variable to check the SD interface
			    variable */
  
  char    sdsname[HE4HE5_MAX_NC_NAME]; /* SDS name */
  char    sdslabel[HE4HE5_MAX_NC_NAME]; /* SDS label */
  size_t  h4size; /* file size of the SDS data type */
  size_t  h4memsize; /* memory size of the SDS data type */
  size_t  bufsize;
 

  /* varibles for hdf5. */
  hid_t   h5_group; /* HDF5 group id */
  hid_t   h5dset; /* HDF5 dataset id */
  hid_t   h5d_sid; /* HDF5 data space id */
  hid_t   h5ty_id; /* HDF5 data type(in disk) id */
  hid_t   h5_memtype; /* HDF5 data type(in memory) id */
  hid_t   create_plist; /* creation property list of dataset*/
  hid_t   write_plist;
  hsize_t h5dims[HE4HE5_MAX_VAR_DIMS]; /* an array of size of each dimension of HDF5
				   dataset converted from SDS*/
  hsize_t max_h5dims[HE4HE5_MAX_VAR_DIMS]; /* the array of maximum size of each
				       dimension of HDF5 dataset converted from SDS */
  char*   h5csds_name; /* absolute path of HDF5 dataset converted from SDS */
  char*   ori_h5csds_name; /* The absolute path of the original HDF5 dataset
			      if the current SDS is visited */
  herr_t  ret; /* the temporary variable to check the "return" status of HDF5 APIs */
  int     check_lookup; /* variable to check the status of SDS look-up table */
  int     temp; /* temporary variable */
  int*    check_lookupptr; /* an integer pointer to pass the check_lookup value */
  hsize_t* chunk_dims;
  Hdf4DimScaleInfo DimscaleInfo1;
  Hdf4DimScaleInfo *DimscaleInfo;
  int    loopCount;
  temp =0;
  herr_t    reth5=CONVERT_SUCCESS;

  /*** part I: SDS object handling. ***/
 
  DimscaleInfo = &DimscaleInfo1;

  /* zeroing out the memory for sdsname and sdslabel.*/
  ZeroMemory(sdsname,HE4HE5_MAX_NC_NAME);
  ZeroMemory(sdslabel,HE4HE5_MAX_NC_NAME);

  /* check whether the SDS is empty. */
  if(SDcheckempty(sds_id,&sds_empty)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot run SDcheckempty routine\n");
	}
      return CONVERT_FAIL;
  }

   /*obtain name,rank,dimsizes,datatype and num of attributes of sds */
  if (SDgetinfo(sds_id,sdsname,&sds_rank,sds_dimsizes,&sds_dtype,
		&num_sdsattrs)==CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to get information of SDS object\n");
	}
      return CONVERT_FAIL;
  }

  if(sds_rank > 1) {/* this sds is not dimension scale */
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("The SDS %s is not dimension scale (has more than 1 dimension). Skipping it.\n");
	}
      return CONVERT_SUCCESS;
  }

  /* part II. obtain the hdf5  name. */
  /*obtaining reference number and name of h5 dataset 
    corresponding to sds. */

  sds_ref = SDidtoref(sds_id);
  /* Just for netcdf conversion */
  if(sds_ref == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot obtain SDS reference number\n");
	}
      return CONVERT_FAIL;
  }

   /* if sds_empty !=0, then it means that sds is empty (because FALSE = 0) . 
     We will not convert empty SDS at this time. If it is requested, it will 
     be implemented later */
  if(sds_empty != 0) return CONVERT_SUCCESS;
   
  h5csds_name= (char *)malloc((strlen(sdsname) +1) * sizeof(char));
  if(h5csds_name == NULL) {               
    return (CallocErrorMsg(__LINE__));
  }
  strcpy(h5csds_name,sdsname); 

  /* part III. obtain start,edge, stride and number of sds data. */
  sds_start     = malloc(sizeof(int32)*sds_rank);
  if(sds_start == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS starting array\n");
	}
    free(h5csds_name);
    return (CallocErrorMsg(__LINE__));
  }

  sds_edge      = malloc(sizeof(int32)*sds_rank);
  if(sds_edge == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS edging array\n");
	}
    free(sds_start);
    free(h5csds_name);
    return (CallocErrorMsg(__LINE__));
  }

  sds_stride    = malloc(sizeof(int32)*sds_rank);
  if(sds_stride == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS stride\n");
	}
    free(sds_start);
    free(sds_edge);
    free(h5csds_name);
    return (CallocErrorMsg(__LINE__));
  }

  count_sdsdata = 1;
  for (i=0;i<sds_rank;i++){
    sds_stride[i] = 1;
    sds_start[i]  = 0;
    sds_edge[i]   = sds_dimsizes[i];
    count_sdsdata = count_sdsdata*sds_dimsizes[i];
  }

  for (i=0;i<sds_rank;i++) {
    h5dims[i] = sds_edge[i]-sds_start[i];
    max_h5dims[i] = h5dims[i];
  }

  if(SDisrecord(sds_id)) max_h5dims[0] = H5S_UNLIMITED;

  /* Datatype conversion is done here since memory allocation size
     should be known first. */

  /* convert HDF4 data type to HDF5 data type. */
  if  (h4type_to_h5type(sds_dtype,&h5_memtype,&h4memsize,&h4size,
			&h5ty_id) == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert HDF4 data type to HDF5 data type\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(h5csds_name);
    return CONVERT_FAIL;
  }

  /* check whether the datatype is string, if we find string format,
     we will change them back into integer format.*/
  if (h5ty_id == H5T_STRING) {

    /* rechange string datatype into numerical datatype.*/
    if(h5string_to_int(sds_dtype,&h5_memtype,h4memsize,
		       &h5ty_id)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot translate H5T_STRING into int\n");
	}
      free(sds_start);
      free(sds_edge);
      free(sds_stride);
      free(h5csds_name);
      return CONVERT_FAIL;
    }
  }
 
  sds_data = malloc(h4memsize*count_sdsdata);
  if(sds_data == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS data\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(h5csds_name);
    return (CallocErrorMsg(__LINE__));
  }

  istat    = SDreaddata(sds_id, sds_start, sds_stride, sds_edge, 
			(VOIDP)sds_data);
  if (istat == CONVERT_FAIL)  { 
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot read SDS data\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    free(h5csds_name);
    return CONVERT_FAIL;
  }

  /***************************************************************************
  If there are attributes..... get them too
  ***************************************************************************/
  if (num_sdsattrs > 0)
    {
      DimscaleInfo->totalNumAttr = num_sdsattrs;
      strcpy(DimscaleInfo->name, h5csds_name);
      DimscaleInfo->sds_buffsize = h4memsize*count_sdsdata;
      DimscaleInfo->sds_data = (void *)sds_data;
      DimscaleInfo->sds_dtype = sds_dtype; /*hdf4 datype for dimscale */
      DimscaleInfo->dimsize = sds_dimsizes[0];

      if ((reth5=getHDF4Dimscaleattrs(sds_id, DimscaleInfo)) != CONVERT_SUCCESS) {
	free(sds_start);
	free(sds_edge);
	free(sds_stride);
	free(sds_data);
	free(h5csds_name);
	return CONVERT_FAIL;
      }
    }
  /* we now have SDS dimscale name, values, and attrributes. we will
  /* be converting SDS dimscale into hdfeos5 dataset dimscale. */

  status = ConvertHDF4DimScales(outFileName, DimscaleInfo);

  /*
    free all allocated memeory in DimscaleInfo for attributes here
 
   */ 
  for (loopCount = 0; loopCount < DimscaleInfo->totalNumAttr ; loopCount++)
    {
      if(DimscaleInfo->sds_attdata[loopCount] !=NULL)
	{
	  free(DimscaleInfo->sds_attdata[loopCount]);
	  DimscaleInfo->sds_attdata[loopCount] = NULL;
	}

      if(DimscaleInfo->sdsattr_name[loopCount] !=NULL)
	{
	  free(DimscaleInfo->sdsattr_name[loopCount]);
	  DimscaleInfo->sdsattr_name[loopCount]= NULL;
	}
    }

  if(h5csds_name != NULL) free(h5csds_name);
  free(sds_start);
  free(sds_edge);
  free(sds_stride);
  free(sds_data);
  return CONVERT_SUCCESS;
}



/*-------------------------------------------------------------------------
 * Function:	CopysdsToHe5
 *
 * Purpose:     translate a SDS object into an HDF5 dataset (with attributes)
 *              
 * Return:	CONVERT_FAIL if failed, CONVERT_CONVERT_SUCCESS if successful.
 *
 * In :	        
 outFid:      h5 identifier
 sds_id:      SDS identifier
 h5groupname: hdf5 group name(must be provided)
 
 *-------------------------------------------------------------------------
 */	 

int CopysdsToHe5(hid_t outFid,
		 int32 sds_id,
		 char* h5groupname)
{
  /* HDF4 variables. */
  int32   sds_dtype; /* SDS data type */
  int32   sds_rank; /* rank of SDS */
  int32   sds_dimsizes[HE4HE5_MAX_VAR_DIMS]; /* array to describe dimensional 
					 size of SDS */
  int32*  sds_start; /* The starting point of SDS array to convert */
  int32*  sds_edge; /* The ending point of SDS array to convert */
  int32*  sds_stride; /* The stride of SDS array to convert */
  int32   count_sdsdata; /* The count of SDS array */
  int32   sds_ref; /* SDS reference number */
  intn    sds_empty; /* the variable to check whether SDS is empty. */
  int32   istat; /* check the "return" status  of HDF4 APIs */
  int     i; /* temporary variable to represent the index of for loop */
  int     check_sds; /* temporary variable to check SDS */
  int32   num_sdsattrs; /* number of SDS attributes */
  void*   sds_data; /* buffer to store SDS data */

  int     check_sdsname; /* temporary variable to check the confliction 
			    of SDS name */
  int     check_gloattr; /* temporary variable to check the SD interface
			    variable */
  
  char    sdsname[HE4HE5_MAX_NC_NAME]; /* SDS name */
  char    sdslabel[HE4HE5_MAX_NC_NAME]; /* SDS label */
  size_t  h4size; /* file size of the SDS data type */
  size_t  h4memsize; /* memory size of the SDS data type */
  size_t  bufsize;
 

  /* varibles for hdf5. */
  hid_t   h5_group; /* HDF5 group id */
  hid_t   h5dset; /* HDF5 dataset id */
  hid_t   h5d_sid; /* HDF5 data space id */
  hid_t   h5ty_id; /* HDF5 data type(in disk) id */
  hid_t   h5_memtype; /* HDF5 data type(in memory) id */
  hid_t   create_plist; /* creation property list of dataset*/
  hid_t   write_plist;
  hsize_t h5dims[HE4HE5_MAX_VAR_DIMS]; /* an array of size of each dimension of HDF5
				   dataset converted from SDS*/
  hsize_t max_h5dims[HE4HE5_MAX_VAR_DIMS]; /* the array of maximum size of each
				       dimension of HDF5 dataset converted from SDS */
  char*   h5csds_name; /* absolute path of HDF5 dataset converted from SDS */
  char*   ori_h5csds_name; /* The absolute path of the original HDF5 dataset
			      if the current SDS is visited */
  herr_t  ret; /* the temporary variable to check the "return" status of HDF5 APIs */
  int     check_lookup; /* variable to check the status of SDS look-up table */
  int     temp; /* temporary variable */
  int*    check_lookupptr; /* an integer pointer to pass the check_lookup value */
  hsize_t* chunk_dims; 


  temp =0;

  /*** part I: SDS object handling. ***/
  
  /* zeroing out the memory for sdsname and sdslabel.*/
  ZeroMemory(sdsname,HE4HE5_MAX_NC_NAME);
  ZeroMemory(sdslabel,HE4HE5_MAX_NC_NAME);

  /* check whether the SDS is empty. */
  if(SDcheckempty(sds_id,&sds_empty)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot run SDcheckempty routine\n");
	}
      return CONVERT_FAIL;
  }

  /*obtain name,rank,dimsizes,datatype and num of attributes of sds */
  if (SDgetinfo(sds_id,sdsname,&sds_rank,sds_dimsizes,&sds_dtype,
		&num_sdsattrs)==CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to get information of SDS object\n");
	}
      return CONVERT_FAIL;
  }

  /* Due to the bug in SDcheckempty(), sds_empty may not be correct.
   * In particular, sds_empty can be zero even when sds_dimsizes[0] is zero.
   * So, sds_empty has the second chance to be non-zero, here.
   * This is related to the bugzilla #1382.
   * This second chance is not necessary for HDF4 later than 4.2r3.
   */
  /*
  if (sds_empty == 0)
    sds_empty = sds_dimsizes[0] == 0;
  */

  /* part II. obtain the hdf5  name. */
  /*obtaining reference number and name of h5 dataset 
    corresponding to sds. */

  sds_ref = SDidtoref(sds_id);
  /* Just for netcdf conversion */
  if(sds_ref == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot obtain SDS reference number\n");
	}
      return CONVERT_FAIL;
  }

  /* obtain h5_group id */
  h5_group    =  get_h5groupid(h5groupname,outFid);
  if(h5_group < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot obtain HDF5 group ID for %s\n", h5groupname);
	}
    return CONVERT_FAIL;
  }
  /* if sds_empty !=0, then it means that sds is empty (because FALSE = 0) . 
     We will not convert empty SDS at this time. If it is requested, it will 
     be implemented later */
  if(sds_empty != 0) return CONVERT_SUCCESS;
   
  h5csds_name= (char *)malloc((strlen(sdsname) +1) * sizeof(char));
  if(h5csds_name == NULL) {               
    return (CallocErrorMsg(__LINE__));
  }
  strcpy(h5csds_name,sdsname); 

  /* part III. obtain start,edge, stride and number of sds data. */
  sds_start     = malloc(sizeof(int32)*sds_rank);
  if(sds_start == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS starting array\n");
	}
    free(h5csds_name);
    H5Gclose(h5_group);
    return (CallocErrorMsg(__LINE__));
  }

  sds_edge      = malloc(sizeof(int32)*sds_rank);
  if(sds_edge == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS edging array\n");
	}
    free(sds_start);
    free(h5csds_name);
    H5Gclose(h5_group);
    return (CallocErrorMsg(__LINE__));
  }

  sds_stride    = malloc(sizeof(int32)*sds_rank);
  if(sds_stride == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS stride\n");
	}
    free(sds_start);
    free(sds_edge);
    free(h5csds_name);
    H5Gclose(h5_group);
    return (CallocErrorMsg(__LINE__));
  }

  count_sdsdata = 1;
  for (i=0;i<sds_rank;i++){
    sds_stride[i] = 1;
    sds_start[i]  = 0;
    sds_edge[i]   = sds_dimsizes[i];
    count_sdsdata = count_sdsdata*sds_dimsizes[i];
  }

  for (i=0;i<sds_rank;i++) {
    h5dims[i] = sds_edge[i]-sds_start[i];
    max_h5dims[i] = h5dims[i];
  }

  if(SDisrecord(sds_id)) max_h5dims[0] = H5S_UNLIMITED;

  /* Datatype conversion is done here since memory allocation size
     should be known first. */

  /* convert HDF4 data type to HDF5 data type. */
  if  (h4type_to_h5type(sds_dtype,&h5_memtype,&h4memsize,&h4size,
			&h5ty_id) == CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot convert HDF4 data type to HDF5 data type\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  /* check whether the datatype is string, if we find string format,
     we will change them back into integer format.*/
  if (h5ty_id == H5T_STRING) {

    /* rechange string datatype into numerical datatype.*/
    if(h5string_to_int(sds_dtype,&h5_memtype,h4memsize,
		       &h5ty_id)== CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot translate H5T_STRING into int\n");
	}
      free(sds_start);
      free(sds_edge);
      free(sds_stride);
      free(h5csds_name);
      H5Gclose(h5_group);
      return CONVERT_FAIL;
    }
  }
 
  sds_data = malloc(h4memsize*count_sdsdata);
  if(sds_data == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for SDS data\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(h5csds_name);
    H5Gclose(h5_group);
    return (CallocErrorMsg(__LINE__));
  }

  istat    = SDreaddata(sds_id, sds_start, sds_stride, sds_edge, 
			(VOIDP)sds_data);
  if (istat == CONVERT_FAIL)  { 
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot read SDS data\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  /* part IV. converting SDS into hdf5 dataset. */

  h5d_sid = H5Screate_simple(sds_rank,h5dims,max_h5dims);	
  if (h5d_sid < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot create HDF5 data space\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data); 
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;							      
  }									      
	
  /* create property list. */

  /*create_plist = get_chunking_plistid(sds_id);*/

  create_plist = H5Pcreate(H5P_DATASET_CREATE);
  if(create_plist == -1) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to create property list\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data); 
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  chunk_dims   = malloc(sizeof(hsize_t)*sds_rank);
  if(chunk_dims == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to allocate memory for chunk_dims\n");
      }
    H5Pclose(create_plist);
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data); 
    free(h5csds_name);
    H5Gclose(h5_group);
    return (CallocErrorMsg(__LINE__));
  }
  for(i=0;i<sds_rank;i++){
    chunk_dims[i] = (hsize_t)(sds_dimsizes[i]);
  }
  if(H5Pset_chunk(create_plist, sds_rank, chunk_dims)<0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("unable to set chunk property list\n");
      }
    H5Pclose(create_plist);
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data); 
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  if(create_plist < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot generate HDF4 chunking list \n");
	}
    H5Pclose(create_plist);
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }

  /*
  if (add_fillvalue(sds_id,create_plist,h4memsize,h5_memtype) < 0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot add fill value to the chunking list \n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;
  }
  */

  h5dset = H5Dcreate(h5_group,h5csds_name,h5ty_id,h5d_sid,create_plist); 
  if (h5dset < 0) {	
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot create HDF5 dataset converted from SDS \n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    H5Sclose(h5d_sid);
    H5Pclose(create_plist);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;	
  }									      


  write_plist = H5Pcreate(H5P_DATASET_XFER);
  bufsize = h4memsize;
  for(i=0;i<sds_rank;i++)
    bufsize *= h5dims[i];

  if(H5Pset_buffer(write_plist,bufsize,NULL,NULL)<0) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to set HDF5 buffer for property list\n");
	}
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    H5Sclose(h5d_sid);
    H5Pclose(create_plist);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;		
  }

  if (H5Dwrite(h5dset,h5_memtype,h5d_sid,h5d_sid,write_plist,	    
	       (void *)sds_data)<0) {	
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to write HDF5 dataset\n");
	}
    H5Sclose(h5d_sid);
    H5Dclose(h5dset);
    H5Pclose(create_plist);
    H5Pclose(write_plist);
    free(sds_start);
    free(sds_edge);
    free(sds_stride);
    free(sds_data);
    free(h5csds_name);
    H5Gclose(h5_group);
    return CONVERT_FAIL;							      
  }							      
  
  /*  if (H5Dwrite(h5dset,h5_memtype,h5d_sid,h5d_sid,H5P_DEFAULT,	    
      (void *)sds_data)<0) {	
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot write HDF5 dataset \n");
	  }
      H5Sclose(h5d_sid);
      H5Dclose(h5dset);
      H5Pclose(create_plist);
      free(sds_start);
      free(sds_edge);
      free(sds_stride);
      free(sds_data);	
      free(h5csds_name);
      H5Gclose(h5_group);
      return CONVERT_FAIL;							      
      }
  
  */  								      
  free(sds_start);
  free(sds_edge);
  free(sds_stride);
  free(sds_data);
  free(chunk_dims);
  H5Pclose(write_plist);

  printf("Done: HDF4 written SDS %s converted\n", h5csds_name);

  /*obtain name,rank,dimsizes,datatype and num of attributes of sds */
  if (SDgetinfo(sds_id,sdsname,&sds_rank,sds_dimsizes,&sds_dtype,
		&num_sdsattrs)==CONVERT_FAIL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("unable to get information of SDS object\n");
	}
      return CONVERT_FAIL;
  }
  
  /***************************************************************************
  If there are attributes......convert.
  ***************************************************************************/
  if (num_sdsattrs > 0)
    {
      if (sds_HDF4convertattrs(h5_group, sds_id,h5csds_name,num_sdsattrs
			       )!=CONVERT_SUCCESS) {
	return CONVERT_FAIL;
      }
    }

  free(h5csds_name);
  return CONVERT_SUCCESS;
}



/*-------------------------------------------------------------------
 * Function:	h5string_to_int
 *
 * Purpose:	This function will convert H5T_STRING into integer.
                This is a correction routine when the user defines the 
		numerical datatype int8 as DFNT_CHAR8 and DFNT_UCHAR8

 * Errors:      will return error message to the interface
 * Return:	FAIL if failed, CONVERT_SUCCESS if success
  *
 * In :	        h4type: HDF4 datatype
                h4memsize: the real memory size of h4type

 * Out:         h5memtype: pointer of which value should be h5memtype
                (the real data type stored at the memory)
                h5type:    pointer of which value should be h5type
		(the HDF5 type stored at the disk).
 *
 *--------------------------------------------------------------*/
		
	
int h5string_to_int(const int32 h4type, 
		    hid_t* h5memtype,
		    const size_t h4memsize,
		    hid_t* h5type) {

   switch(h4type) {

  case DFNT_CHAR8:

    *h5type = H5T_STD_I8BE;
    if (h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype = H5T_NATIVE_SCHAR;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_SHORT;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_INT;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_LONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert HDF4 character into integer\n");
	}
      return CONVERT_FAIL;
    }
    break;

  case DFNT_UCHAR8:
     
    *h5type = H5T_STD_U8BE;
    if (h4memsize == H5Tget_size(H5T_NATIVE_CHAR))
      *h5memtype =  H5T_NATIVE_UCHAR;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_SHORT))
      *h5memtype = H5T_NATIVE_USHORT;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_INT))
      *h5memtype = H5T_NATIVE_UINT;
    else if(h4memsize == H5Tget_size(H5T_NATIVE_LONG))
      *h5memtype = H5T_NATIVE_ULONG;
    else {
      if (verboseModeGlobal == CONVERT_TRUE)
	{
	  printf("cannot convert HDF4 character into integer\n");
	}
      return CONVERT_FAIL;
    }
    break;
  }
  return CONVERT_SUCCESS;
}



/*------------------------------------------------------------------
 * Function:    get_h5datasetname
 *
 * Purpose:     get HDF5 dataset name
 *              
 * Return:      The absolute path of HDF5 dataset name
 *
 * In :	       
                char* h5groupname: absolute path of HDF5 group
                char*h5dsetname:   relative HDF5 dataset name to 
                                   h5groupname
                obj_ref:           object reference number
                char* HDF4obj:     constant HDF4 object identifier
		                   (HDF4_SDS)
                int* check_nameptr: integer pointer to check
		                    whether the name is used or not
               
   Out:
   Modification:         
 *-------------------------------------------------------------------
 */

char* get_h5datasetname(
			const char* h5groupname,
                        const char* h5dsetname, 
			uint16 obj_ref,
                        const char* hdf4obj,
			int tablesize,
			int* check_nameptr) {

  char  refstr[MAXREF_LENGTH]; /* the string format of HDF4 reference number */
  char* cor_h5dsetname = NULL; /* the corrected HDF5 dataset name */
  char* h5cdataset_name = NULL; /* The absolute path of HDF5 dataset name */

   /* change format of reference number. */
  if(conv_int_str(obj_ref,refstr)== CONVERT_FAIL) {
    
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("failed to convert reference number into string");
	}
    return NULL;
  }
  
  /* obtaining HDF5 dataset name, changing "/" in the dsetname into "_". */
  if(h5dsetname != NULL) {
    cor_h5dsetname = correct_name(h5dsetname);
    if(cor_h5dsetname == NULL) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("failed to get HDF5 dataset name");
	}
      return NULL;
    }
  }
  if(hdf4obj != NULL) 
    h5cdataset_name= get_obj_aboname(cor_h5dsetname,refstr,
				     h5groupname,hdf4obj,/*
				     temph4toh5id->name_hashtab,*/tablesize,
				     check_nameptr);
   if(h5dsetname != NULL)
     free(cor_h5dsetname);
  return h5cdataset_name;  
}



/*---------------------------------------------------------------
 * Function:	conv_int_str
 *
 * Purpose:	this function will convert numerical number into the 
                string format for a reference.
 * Return:	SUCCEED if success, FAIL if failed.
  *
 * In :	        num: an unsigned digital number.
                  
 * Out:         str_num: character string format of the number. 

 *	     
 *
 *-------------------------------------------------------------------------
 */			

int conv_int_str(
		 uint32 num, 
		 char* str_num) {
   
  if(str_num == NULL) {
    
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("not enough memory for allocating str_num");
	}
    return CONVERT_FAIL;
  }

  /*  Adding this line will cause problems, investigating this later.
      ZeroMemory(str_num,strlen(str_num)+1);*/
  sprintf(str_num,"%u",(unsigned int)num);
  return CONVERT_SUCCESS;
}


/*-----------------------------------------------------------------
 * Function:    correct_name
 *
 * Purpose:     modify the HDF4 object name when the name contains 
                '/'. Change this character into '_'.
                
 *              
 * Return:      the corrected name
 *
 * In :	        old name
                h4toh5id:  h4toh5 identifier
 
 *-----------------------------------------------------------------
 */	  
char *correct_name(const char* oldname){

  char * cptr; /* temproary pointer that stores the address
                  of the string where the ORI_SLASH is located. */
  char * newname; /* the new name after changing the ORI_SLASH to
                     CHA_SLASH. */
  if(oldname == NULL) {
   if (verboseModeGlobal == CONVERT_TRUE)
     {
       printf("inputting name is wrong");
     }
    return NULL;
  }

  newname = malloc(strlen(oldname)+1);
  if(newname == NULL) {
    
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("no enough memory for allocating newname");
	}
    return NULL;
  }

  ZeroMemory(newname,strlen(oldname)+1);

  newname = strncpy(newname, oldname, strlen(oldname));
  while(strchr(newname,ORI_SLASH)!= NULL){
    cptr = strchr(newname,ORI_SLASH);
    *cptr = CHA_SLASH;
  }

  return newname;
}




char *correct_name2(const char* oldname){

  char * cptr; /* temproary pointer that stores the address
                  of the string where the ORI_SLASH is located. */
  char * newname; /* the new name after changing the ORI_SLASH to
                     CHA_SLASH. */
  if(oldname == NULL) {
   if (verboseModeGlobal == CONVERT_TRUE)
     {
       printf("inputting name is wrong");
     }
    return NULL;
  }

  newname = malloc(strlen(oldname)+1);
  if(newname == NULL) {
    
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("no enough memory for allocating newname");
	}
    return NULL;
  }

  ZeroMemory(newname,strlen(oldname)+1);

  newname = strncpy(newname, oldname, strlen(oldname));
  while(strchr(newname,' ')!= NULL){
    cptr = strchr(newname,' ');
    *cptr = CHA_SLASH;
  }

  return newname;
}



/*-------------------------------------------------------------------------
 * Function:	get_obj_aboname
 *
 * Purpose:     get absolute path name of HDF5 object
                In this function, we will deal with name clashing.
		If we find an object name(by using lookup_updatename routine)
		that has already been used,
		we will create a new name for this object. We will follow
		object type(vgroup,sds,image,palette, vdata) plus reference
		number to make it unique.
 *              
 * Return:	NULL if failed, object name if successful.
 *
 * In :
                obj_name: relative object name 
		ref_str: reference number in character format
		path_name: absolute path 
		objstr: object type in character format
                name_hashtab: the name hash table that is checked
                tablesize:  the size of the table
                check_nameptr: pointer to check whether this HDF4 object
                name is used to create HDF5 object name
 
 *-------------------------------------------------------------------------
 */	

char* get_obj_aboname(
		      char* obj_name,
		      char* refstr,
		      const char* path_name,
		      const char*objstr,/*
		      struct name_table*name_hashtab,*/
		      int tablesize,
		      int* check_nameptr ) {
  char *abo_objname; /* absolute object name */
  int check_name;   /* a temporary flag to check name*/
  char check_char; /* a temporary flag to check character */

  check_name = 0;  

  /* sometimes although the object name is not NULL, but it is empty. 
     We will use make_objname_no in this case. */
  if(obj_name != NULL) check_char = *obj_name;

  /* obtain the absolute name of the object. */
  if (obj_name == NULL || check_char == '\0')
    abo_objname = make_objname_no(refstr,path_name,objstr);
  else 
    abo_objname = make_objname_yes(obj_name,path_name);
 
  if(abo_objname == NULL) {
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("abosolute path object name is NULL");
	}
    return NULL;
  }

  /* look up the name and see whether there is name clashing here.
     if yes, remake the object name.*/
  /*check_name = lookup_name(abo_objname,tablesize,name_hashtab);*/
  check_name = 0;
  *check_nameptr = check_name;
  if(check_name == 1) {
    /* name_clashing is found. We use make_objname_no to build up
       new abo_objname.*/

    if(objstr != NULL && refstr != NULL){
      free(abo_objname);
      abo_objname = make_objname_no(refstr,path_name,objstr);
    }
  }
 
  return abo_objname;
}



/*-------------------------------------------------------------------
 * Function:	make_objname_no
 *
 * Purpose:     make absolute path name of HDF5 object when object 
                name is not defined. We will use path name and 
		object type(vgroup,SDS,image,palette, vdata) 
		plus reference number to make the new name unique.
 *              
 * Return:	NULL if failed, object name if successful.
 *
 * In :	        
 		ref_str:   reference number in character format
		path_name: absolute path 
		objstr:    object type in character format
 
 *-------------------------------------------------------------------
 */	

char* make_objname_no(
		      char* refstr,
		      const char* path_name,
		      const char*objstr)
{

  char *new_objname; /* new object name */
  
  char * root_path;

  root_path =malloc(2);
  ZeroMemory(root_path,2);
  strcpy(root_path,"/");

  if(objstr == NULL || refstr == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("object type and ref. number should be defined.");
      }
    return NULL;
  }

  if (path_name == NULL || strcmp(path_name,root_path) == 0) {/* under root group. */
      
    new_objname= malloc(strlen(objstr)+strlen(refstr)+3);
    if(new_objname == NULL) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for object name");
	}
      return NULL;
    }
    ZeroMemory(new_objname,strlen(objstr)+strlen(refstr)+strlen(root_path)+1);
    strcpy(new_objname,root_path);
    strcat(new_objname,objstr);
    strcat(new_objname,"_");
    strcat(new_objname,refstr);
  }

  else {

    new_objname= malloc(strlen(path_name)+strlen(objstr)
			+strlen(refstr)+3);
    if(new_objname == NULL) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate memory for new_objname");
	}
      return NULL;
    }
    ZeroMemory(new_objname,strlen(path_name)+
		      strlen(objstr)+strlen(refstr)+3);
    strcpy(new_objname,path_name);
    strcat(new_objname,"/");
    strcat(new_objname,objstr);
    strcat(new_objname,"_");
    strcat(new_objname,refstr);
  }
  free(root_path);
  return new_objname;
}

/*-------------------------------------------------------------------
 * Function:	make_objname_yes
 *
 * Purpose:     get absolute path name of HDF5 object when object 
                name is defined. 
               
 *              
 * Return:	NULL if failed, object name if successful.
 *
 * In :	        obj_name: object name
		path_name: absolute path 

 *-------------------------------------------------------------------
 */	

char* make_objname_yes(
		       char* obj_name,
		       const char* path_name){

  char*new_objname; /* new HDF5 object name */
  char* root_path; /* temporary variable to store root path "/" */

  root_path =malloc(2);
  ZeroMemory(root_path,2);
  strcpy(root_path,"/");

  if(path_name == NULL || strcmp(path_name,root_path) == 0) {
    /* under root group */
    new_objname = malloc(strlen(obj_name)+2);
    if(new_objname == NULL) {
       
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot make H5 String type");
	}
      return NULL;
    }
    ZeroMemory(new_objname,strlen(obj_name)+2);
    strcpy(new_objname,root_path);
    strcat(new_objname,obj_name);
  }
  
  else {
    new_objname = malloc(strlen(path_name)+strlen(obj_name)+2);
    if(new_objname == NULL) {
      
      if (verboseModeGlobal == CONVERT_TRUE)
        {
	  printf("cannot allocate enough memory for new_objname");
	}
      return NULL;
    }
    ZeroMemory(new_objname,strlen(path_name)
		      +strlen(obj_name)+2);
    strcpy(new_objname,path_name);
    strcat(new_objname,"/");
    strcat(new_objname,obj_name);
  }
  free(root_path);
  return new_objname;
}

/*------------------------------------------------------------------
 * Function:	trans_obj_name
 *
 * Purpose:     make HDF4 attribute name 
                The name will be HDF4 object type
                plus ATTR plus reference number.
 *              
 * Return:      object name;
 *
 * In :	        
                obj_tag:  HDF4 tag 
		index  :  HDF5 group id
 
 *------------------------------------------------------------------
 */	     
char* trans_obj_name(
		     int32 obj_tag,
		     int32 h4toh5_index) {

  char* obj_name; /* the generated HDF5 object name */
  char indstr[MAXREF_LENGTH]; /* index string format converted from
				 decimal index */
   /* the reason why we allocate memory with strlen(HDF4_PALETTE) is 
     HDF4_PALETTE is the longest string among HDF4_??? 
     number 8 is chosen "arbitrarily", which assures the maximum number 
     for the object name. */
  obj_name = malloc(strlen(HDF4_PALETTE)+strlen(ATTR)+8);
  if(obj_name == NULL) { 
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot allocate memory for object name");
      }
    return NULL;
  }

  ZeroMemory(obj_name,strlen(HDF4_PALETTE)+strlen(ATTR)+8);
  if(conv_int_str((uint32)h4toh5_index,indstr)== CONVERT_FAIL) { 
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot convert integer into string");
      }
    return NULL;
  }

  /*find the appropriate object tag and make an object name. */

  switch(obj_tag) {

  case DFTAG_SD:
  case DFTAG_NDG:
  case DFTAG_SDG:
    strcpy(obj_name,HDF4_SDS);
    break;

  case DFTAG_RIG:
  case DFTAG_RI:
  case DFTAG_RI8:
    strcpy(obj_name,HDF4_IMAGE);
    break;

  case DFTAG_VG:
    strcpy(obj_name,HDF4_VGROUP);
    break;

  case DFTAG_VH:
  case DFTAG_VS:
    strcpy(obj_name,HDF4_VDATA);
    break;

  case DFTAG_LUT: 
    strcpy(obj_name,HDF4_PALETTE);
    break;

  default: 
    printf("object tag cannot be transferred ");
    free(obj_name);
    return NULL;
  }
    
  strcat(obj_name,"_");
  strcat(obj_name,ATTR);
  strcat(obj_name,"_");
  strcat(obj_name,indstr);

  return obj_name;
}


/*-----------------------------------------------------------------
 * Function:	he4_transpredattrs
 *
 * Purpose:     translate predefined attributes into HDF5 attribute
 *              predefined attributes including HDF4 OBJECT TYPE, 
                HDF4 OBJECT NAME, HDF4 CLASS etc. They are all in
		H5T_STRING format.

 * Return:	FAIL if failed, SUCCEED if successful.
 *
 * In :	        
                h4toh5id:  h4toh5 identifier
                h5g: group id
		attrname: attribute name
		data: attribute data
 
 *-----------------------------------------------------------------
*/	   
int he4_transpredattrs(
		      hid_t h5g,
		      const char *attrname,
		      char* data){

  size_t    h5str_size; /* HDF5 string size */
  hid_t     h5a_id; /* HDF5 attribute id */
  hid_t     h5a_sid; /* HDF5 attribute space id */
  hid_t     h5str_type; /* HDF5 string type */
  herr_t    ret; /* temporary variable to check the "return" status of HDF5 */

  if(data == NULL) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("predefined attribute data is not available");
      }
    return CONVERT_FAIL;
  }
  
  h5str_size = strlen(data);
  if ((h5str_type = mkstr(h5str_size,H5T_STR_SPACEPAD))<0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot make H5 String type for h4 predefined attribute");
      }
    return CONVERT_FAIL;
  }
  
  h5a_sid = H5Screate(H5S_SCALAR);
  if (h5a_sid < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create attribute space");
      }
    return CONVERT_FAIL;
  }
  
  h5a_id = H5Acreate_safe(h5g,attrname,h5str_type,h5a_sid,H5P_DEFAULT);
  if(h5a_id <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create attribute");
      }
    H5Sclose(h5a_sid);
    return CONVERT_FAIL;
  }
  
  ret = H5Awrite(h5a_id,h5str_type,(void *)data);
  if(ret <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot write attribute");
      }
    H5Aclose(h5a_id);
    H5Sclose(h5a_sid);
    return CONVERT_FAIL;
  }
  ret = H5Sclose(h5a_sid);
  if (ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot close data space id");
      }
    H5Aclose(h5a_id);
    return CONVERT_FAIL;
  }
  ret = H5Aclose(h5a_id);
  if (ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot close attribute interface");
      }
    return CONVERT_FAIL;
  }
  return CONVERT_SUCCESS;
}



/*-------------------------------------------------------------------
 * Function:	h4_transnumattr
 *
 * Purpose:     translate HDF4 object reference number into an HDF5 attribute
 *              
 * Return:	FAIL if failed, SUCCEED if successful.
 *
 * In :	        
                h5g: HDF5 group id
		refname: reference name
		group_ref: HDF4 reference number
		h4toh5id:  h4toh5 identifier
 
 *----------------------------------------------------------------
*/
    
int h4_transnumattr( 
		    hid_t h5g,
		    const char *refname,
		    uint16 group_ref) {

  hid_t    h5memtype; /* HDF5 memory data type. */
  hid_t    h5a_id; /* HDF5 attribute id */
  hid_t    h5a_sid; /* HDF5 attribute space id. */
  herr_t   ret; /* variable to check HDF5 "return" status */

  /* create attribute space id and attribute id */
  h5a_sid = H5Screate(H5S_SCALAR);

  if (h5a_sid < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create attribute space for HDF4_REF_NUM");
      }
    return CONVERT_FAIL;
  }

  h5a_id = H5Acreate_safe(h5g,refname,H5T_STD_U16BE,h5a_sid,H5P_DEFAULT);
  if(h5a_id <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot create attribute id for HDF4_REF_NUM");
      }
    H5Sclose(h5a_sid);
    return CONVERT_FAIL;
  }

  /* data type conversion from HDF4 uint16 to HDF5 */
  if(H5Tget_size(H5T_NATIVE_CHAR)== sizeof(uint16))
    h5memtype =  H5T_NATIVE_UCHAR;
  else if(H5Tget_size(H5T_NATIVE_SHORT)== sizeof(uint16))
    h5memtype = H5T_NATIVE_USHORT;
  else if(H5Tget_size(H5T_NATIVE_INT) == sizeof(uint16))
    h5memtype = H5T_NATIVE_UINT;
  else if(H5Tget_size(H5T_NATIVE_LONG)== sizeof(uint16))
    h5memtype = H5T_NATIVE_ULONG;
  else {
    H5Sclose(h5a_sid);
    H5Aclose(h5a_id);
    return CONVERT_FAIL;
  }

  ret = H5Awrite(h5a_id,h5memtype,(void *)&group_ref);

  if(ret <0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot write attribute");
      }
    H5Sclose(h5a_sid);
    H5Aclose(h5a_id);
    return CONVERT_FAIL;
  }

  ret = H5Sclose(h5a_sid);
  if( ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot close HDF5 data space ");
      }
    H5Aclose(h5a_id);
    return CONVERT_FAIL;
  }

  ret = H5Aclose(h5a_id);
  if(ret < 0) {
    if (verboseModeGlobal == CONVERT_TRUE)
      {
	printf("cannot close HDF5 attribute id ");
      }
    return CONVERT_FAIL;
  }

  return CONVERT_SUCCESS;
}

